/*
 *
 * Copyright 2010 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt;

import de.bsvrz.dav.daf.main.ClientDavInterface;

/**
 * Schnittstelle zur Verwaltung und Dokumentation der Zustände von Tasks und den Fortschritt von langwierigen Aktionen zu dokumentieren
 *
 * @author Kappich Systemberatung
 * @version $Revision: none $
 */
public interface LongTermTaskStatePublisher {

	/**
	 * Nimmt eine neue Task in die Verwaltung der Tasks auf. Diese Methode sollte von jeder Task einmal aufgerufen werden. Über das zurückgelieferte Task-Objekt
	 * kann die Task ihren Zustand und ihren Fortschritt mit entsprechenden Methoden bekannt geben.
	 * @param taskName Name der Task
	 * @return Task-Objekt über das die Task ihren Zustand und ihren Fortschritt mit entsprechenden Methoden bekannt geben kann.
	 */
	Task registerTask(final String taskName);

	/**
	 * Übergibt die Verbindung zum Datenverteiler. Nach Aufruf dieser Methode kann die Task-Verwaltung die Zustände der einzelnen Task über den Datenverteiler
	 * publizieren. Der Aufruf dieser Methode ist optional.
	 * @param davConnection
	 */
	void setDavConnection(ClientDavInterface davConnection);

	/**
	 * Schnittstelle, die von den einzelnen Tasks verwendet wird, um den Zustand der Task und den Fortschritt von langwierigen Aktionen zu dokumentieren.
	 */
	public interface Task {

		/**
		 * Setzt den Zustand einer Task.
		 * @param taskState Neuer Zustand der Task.
		 */
		void setTaskState(TaskState taskState);

		/**
		 * Setzt den Fortschritt einer Task. Im Falle einer Ausgabe werden der übergebene Formatstring und die weiteren Parameter an die Methode
		 * Formatter.format() übergeben um den auszugebenden Text zu ermitteln.
		 * @param format Formatstring, der bei Ausgaben dazu verwendet wird, eine textuelle Beschreibung des aktuellen Fortschritts zu erzeugen.
		 * @param progress Fortschritt der aktuellen Aktion oder <code>-1</code>, wenn der Fortschritt nicht bekannt ist.
		 * @param maximumProgress Maximum des Fortschritts einer Aktion oder <code>-1</code>, wenn das Maximum nicht bekannt ist.
		 * @param result Zwischenergebnis der aktuell durchgeführten Aktion oder <code>-1</code>, wenn kein Zwischenergebnis bekannt ist.
		 */
		void setTaskProgress(String format, long progress, long maximumProgress, long result);
	}

	public static enum TaskState {

		/**
		 * Initialer Zustand einer Task nach dem Erzeugen.
		 */
		INIT("neu"),

		/**
		 * Zustand einer Task, die momentan bereit für neue Aufgaben ist.
		 */
		IDLE("bereit"),

		/**
		 * Zustand einer Task, die soeben eine neue Aufgabe bekommen hat und sich auf die Bearbeitung vorbereitet.
		 */
		START_WORK("starten"),

		/**
		 * Zustand einer Task, die mit der Bearbeitung einer Aufgabe beschäftigt ist.
		 */
		WORKING("bearbeiten"),

		/**
		 * Zustand einer Task, die mit den abschließenden Arbeiten nach Bearbeitung einer Aufgabe beschäftigt ist.
		 */
		FINISH_WORK("abschließen"),

		/**
		 * Zustand einer Task, die terminiert wurde.
		 */
		TERMINATED("terminiert");

		/**
		 * Name des Zustands, der für Ausgaben etc. verwendet werden soll.
		 */
		private String _stateName;

		/**
		 * Der Konstruktor wird verwendet, um den in Ausgaben zu verwendenden Namen des Zustands zu übergeben.
		 * @param stateName Name des Zustands, der für Ausgaben etc. verwendet werden soll.
		 */
		private TaskState(final String stateName) {
			_stateName = stateName;
		}

		/**
		 * Liefert den Namen des Zustands zurück.
		 * @return Name des Zustands.
		 */
		public String getStateName() {
			return _stateName;
		}

		@Override
		public String toString() {
			return name() + "(" + _stateName + ")";
		}
	}

}
