/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.datatree;

import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTree.AspNode;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTree.AtgNode;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTree.ObjNode;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;

import java.util.Iterator;
import java.util.Map;


/**
 * Iterator zum Durchlaufen eines DataIdentTrees. Im Konstruktor muss der zu durchlaufende DataIdentTree uebergeben werden. Für jeden Iterationsschritt, auch
 * den ersten,  muss einmal getNext() aufgerufen werden. Wenn die Methode getNext() true zurueckliefert, ist die naechste Datenidentifikation des Baums
 * verfuegbar. Ihre zugehoerigen IDs koennen mit getObjId(), etc. abgefragt werden. Sobald getNext() false liefert ist der Durchlauf beendet und das
 * Iterator-Objekt muss verworfen werden. <br> Vorraussetzung: Alle Pfade im DataIdentTree muessen immer zu einem DataIdentNode fuehren. <br><br>
 * Beispiel:<br><br>
 * <blockquote><pre>
 * DataIdentTreeIterator it = new DataIdentTreeIterator(dataIdentTree);
 * while(it.getNext()) {
 *     myMethod(it.getObjId(), it.getAtgId(), it.getAspId(), it.getSV());
 * }
 * </pre></blockquote>
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schaefer
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class DataIdentTreeIterator {

	private boolean emptyTree;

	private long objId = 0;

	private ObjNode objNode;

	private Map<Long, ObjNode> objNodes;

	private Iterator<Long> objIdsIt;

	private long atgId = 0;

	private AtgNode atgNode;

	private Map<Long, AtgNode> atgNodes;

	private Iterator<Long> atgIdsIt;

	private long aspId = 0;

	private AspNode aspNode;

	private Map<Long, AspNode> aspNodes;

	private Iterator<Long> aspIdsIt;

	private int svId = 0;

	private DataIdentNode dataIdentNode;

	private Map<Integer, DataIdentNode> dataIdentNodes;

	private Iterator<Integer> svIt; // SV ist Key zum DIN

	/**
	 * Erstellt einen DataIdentTreeIterator zum einmaligen Durchlaufen des uebergebenen DataIdentTree.
	 *
	 * @param tree Zu durchlaufender DataIdentTree
	 */
	public DataIdentTreeIterator(DataIdentTree tree) {
		objNodes = tree.getRootNode().getObjMapCopy();
		objIdsIt = objNodes.keySet().iterator();
		emptyTree = !objIdsIt.hasNext();
		if(!emptyTree) {
			objId = objIdsIt.next();
			objNode = objNodes.get(objId);
			enterFirstAtgNode();
			enterFirstAspNode();
			dataIdentNodes = aspNode.getDinMapCopy();
			svIt = dataIdentNodes.keySet().iterator();
			// svIt.next() nicht hier sondern erst in getNext() aufrufen.
		}
	}

	/**
	 * Fuehrt einen Iterations-Schritt durch. Wenn die naechste Datenidentifikation des Baumes verfuegbar ist, wird true zurueckgeliefert, ansonsten false am
	 * Ende des kompletten Durchlaufs.
	 *
	 * @return Kennzeichen, ob naechste Datenidentifikation verfuegbar ist.
	 */
	public boolean getNext() {
		if(emptyTree) {
			return false;
		}
		if(svIt.hasNext()) {
			svId = svIt.next();
			dataIdentNode = dataIdentNodes.get(svId);
			return true;
		}
		else if(aspIdsIt.hasNext()) {
			aspId = aspIdsIt.next();
			aspNode = aspNodes.get(aspId);
			enterFirstDIN();
			return true;
		}
		else if(atgIdsIt.hasNext()) {
			atgId = atgIdsIt.next();
			atgNode = atgNodes.get(atgId);
			enterFirstAspNode();
			enterFirstDIN();
			return true;
		}
		else if(objIdsIt.hasNext()) {
			objId = objIdsIt.next();
			objNode = objNodes.get(objId);
			enterFirstAtgNode();
			enterFirstAspNode();
			enterFirstDIN();
			return true;
		}
		return false;
	}

	/**
	 * Liefert die Objekt-ID der Datenidentifikation.
	 *
	 * @return Objekt-ID
	 */
	public long getObjId() {
		return objId;
	}

	/**
	 * Liefert die Attributgruppen-ID der Datenidentifikation.
	 *
	 * @return Attributgruppen-ID
	 */
	public long getAtgId() {
		return atgId;
	}

	/**
	 * Liefert die Aspekt-ID der Datenidentifikation.
	 *
	 * @return Aspekt-ID
	 */
	public long getAspId() {
		return aspId;
	}

	/**
	 * Liefert die Simulationsvariante der Datenidentifikation.
	 *
	 * @return Simulationsvariante
	 */
	public int getSV() {
		return svId;
	}

	/**
	 * Liefert den DataIdentNode der Datenidentifikation.
	 *
	 * @return DataIdentNode
	 */
	public DataIdentNode getDataIdentNode() {
		return dataIdentNode;
	}

	public String toString() {
		return "objID=" + getObjId() + ", atgID=" + getAtgId() + ", aspID=" + getAspId() + ", sv=" + getSV();
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Private Methoden
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/** Wandert zum ersten Attributgruppen-Node des aktuellen Objekt-Node. */
	private void enterFirstAtgNode() {
		atgNodes = objNode.getAtgMapCopy();
		atgIdsIt = atgNodes.keySet().iterator();
		atgId = atgIdsIt.next();
		atgNode = atgNodes.get(atgId);
	}

	/** Wandert zum ersten Aspekt-Node des aktuellen Attributgruppen-Node. */
	private void enterFirstAspNode() {
		aspNodes = atgNode.getAspMapCopy();
		aspIdsIt = aspNodes.keySet().iterator();
		aspId = aspIdsIt.next();
		aspNode = aspNodes.get(aspId);
	}

	/** Wandert zur ersten DataIdentNode des aktuellen Aspekt-Node. */
	private void enterFirstDIN() {
		dataIdentNodes = aspNode.getDinMapCopy();
		svIt = dataIdentNodes.keySet().iterator(); // SV ist Key vom Aspekt-Node zum DIN.
		svId = svIt.next();
		dataIdentNode = dataIdentNodes.get(svId);
	}

	public IdDataIdentification getDataIdentification() {
		return new IdDataIdentification(getObjId(), getAtgId(), getAspId(), getSV());
	}
}
