/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.LongTermTaskStatePublisher;
import de.bsvrz.ars.ars.mgmt.datatree.DIdNodeNotFoundException;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.persistence.*;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.ars.ars.persistence.index.result.IndexResult;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dataSerializer.Serializer;
import de.bsvrz.sys.funclib.dataSerializer.SerializingFactory;
import de.bsvrz.sys.funclib.losb.datk.MessageType;
import de.bsvrz.sys.funclib.losb.datk.PidQuery;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.util.Calendar;
import java.util.concurrent.BlockingQueue;

/**
 * Task zur Bearbeitung von Loeschzeit-Verlaengerungsauftraegen.
 *
 * @author beck et al. projects GmbH
 * @author Christian Wied
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class LZVTask extends SingleClientSenderTask {

	private static final long CHANGETIME_SPONTAN = -1;
	public static final long CHANGETIME_REGULAR = 0;
	private static final long DELETE_IMMEDIATELY = -1;
	private static final long DELETE_LATER = 0;
	/** Um wieviele Monate maximal der Vorhaltezeitraum angepasst wird. */
	private static final int MAX_TIMESHIFTING = 2 * 12; //
	// Nur für Testzwecke: Löschschutz ignorieren
	public static boolean IGNORE_LOESCH_SCHUTZ = false;
	private static long MIN_REM_DSK_CAPA = 2 * 1000 * 1000 * 1000; // Default 2GB gemaess DatK
	private static CapacityCheckResultData capacityCheckResultData;
	private static DeleteRegularResultData deleteRegularResultData;
	private final LongTermTaskStatePublisher.Task _longTermTaskStatePublisherTask;
	private byte[] indexMem;
	private int numOfVisitedCont, numOfDeletedCont;

	;

	/**
	 * Erzeugt eine neue Loeschzeit-Verlaengerungs-Task.
	 *  @param rodl       Der Listener wird benachrichtigt beim Beenden der Task.
	 * @param archiveMgr Der Archiv-Manager.
	 * @param inputData  Eingangswarteschlange.
	 */
	public LZVTask(ArchiveManager archiveMgr,BlockingQueue<Object> inputData, final LongTermTaskStatePublisher longTermTaskStatePublisher) {
		super(archiveMgr, inputData);
		setName("LZV-Task");
		_longTermTaskStatePublisherTask = longTermTaskStatePublisher.registerTask("Löschen");
	}

	/**
	 * Liefert die Mindest-Restkapazitaet der Festplatte.
	 *
	 * @return Mindestrestkapazitaet in Byte
	 */
	public static long getMinRemainingDiskCapa() {
		return MIN_REM_DSK_CAPA;
	}

	/**
	 * Setzt die Mindest-Restkapazitaet der Festplatte.
	 *
	 * @param minRemainingDiskCapa Neue Mindestrestkapazitaet in Byte.
	 */
	public static void setMinRemainingDiskCapa(long minRemainingDiskCapa) {
		MIN_REM_DSK_CAPA = minRemainingDiskCapa;
	}

	public static CapacityCheckResultData getCapacityCheckResultData(ArchiveManager archMgr) {
		if(capacityCheckResultData == null) {
			Data data = archMgr.getDavCon().createData(archMgr.getAtg("atg.archivAnfrageSchnittstelle"));
			data.getReferenceValue(PidQuery.ATT_SENDER_NAME).setSystemObject(archMgr.getDavCon().getLocalApplicationObject());
			data.getItem(PidQuery.ATT_QUERY_IDX_NAME).asUnscaledValue().set(0);
			data.getUnscaledValue(PidQuery.ATT_MESSAGE_TYP_NAME).set(MessageType.DELETION_TIME);
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			try {
				Serializer serializer = SerializingFactory.createSerializer(bos);
				serializer.writeInt(serializer.getVersion());
				serializer.writeLong(CHANGETIME_SPONTAN);
			}
			catch(Exception e) {
//                logger.warning("Konnte lzvResultData nicht richtig initialisieren: ", e.getMessage());  // TODO Auskommentierter Code !?
			}
			data.getUnscaledArray(PidQuery.ATT_DATA_NAME).set(bos.toByteArray());

			capacityCheckResultData = new CapacityCheckResultData(null, null, 0, data);
		}
		return capacityCheckResultData;
	}

	public static DeleteRegularResultData getDeleteRegularResultData(ArchiveManager archMgr) {
		if(deleteRegularResultData == null) {
			Data data = archMgr.getDavCon().createData(archMgr.getAtg("atg.archivAnfrageSchnittstelle"));
			data.getReferenceValue(PidQuery.ATT_SENDER_NAME).setSystemObject(archMgr.getDavCon().getLocalApplicationObject());
			data.getItem(PidQuery.ATT_QUERY_IDX_NAME).asUnscaledValue().set(0);
			data.getUnscaledValue(PidQuery.ATT_MESSAGE_TYP_NAME).set(MessageType.DELETION_TIME);
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			try {
				Serializer serializer = SerializingFactory.createSerializer(bos);
				serializer.writeInt(serializer.getVersion());
				serializer.writeLong(CHANGETIME_REGULAR);
			}
			catch(Exception e) {
//                logger.warning("Konnte lzvResultData nicht richtig initialisieren: ", e.getMessage()); // TODO Auskommentierter Code !?
			}
			data.getUnscaledArray(PidQuery.ATT_DATA_NAME).set(bos.toByteArray());

			deleteRegularResultData = new DeleteRegularResultData(null, null, 0, data);
		}
		return deleteRegularResultData;
	}

	@Override
	public void execute() {
		_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		try {
			super.execute();
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.TERMINATED);
		}
	}

	/**
	 * Führt das Löschen, spontanes Löschen oder die Löschzeitverlängerung aus.
	 *
	 * @see de.bsvrz.ars.ars.mgmt.tasks.SingleTask#work(ResultData)
	 */
	protected void work(ResultData resultData) {
		if(!resultData.hasData()) return;
		try {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.START_WORK);
			indexMem = new byte[INDEX_MEMORY_SIZE];
			init(resultData.getData());
			changeDeletionTime(resultData);
			if(!(resultData instanceof CapacityCheckResultData) && !(resultData instanceof DeleteRegularResultData)) sendSuccessResponse();
			_debug.info(getName() + ": " + numOfVisitedCont + " Container besucht, " + numOfDeletedCont + " davon betroffen");
		}
		catch(Exception e) {
			_debug.error("Loeschzeitverlaengerung fehlgeschlagen", e);
			sendErrorResponse("Loeschzeitverlaengerung fehlgeschlagen: " + e.getMessage());
		}
		finally {
			indexMem = null;
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		}
	}

	/**
	 * Prüft, welche Operation (Löschen, spontanes Löschen oder die Löschzeitverlängerung) ausgeführt werden soll, und führt diese dann durch.
	 *
	 * @param resultData Datensatz mit Informationen über den Auftrag
	 * @throws Exception Fehler
	 */
	private void changeDeletionTime(ResultData resultData) throws Exception {
		_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.WORKING);
		try {
			long changeTime = deserializer.readLong();
			ArchiveDataSpecification[] ads = new ArchiveDataSpecification[0];
			if(!(resultData instanceof CapacityCheckResultData) && !(resultData instanceof DeleteRegularResultData)) {
				// nur externe Aufträge werden geparst
				ads = parseArchiveDataSpec(deserializer, resultData);
			}

			if(resultData instanceof CapacityCheckResultData) {
				deleteSpontan(ads);
			}
			else if(resultData instanceof DeleteRegularResultData) {
				// Reguläres löschen mit Betrachtung des Löschschutzes
				long t0 = System.currentTimeMillis();
				_debug.info("Reguläres Löschen");
				containerWalk(ads, new DeleteRegular());
				_debug.info("Reguläres Löschen beendet, Dauer: " + ((System.currentTimeMillis() - t0) / 1000) + " Sekunden");
			}
			else if(changeTime == DELETE_IMMEDIATELY) {
				// Sofortiges Löschen der angegebenen Daten ohne Betrachtung des Löschschutzes
				long t0 = System.currentTimeMillis();
				_debug.info("Sofortiges Löschen");
				containerWalk(ads, new DeleteNow());
				_debug.info("Sofortiges Löschen beendet, Dauer: " + ((System.currentTimeMillis() - t0) / 1000) + " Sekunden");
			}
			else if(changeTime == DELETE_LATER) {
				// Löschschutzaufhebung (späteres Löschen) der angegebenen Daten
				long t0 = System.currentTimeMillis();
				_debug.info("Späteres Löschen");
				containerWalk(ads, new DeleteLater());
				_debug.info("Späteres Löschen beendet, Dauer: " + ((System.currentTimeMillis() - t0) / 1000) + " Sekunden");
			}
			else if(changeTime > 0) {
				// Manuelle Löschschutzverlängerung
				containerWalk(ads, new ChangeTime(changeTime));
			}
			else {
				_debug.warning(getName() + ": Unbekannte Loeschzeit: " + changeTime);
				throw new Exception(getName() + ": Loeschzeitverlaengerung: Unbekannte Loeschzeit: " + changeTime);
			}
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.FINISH_WORK);
		}
	}

	/**
	 * Der <code>DeleteSpontanTask</code> löscht bei knapp werdendem Speicherplatz unwichtige Container.
	 * <p>
	 * Dazu ueberprueft er in einem parametrierbaren Zyklus, wieviel Speicherplatz dem Archivsystem noch zur Verfuegung steht. Erreicht die vorhandene Kapazität
	 * einen Schwellwert, werden folgende Daten geloescht: <ul> <li>archivierte Daten, deren Loeschschutz abgelaufen ist und die nicht (mehr) gesichert werden
	 * muessen,</li> <li>wiederhergestellte Daten,</li> <li>archivierte Simulationsdaten,</li> <li>archivierte Daten, deren Loeschschutz nicht abgelaufen ist,
	 * die aber schon gesichert wurden.</li> </ul>
	 */
	public void deleteSpontan(ArchiveDataSpecification[] ads) throws Exception {
		long freeDiskSpace = getFreeDiskSpace();
		Calendar compareDate = Calendar.getInstance();
		boolean someDataDeleted = false;

		for(DELETE_LEVEL level : DELETE_LEVEL.values()) {
			if(freeDiskSpace < getMinRemainingDiskCapa()) {
				someDataDeleted = true;
				switch(level) {
					// Verschieben des Vorhaltezeitraums
					case TIME_SHIFTING:
						for(int i = 0; i < MAX_TIMESHIFTING; i++) {
							compareDate.add(Calendar.MONTH, 1);
							if(freeDiskSpace < getMinRemainingDiskCapa()) {
								long t0 = System.currentTimeMillis();
								_debug.info(
										"Spontanes Löschen, Level " + level + " mit " + (i + 1) + " Monaten Verkürzung des Vorhaltezeitraums: "
												+ "Verbleibender Plattenplatz in Bytes: " + freeDiskSpace
								);
								containerWalk(ads, new DeleteSpontan(compareDate, level));
								freeDiskSpace = getFreeDiskSpace();
								_debug.info(
										"Nach spontanem Löschen, Level " + level + " mit " + (i + 1) + " Monaten Verkürzung des Vorhaltezeitraums: "
												+ "Verbleibender Plattenplatz in Bytes: " + freeDiskSpace + ", " + "Dauer: "
												+ ((System.currentTimeMillis() - t0) / 1000) + " Sekunden"
								);
							}
							else {
								MessageSender.getInstance().sendMessage(
										de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
										MessageGrade.INFORMATION,
										"Das spontane Löschen hat Daten gelöscht"
								);
								_debug.info("Das spontane Löschen hat Daten gelöscht");
								return;
							}
						}
						break;

					// Löschung nach den 4 Kriterien in der TAnf 5.1.2.7
					default:
						long t0 = System.currentTimeMillis();
						_debug.info("Spontanes Löschen, Level " + level + ": Verbleibender Plattenplatz in Bytes: " + freeDiskSpace);
						containerWalk(ads, new DeleteSpontan(compareDate, level));
						freeDiskSpace = getFreeDiskSpace();
						_debug.info(
								"Nach spontanem Löschen, Level " + level + ": Verbleibender Plattenplatz in Bytes: " + freeDiskSpace + ", " + "Dauer: "
										+ ((System.currentTimeMillis() - t0) / 1000) + " Sekunden"
						);
				}
			}
			else {
				if(someDataDeleted) {
					MessageSender.getInstance().sendMessage(
							de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
							MessageGrade.INFORMATION,
							"Das spontane Löschen hat Daten gelöscht"
					);
					_debug.info("Das spontane Löschen hat Daten gelöscht");
				}
				return; // Es wurde genug PLatz freigegeben -> fertig
			}
		}

		// Wenn jetzt noch immer nicht genug Platz frei ist, belegt u.U. die Verwaltungsinfo
		// alleine schon so viel, dass wir durch Löschen keinen Platz mehr bekommen
		MessageSender.getInstance().sendMessage(
				de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
				MessageGrade.WARNING,
				"Achtung: Der Speicherplatz geht zu Ende! Durch spontanes Löschen konnte nicht so viel Platz freigegeben werden, wie laut Parametrierung mindestens frei sein soll."
		);
		_debug.warning(
				"Achtung: Der Speicherplatz geht zu Ende! Durch spontanes Löschen konnte nicht so viel Platz freigegeben werden, wie laut Parametrierung mindestens frei sein soll."
		);
	}

	/**
	 * Gibt den freien Speicherplatz zurück und veröffentlicht ihn ebenfalls über den Datenverteiler (falls möglich)
	 * @return frier Speicherpaltz in Bytes
	 * @throws Exception
	 */
	public long getFreeDiskSpace() throws Exception {
		long freeDiskSpace = Util.calcFreeDiskSpace(persMgr.getArchivePath());
		archMgr.getInQueuesMgr().setFreeDiskSpace(freeDiskSpace);
		return freeDiskSpace;
	}

	/**
	 * Die generelle Container-Walk-Methode.
	 * <p>
	 * Je nachdem, ob die ArchiveDataSpec gefuellt ist oder nicht, wird ein ContainerWalk über das komplette Persistenz-Verzeichnis (containerWalkAllFromDir)
	 * oder alle spezifizierten Container (containerWalkAllByDataSpec) ausgeführt.<br><br>
	 * <p>
	 * Achtung!<br>Die Methode löscht je nach übergebenen Parametern Archivdaten!
	 *
	 * @param ads
	 * @param changeTime
	 * @param compareDate
	 * @throws PersistenceException
	 * @throws IndexException
	 */
	public void containerWalk(ArchiveDataSpecification[] ads, ContainerHandler handler) throws Exception {
		numOfVisitedCont = numOfDeletedCont = 0;
		if(ads.length == 0) {
			containerWalkAllFromDir(handler);
		}
		else {
			containerWalkByArchiveDataSpec(ads, handler);
		}
	}

	private void containerWalkAllFromDir(ContainerHandler handler) throws Exception {
		File didDir;
		ContainerFileDirIterator iter = new ContainerFileDirIterator(persMgr.getArchivePath());
		while((didDir = iter.getNextDIdDirectory()) != null) {
			long[] did = PersistenceManager.extractDIdFromPath(didDir);
			ArchiveDataKind adk = PersistenceManager.getArchiveDataKindFromDir(didDir);

			File[] files = didDir.listFiles(ContainerFile.CONT_FILENAME_FILTER);
			if(files != null) {
				for(File cont : files) {
					if(shouldTerminate()) throw new Exception(getName() + ": Archivsystem wurde beendet.");

					DataIdentNode din = didTree.existsCreate(did[0], did[1], did[2], (short) did[3]); // nie null
					boolean busyAreaEntered = false;
					try {
						//Prüfen, ob Archivsystem überlastet und Task deshalb angehalten werden soll
						//Spontanes Löschen darf aber nicht angehalten werden, sondern nur Löschzeitverlängerung und (nicht-spontanes) Löschen.
						if(handler.maySuspend()) suspendTaskIfNecessary();

						// Synchronisierung ueber DataIdentNode:
						busyAreaEntered = din.enterBusyArea();
						if(!busyAreaEntered) {
							_debug.warning(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
							throw new Exception(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
						}
						// Auf Container zugreifen und bearbeiten:
						persMgr.accessContainer(this, did[0], did[1], did[2], (short) did[3], adk, PersistenceManager.getContID(cont));
						numOfVisitedCont++;
						handler.handleContainer();
					}
					catch(PersistenceException e) {
						throw e;
					}
					finally {
						persMgr.leaveContainer(this);
						if(din != null && busyAreaEntered) din.exitBusyArea();
					}
				}
			}
		}
	}

	private void containerWalkByArchiveDataSpec(ArchiveDataSpecification[] ads, ContainerHandler handler) throws Exception {

		for(int i = 0; i < ads.length; i++) {
			ArchiveDataSpecification arcDataSpec = ads[i];
			DataIdentNode din = didTree.existsCreate(arcDataSpec); // DataIdentNode ist hier nie null

			boolean criticalAreaEntered = false;
			try {
				// Synchronisierung ueber DataIdentNode:
				criticalAreaEntered = din.enterCriticalArea();
				if(!criticalAreaEntered) {
					_debug.warning(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
					throw new Exception(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
				}

				// Indexe aller Datensatzarten anlegen:
				// (Alle, damit die Container all der Datensatzarten behandelt werden koennen, fuer die Daten vorhanden sind.)
				for(ArchiveDataKind dataKind : Util.getDataKinds(arcDataSpec.getDataKinds())) {
					if(!din.indexesAssigned(dataKind)) {
						try {
							persMgr.assignIndexes(didTree, arcDataSpec, dataKind, this, indexMem);
						}
						catch(DIdNodeNotFoundException e) {
						} // Exception tritt nicht auf, da DIN oben angelegt worden ist.
					}
				}
			}
			finally {
				if(din != null && criticalAreaEntered) din.exitCriticalArea();
			}

			ArchiveTimeSpecification atSpec = arcDataSpec.getTimeSpec();
			long start = atSpec.getIntervalStart();
			long end = atSpec.getIntervalEnd();

			for(ArchiveDataKind dataKind : Util.getDataKinds(arcDataSpec.getDataKinds())) {
				if(shouldTerminate()) throw new Exception(getName() + ": Archivsystem wurde beendet.");

				IndexResult ir;
				criticalAreaEntered = false;
				try {
					// Synchronisierung ueber DataIdentNode:
					criticalAreaEntered = din.enterCriticalArea();
					if(!criticalAreaEntered) {
						_debug.warning(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
						throw new Exception(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
					}
					// Pruefen, ob jemals Daten fuer diese Datenidentifikation und Datensatzart archiviert wurden:
					if(atSpec.getTimingType().equals(TimingType.DATA_INDEX)) {
						ir = din.getDataIdxContIDs(dataKind, persMgr.getPathName(arcDataSpec, dataKind), indexMem, start, end);
					}
					else if(atSpec.getTimingType().equals(TimingType.DATA_TIME)) {
						ir = din.getDataTimeContIDs(dataKind, persMgr.getPathName(arcDataSpec, dataKind), indexMem, start, end);
					}
					else if(atSpec.getTimingType().equals(TimingType.ARCHIVE_TIME)) {
						ir = din.getArSTimeContIDs(dataKind, persMgr.getPathName(arcDataSpec, dataKind), indexMem, start, end);
					}
					else {
						return;
					}
				}
				finally {
					if(din != null && criticalAreaEntered) din.exitCriticalArea();
				}

				// Alle ueber die Indexe gefundenen Container behandeln:
				for(int j = 0; j < ir.size(); j++) {
					boolean busyAreaEntered = false;
					try {
						//Prüfen, ob Archivsystem überlastet und Task deshalb angehalten werden soll
						//Spontanes Löschen darf aber nicht angehalten werden, sondern nur Löschzeitverlängerung und (nicht-spontanes) Löschen.
						if(handler.maySuspend()) suspendTaskIfNecessary();

						// Synchronisierung ueber DataIdentNode:
						busyAreaEntered = din.enterBusyArea();
						if(!busyAreaEntered) {
							_debug.warning(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
							throw new Exception(getName() + ": Synchronisierung fehlgeschlagen auf DataIdentNode: " + din);
						}
						// Auf Container zugreifen und bearbeiten:
						persMgr.accessContainer(this, persMgr.getPathName(arcDataSpec, dataKind), ir.get(j).getContainerId());
						numOfVisitedCont++;
						handler.handleContainer();
					}
					finally {
						persMgr.leaveContainer(this);
						if(din != null && busyAreaEntered) din.exitBusyArea();
					}
				}
			}
		}
	}

	/**
	 * Prueft, ob das ContainerFile geloescht werden muss. Folgende Faktoren werden betrachtet: <ul> <li>archivierte Daten, deren Loeschschutz abgelaufen ist
	 * und die nicht (mehr) gesichert werden muessen,</li> <li>wiederhergestellte Daten,</li> <li>archivierte Simulationsdaten,</li> <li>archivierte Daten,
	 * deren Loeschschutz nicht abgelaufen ist, die aber schon gesichert wurden.</li> </ul>
	 *
	 * @param compareDate
	 * @return boolean
	 * @throws PersistenceException
	 */
	private boolean hasToBeDeletedSpontan(Task task, Calendar compareDate, DELETE_LEVEL level) throws PersistenceException {
		try {
			persMgr.readContainerHeader(task);
			// aktueller, offener Container wird nicht gelöscht
			if(persMgr.getContainerHeaderParamAsInt(task, ContainerHdr.CHP_ANZ_DS) < 0) return false;

			// schon gelöschte Container nicht nochmal löschen
			if(persMgr.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_DELETED)) return false;

			// Kein level gesetzt -> keine weiteren Checks zum Löschen
			if(level == null) return false;

			switch(level) {
				case SAVED_EXPIRED: // Fallthrough
				case TIME_SHIFTING:
					// zu sichernder, aber noch nicht gesicherter, Container wird nicht gelöscht
					if(persMgr.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_TO_SAVE) && persMgr.getContainerHeaderParamAsString(
							task, ContainerHdr.CHP_MEDIUM_ID
					).equals(ContainerHdr.MEDIUM_UNSAVED)) {
						return false;
					}
					// TAnf 5.1.2.7 erster Absatz: (gesichert || nicht_zu_sichern) && Vorhaltezeitraum_abgelaufen && (kein_Löschschutz || Löschschutz_abgelaufen)
					long chpLoeschutz = persMgr.getContainerHeaderParamAsLong(task, ContainerHdr.CHP_LOESCHUTZ);

					long delAt = persMgr.getContainerHeaderParamAsLong(task, ContainerHdr.CHP_LOESCHEN);
					long compareDateTimeInMillis = compareDate.getTimeInMillis();
					return delAt < compareDateTimeInMillis && (chpLoeschutz == 0L || chpLoeschutz < compareDateTimeInMillis);
				case RESTORED:
					return persMgr.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_RESTORED);
				case SIMULATION:
					return persMgr.getContainerHeaderParamAsInt(task, ContainerHdr.CHP_SIM_VAR) != 0;
				case SAVED_NOTEXPIRED:
					return !persMgr.getContainerHeaderParamAsString(task, ContainerHdr.CHP_MEDIUM_ID).equals(ContainerHdr.MEDIUM_UNSAVED);
			}
		}
		catch(Exception e) {
			_debug.warning("Problem bei der Überprüfung, ob ein Container spontan gelöscht werden kann", e);
		}

		return false;
	}

	/**
	 * Prueft, ob das ContainerFile geloescht werden muss.
	 *
	 * @return boolean
	 * @throws PersistenceException
	 */
	private boolean hasToBeDeletedRegular(Task task) throws PersistenceException {
		try {
			persMgr.readContainerHeader(task);
			// aktueller, offener Container wird nicht gelöscht
			if(!persMgr.isContainerClosed(task)) return false;

			// zu sichernder, aber noch nicht gesicherter, Container wird nicht gelöscht
			if(persMgr.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_TO_SAVE) && persMgr.getContainerHeaderParamAsString(
					task, ContainerHdr.CHP_MEDIUM_ID
			).equals(ContainerHdr.MEDIUM_UNSAVED)) {
				return false;
			}

			// schon gelöschte Container nicht nochmal löschen
			if(persMgr.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_DELETED)) return false;

			if(IGNORE_LOESCH_SCHUTZ) return true;

			// TAnf 5.1.2.7 erster Absatz: (gesichert || nicht_zu_sichern) && Vorhaltezeitraum_abgelaufen && (kein_Löschschutz || Löschschutz_abgelaufen)
			long chpLoeschutz = persMgr.getContainerHeaderParamAsLong(task, ContainerHdr.CHP_LOESCHUTZ);
			long now = System.currentTimeMillis();

			if(!persMgr.getContainerHeaderParamAsString(task, ContainerHdr.CHP_MEDIUM_ID).equals(ContainerHdr.MEDIUM_UNSAVED)
					|| !persMgr.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_TO_SAVE)) {

				long delAt = persMgr.getContainerHeaderParamAsLong(task, ContainerHdr.CHP_LOESCHEN);
				if(delAt < now && (chpLoeschutz == 0L || chpLoeschutz < now)) return true;
			}
		}
		catch(Exception e) {
			_debug.warning("Problem bei der Überprüfung, ob ein Container regulär gelöscht werden kann", e);
		}

		// Sonst nichts löschen
		return false;
	}

	/**
	 * Prueft, ob das ContainerFile gelöscht beim manuellen löschen werden kann.
	 *
	 * @return boolean
	 * @throws PersistenceException
	 */
	private boolean hasToBeDeletedExplicit(Task task) throws PersistenceException {
		try {
			persMgr.readContainerHeader(task);
			// aktueller, offener Container wird nicht gelöscht
			if(!persMgr.isContainerClosed(task)) return false;

			// zu sichernder, aber noch nicht gesicherter, Container wird nicht gelöscht
			boolean toSafe = persMgr.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_TO_SAVE);
			boolean isUnsaved = persMgr.getContainerHeaderParamAsString(
					task, ContainerHdr.CHP_MEDIUM_ID
			).equals(ContainerHdr.MEDIUM_UNSAVED);
			if(toSafe && isUnsaved) {
				return false;
			}

			// schon gelöschte Container nicht nochmal löschen
			if(persMgr.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_DELETED)) return false;

			long now = System.currentTimeMillis();

			long delAt = persMgr.getContainerHeaderParamAsLong(task, ContainerHdr.CHP_LOESCHEN);
			if(delAt < now) return true;

		}
		catch(Exception e) {
			_debug.warning("Problem bei der Überprüfung, ob ein Container manuell gelöscht werden kann", e);
		}

		// Sonst nichts löschen
		return false;
	}

	/**
	 * Für spontantes Löschen: 1. Vorhaltezeitraum abgelaufen, und schon gesichert 2. wiederhergestellte Daten 3. Simulationsdaten 4. Vorhaltezeitraum noch
	 * nicht abgelaufen, aber schon gesichert 5. Löschen durch schrittweises Verschieben des Vorhaltezeitraums
	 * <p>
	 * Die Reihenfolge der Deklaration ist von Bedeutung!
	 */
	private enum DELETE_LEVEL {

		SAVED_EXPIRED("1 (Vorhaltezeitraum abgelaufen, und schon gesichert)"),
		RESTORED("2 (wiederhergestellte Daten)"),
		SIMULATION("3 (Simulationsdaten)"),
		SAVED_NOTEXPIRED("4 (Vorhaltezeitraum noch nicht abgelaufen, aber schon gesichert)"),
		TIME_SHIFTING("5 (Löschen durch schrittweises Verschieben des Vorhaltezeitraums)");

		private final String _name;

		DELETE_LEVEL(final String name) {
			_name = name;
		}


		public String toString() {
			return _name;
		}
	}

	private interface ContainerHandler {
		void handleContainer() throws PersistenceException;

		boolean maySuspend();
	}

	public static class CapacityCheckResultData extends ResultData {

		public CapacityCheckResultData(SystemObject object, DataDescription dataDescription, long time, Data data) {
			super(object, dataDescription, time, data);
		}
	}

	public static class DeleteRegularResultData extends ResultData {

		public DeleteRegularResultData(SystemObject object, DataDescription dataDescription, long time, Data data) {
			super(object, dataDescription, time, data);
		}
	}

	/**
	 * Sofortiges Löschen auf Anfrage. Es werden alle Container gelöscht, die nicht zu sichern sind oder schon gesichert wurden.
	 */
	public class DeleteNow implements ContainerHandler {
		@Override
		public void handleContainer() throws PersistenceException {
			if(persMgr.existsContainer(LZVTask.this) && hasToBeDeletedExplicit(LZVTask.this)) {
				persMgr.deleteContainerData(LZVTask.this);
				persMgr.setContainerHeaderParam(LZVTask.this, ContainerHdr.CHP_DELETED, "1");
				persMgr.writeContainerHeader(LZVTask.this);
				numOfDeletedCont++;
				_debug.fine("DeleteNow: Container with id <" + persMgr.getContainerHeaderParamAsString(LZVTask.this, ContainerHdr.CHP_CONT_ID) + "> deleted.");
			}
			_longTermTaskStatePublisherTask.setTaskProgress("Sofortiges Löschen: %1$d Container untersucht, %3$d Container gelöscht", numOfVisitedCont, -1, numOfDeletedCont);
		}

		@Override
		public boolean maySuspend() {
			return true;
		}
	}

	/**
	 * Späteres Löschen, wie DeleteNow, nur wird nur der Löschschutz aufgehoben statt den Container zu löschen
	 */
	public class DeleteLater implements ContainerHandler {
		@Override
		public void handleContainer() throws PersistenceException {
			if(persMgr.existsContainer(LZVTask.this) && hasToBeDeletedExplicit(LZVTask.this)) {
				persMgr.setContainerHeaderParam(LZVTask.this, ContainerHdr.CHP_LOESCHUTZ, 0);
				persMgr.writeContainerHeader(LZVTask.this);
				numOfDeletedCont++;
				_debug.fine("DeleteLater: Container with id <" + persMgr.getContainerHeaderParamAsString(LZVTask.this, ContainerHdr.CHP_CONT_ID) + "> deleted.");
			}
			_longTermTaskStatePublisherTask.setTaskProgress("Späteres Löschen: %1$d Container untersucht, %3$d Container gelöscht", numOfVisitedCont, -1, numOfDeletedCont);
		}

		@Override
		public boolean maySuspend() {
			return true;
		}
	}

	/**
	 * Reguläres Löschen. Es werden alle Container gelöscht, die nicht zu sichern sind oder die schon gesichert wurden, aber nur sofern kein Löschschutz
	 * besteht.
	 */
	public class DeleteRegular implements ContainerHandler {
		@Override
		public void handleContainer() throws PersistenceException {
			if(persMgr.existsContainer(LZVTask.this) && hasToBeDeletedRegular(LZVTask.this)) {
				persMgr.deleteContainerData(LZVTask.this);
				persMgr.setContainerHeaderParam(LZVTask.this, ContainerHdr.CHP_DELETED, "1");
				persMgr.writeContainerHeader(LZVTask.this);
				numOfDeletedCont++;
				_debug.fine("DeleteRegular: Container with id <" + persMgr.getContainerHeaderParamAsString(LZVTask.this, ContainerHdr.CHP_CONT_ID) + "> deleted.");
			}
			_longTermTaskStatePublisherTask.setTaskProgress("Reguläres Löschen: %1$d Container untersucht, %3$d Container gelöscht", numOfVisitedCont, -1, numOfDeletedCont);
		}

		@Override
		public boolean maySuspend() {
			return true;
		}
	}

	/**
	 * Spontanes Löschen bei wenig freiem Speicherplatz
	 */
	public class DeleteSpontan implements ContainerHandler {
		private final Calendar _compareDate;
		private final DELETE_LEVEL _deleteLevel;

		/**
		 * Erstellt ein neues DeleteSpontan-Objekt
		 *
		 * @param compareDate            Vergleichsdatum
		 * @param deleteLevelLösch-Level
		 */
		public DeleteSpontan(final Calendar compareDate, final DELETE_LEVEL deleteLevel) {
			this._compareDate = compareDate;
			this._deleteLevel = deleteLevel;
		}

		@Override
		public void handleContainer() throws PersistenceException {
			if(persMgr.existsContainer(LZVTask.this) && hasToBeDeletedSpontan(LZVTask.this, _compareDate, _deleteLevel)) {
				persMgr.deleteContainerData(LZVTask.this);
				persMgr.setContainerHeaderParam(LZVTask.this, ContainerHdr.CHP_DELETED, "1");
				persMgr.writeContainerHeader(LZVTask.this);
				numOfDeletedCont++;
				_debug.fine("DeleteSpontan: Container mit ID <" + persMgr.getContainerHeaderParamAsString(LZVTask.this, ContainerHdr.CHP_CONT_ID) + "> gelöscht.");
			}
			_longTermTaskStatePublisherTask.setTaskProgress("Spontanes Löschen: %1$d Container untersucht, %3$d Container gelöscht", numOfVisitedCont, -1, numOfDeletedCont);
		}

		@Override
		public boolean maySuspend() {
			return false;
		}
	}

	public class ChangeTime implements ContainerHandler {

		private final long _changeTime;

		public ChangeTime(final long changeTime) {
			_changeTime = changeTime;
		}

		@Override
		public void handleContainer() throws PersistenceException {
			persMgr.readContainerHeader(LZVTask.this);
			long loeschschutz = persMgr.getContainerHeaderParamAsLong(LZVTask.this, ContainerHdr.CHP_LOESCHUTZ);
			persMgr.setContainerHeaderParam(LZVTask.this, ContainerHdr.CHP_LOESCHUTZ, loeschschutz + _changeTime);
			persMgr.writeContainerHeader(LZVTask.this);
			numOfDeletedCont++;
			_longTermTaskStatePublisherTask.setTaskProgress("Löschschutzverlängerung: %1$d Container untersucht, %3$d Container angepasst", numOfVisitedCont, -1, numOfDeletedCont);
		}

		@Override
		public boolean maySuspend() {
			return false;
		}
	}
}
