/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.backup.BackupException;
import de.bsvrz.ars.ars.backup.MediumNotAccessibleException;
import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.LongTermTaskStatePublisher;
import de.bsvrz.ars.ars.mgmt.datatree.DIdNodeNotFoundException;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.persistence.*;
import de.bsvrz.ars.ars.persistence.index.result.IndexResult;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;

import java.io.*;
import java.util.*;
import java.util.concurrent.BlockingQueue;

/**
 * Task zur Wiederherstellung gesicherter Container.
 *
 * @author beck et al. projects GmbH
 * @author Christian Wied
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class RestoreTask extends SingleClientSenderTask {

	private static final String UNAVAIL_MEDIUM_ID_FILE = "_unavailable_mediumids.dat";

	// parametrierte Löschschutzverlängerung in Sekunden
	private static long LOESSCHUTZ_VERL = 0;

	private Set<String> mediumIdIgnoreSet = new HashSet<String>();

	private byte[] indexMem;


	/**
	 * Objekt, dem Informationen zum aktuellen Zustand der Task und dem Fortschritt von langwierigen Arbeiten übergeben werden.
	 */

	private LongTermTaskStatePublisher.Task _longTermTaskStatePublisherTask;

	/**
	 * Erzeugt einen neuen Task zur Wiederherstellung gesicherter Container.
	 *  @param rodl       Der Listener wird benachrichtigt beim Beenden der Task.
	 * @param archiveMgr Der Archiv-Manager.
	 * @param inputData  Eingangswarteschlange
	 *
 	 */
	public RestoreTask(ArchiveManager archiveMgr,BlockingQueue<Object> inputData, final LongTermTaskStatePublisher longTermTaskStatePublisher) {
		super(archiveMgr, inputData);
		setName("Restore Task");
		_longTermTaskStatePublisherTask = longTermTaskStatePublisher.registerTask("Wiederherstellung");
	}


	@Override
	public void execute() {
		_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		try {
			super.execute();
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.TERMINATED);
		}
	}

	/**
	 * Prueft, ob der angegebene Dateiname der Name einer _unavailable_mediumids.dat-Datei ist.
	 *
	 * @param name Name einer Datei
	 *
	 * @return Wahr, wenn es der Name einer _unavailable_mediumids.dat-Datei ist
	 */
	public static boolean isUnavailMediumIDsFile(String name) {
		return name.equals(UNAVAIL_MEDIUM_ID_FILE);
	}

	/**
	 * Fuehrt die Wiederherstellung gesicherter Datenvontainer durch.
	 *
	 * @param resultData Zu verarbeitender Datensatz aus der Eingangswarteschlange.
	 */
	protected void work(ResultData resultData) {
		if(!resultData.hasData()) return;
		try {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.START_WORK);
			indexMem = new byte[INDEX_MEMORY_SIZE];
			init(resultData.getData());
			initIgnoreList();
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.WORKING);
			restore(resultData);
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.FINISH_WORK);
			sendSuccessResponse();
		}
		catch(MediumNotAccessibleException e) {
			_debug.warning("Wiederherstellung abgebrochen: Medium nicht verfügbar.", e);
			MessageSender.getInstance().sendMessage(
					"", MessageType.APPLICATION_DOMAIN, "", MessageGrade.WARNING, MessageState.MESSAGE, "Bitte legen Sie Medium ein"
			);
		}
		catch(Exception e) {
			_debug.error("Wiederherstellung fehlgeschlagen.", e);
			sendErrorResponse("Wiederherstellung fehlgeschlagen: " + e.getMessage());
		}
		finally {
			mediumIdIgnoreSet.clear();
			indexMem = null;
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		}
	}

	@SuppressWarnings({"unchecked"})
	private void restore(ResultData resultData) throws Exception {

		// Wir bauen eine Liste mit Medien-IDs und Container-Filenamen auf.
		// Diese Liste sortieren wir danach nach Medien-ID und stellen jeden einzelnen Container wieder her.
		List<MediumidContainerTupel> medienIds;

		long numRestored = 0;

		ArchiveDataSpecification[] ads = parseArchiveDataSpec(deserializer, resultData);
		if(ads.length == 0) {
			medienIds = collectAllMediumIdsFromDir();    // Methode synchronisiert selbst auf DataIdentNode
		}
		else {
			medienIds = collectMediumIdsByDataSpec(ads); // Methode synchronisiert selbst auf DataIdentNode
		}

		Collections.sort(medienIds);
		if(shouldTerminate()) throw new Exception("Archivsystem wurde beendet. Wiederherstellungslauf abgebrochen");

		// Container wiederherstellen:
		for(MediumidContainerTupel tupel : medienIds) {
			suspendTaskIfNecessary();	//Task anhalten, falls das Archivsystem überlastet.

			File contFile = tupel.getContainerFile();
			int mediumID = tupel.getMediumID();
			boolean busyAreaEntered = false;
			DataIdentNode din = null;
			try {
				// Synchronisierung ueber DataIdentNode:
				long[] did = PersistenceManager.extractDIdFromPath(contFile.getParentFile());
				din = archMgr.getDataIdentTree().existsCreate(did[0], did[1], did[2], (int)did[3]);
				busyAreaEntered = din.enterBusyArea();
				if(!busyAreaEntered) {
					_debug.warning("Wiederherstellung fehlgeschlagen fuer Datei: " + contFile.getAbsolutePath());
					continue;
				}

				persMgr.accessContainerReadOnly(this, contFile);
				persMgr.readContainerHeader(this);
				long objId = persMgr.getContainerHeaderParamAsLong(this, ContainerHdr.CHP_OBJ_ID);
				long atgId = persMgr.getContainerHeaderParamAsLong(this, ContainerHdr.CHP_ATG_ID);
				long aspId = persMgr.getContainerHeaderParamAsLong(this, ContainerHdr.CHP_ASP_ID);
				int sv = persMgr.getContainerHeaderParamAsInt(this, ContainerHdr.CHP_SIM_VAR);
				ArchiveDataKind datakind = persMgr.getContainerHeaderParamAsArchiveDataKind(this, ContainerHdr.CHP_DATA_KIND);
				persMgr.leaveContainer(this);
				restoreFromBackup(mediumID, contFile.getAbsolutePath(), objId, atgId, aspId, sv, datakind);

				_longTermTaskStatePublisherTask.setTaskProgress("Wiederherstellung: Von %1$d von %2$d Medien wiederhergestellt, letzte MedienID: %3$d", ++numRestored, medienIds.size(), mediumID);

				if(shouldTerminate()) throw new Exception("Archivsystem wurde beendet. Wiederherstellungslauf abgebrochen");
			}
			finally {
				persMgr.leaveContainer(this);
				if(din != null && busyAreaEntered) din.exitBusyArea();
			}
		}
		_debug.info("Wiederherstellung erfolgreich beendet: " + medienIds.size() + " Container wiederhergestellt.");
	}

	private List<MediumidContainerTupel> collectAllMediumIdsFromDir() throws Exception {
		List<MediumidContainerTupel> medienIds = new ArrayList<MediumidContainerTupel>();
		long directoriesVisitedCount = 0;
		// Ueber alle Verzeichnisse iterieren:
		File didDir;
		ContainerFileDirIterator iter = new ContainerFileDirIterator(persMgr.getArchivePath());
		while((didDir = iter.getNextDIdDirectory()) != null) {
			long[] did = PersistenceManager.extractDIdFromPath(didDir);
			ArchiveDataKind adk = PersistenceManager.getArchiveDataKindFromDir(didDir);
			boolean busyAreaEntered = false;
			DataIdentNode din = null;
			try {
				// Synchronisierung ueber DataIdentNode:
				din = archMgr.getDataIdentTree().existsCreate(did[0], did[1], did[2], (int)did[3]);
				busyAreaEntered = din.enterBusyArea();
				if(!busyAreaEntered) {
					_debug.warning("Wiederherstellung fehlgeschlagen fuer Verzeichnis: " + didDir.getAbsolutePath());
					continue;
				}

				// Ueber alle Dateien im aktuellen Verzeichnis iterieren:
				File[] files = didDir.listFiles(ContainerFile.CONT_FILENAME_FILTER);
				if(files != null) {
					for(File cont : files) {
						if(shouldTerminate()) throw new Exception("Archivsystem wurde beendet. Wiederherstellunglauf abgebrochen");
						try {
							persMgr.accessContainer(this, did[0], did[1], did[2], (short)did[3], adk, PersistenceManager.getContID(cont));
							persMgr.readContainerHeader(this);
							String mediumID = persMgr.getContainerHeaderParamAsString(this, ContainerHdr.CHP_MEDIUM_ID);
							if(isMediumIdToBeRestored(mediumID)) {
								medienIds.add(new MediumidContainerTupel(Integer.parseInt(mediumID), cont));
							}
						}
						catch(PersistenceException e) {
							throw e;
						}
						finally {
							persMgr.leaveContainer(this);
						}
					}
				}
			}
			finally {
				if(din != null && busyAreaEntered) din.exitBusyArea();
			}
			_longTermTaskStatePublisherTask.setTaskProgress("Indizierung: %1$d Verzeichnisse untersucht, %3$d wiederherzustellende Container gefunden", ++directoriesVisitedCount, -1, medienIds.size());
		}
		return medienIds;
	}

	private List<MediumidContainerTupel> collectMediumIdsByDataSpec(ArchiveDataSpecification[] ads) throws Exception {
		List<MediumidContainerTupel> medienIds = new ArrayList<MediumidContainerTupel>();

		for(int i = 0; i < ads.length; i++) {
			ArchiveDataSpecification arcDataSpec = ads[i];
			// DataIdentNode:
			DataIdentNode din = didTree.existsCreate(arcDataSpec); // nie null

			boolean criticalAreaEntered = false;
			try {
				// Synchronisierung ueber DataIdentNode:
				criticalAreaEntered = din.enterCriticalArea();
				if(!criticalAreaEntered) {
					_debug.warning("Wiederherstellung fehlgeschlagen fuer DataIdentNode: " + din);
					continue;
				}

				// Indexe aller Datensatzarten anlegen:
				// (Alle, damit geprueft werden kann, ob jemals Daten für diese Datenidentifikation gespeichert wurden.)
				for(ArchiveDataKind dataKind : Util.getDataKinds(arcDataSpec.getDataKinds())) {
					if(!din.indexesAssigned(dataKind)) {
						try {
							persMgr.assignIndexes(didTree, arcDataSpec, dataKind, this, new byte[INDEX_MEMORY_SIZE]);
						}
						catch(DIdNodeNotFoundException e) {
						} // Exception tritt nicht auf, da DIN oben angelegt worden ist.
					}
				}
			}
			finally {
				if(din != null && criticalAreaEntered) din.exitCriticalArea();
			}

			ArchiveTimeSpecification atSpec = arcDataSpec.getTimeSpec();
			long start = atSpec.getIntervalStart();
			long end = atSpec.getIntervalEnd();
			TimingType timingType = atSpec.getTimingType();

			// Pruefen, ob jemals Daten für diese Datenidentifikation archiviert wurden:
			for(ArchiveDataKind dataKind : Util.getDataKinds(arcDataSpec.getDataKinds())) {

				criticalAreaEntered = false;
				IndexResult<?> ir = null;
				try {
					// Synchronisierung ueber DataIdentNode:
					criticalAreaEntered = din.enterCriticalArea();
					if(!criticalAreaEntered) {
						_debug.warning("Wiederherstellung fehlgeschlagen fuer DataIdentNode: " + din);
						continue;
					}
					// Synchronisiserter Index-Zugriff:
					if(din.getOpenContID(dataKind) > -1) {
						if(timingType == TimingType.ARCHIVE_TIME) {
							ir = din.getArSTimeContIDs(dataKind, persMgr.getPathName(arcDataSpec, dataKind), indexMem, start, end);
						}
						else if(timingType == TimingType.DATA_TIME) {
							ir = din.getDataTimeContIDs(dataKind, persMgr.getPathName(arcDataSpec, dataKind), indexMem, start, end);
						}
						else if(timingType == TimingType.DATA_INDEX) {
							ir = din.getDataIdxContIDs(dataKind, persMgr.getPathName(arcDataSpec, dataKind), indexMem, start, end);
						}
					}
				}
				finally {
					if(criticalAreaEntered) din.exitCriticalArea();
				}

				if(ir != null) {
					boolean busyAreaEntered = false;
					try {
						// Synchronisierung ueber DataIdentNode:
						busyAreaEntered = din.enterBusyArea();
						if(!busyAreaEntered) {
							_debug.warning("Wiederherstellung fehlgeschlagen fuer DataIdentNode: " + din);
							continue;
						}
						// Alle Container inklusive der Medien-ID sammeln:
						for(int j = 0; j < ir.size(); j++) {
							String contFilePath =
									persMgr.getPathName(arcDataSpec, dataKind) + File.separator + PersistenceManager.getContFileName(ir.get(j).getContainerId());
							File contFile = new File(contFilePath);
							persMgr.accessContainerReadOnly(this, contFile);
							persMgr.readContainerHeader(this);
							String mediumID = persMgr.getContainerHeaderParamAsString(this, ContainerHdr.CHP_MEDIUM_ID);
							persMgr.leaveContainer(this);
							if(isMediumIdToBeRestored(mediumID)) {
								medienIds.add(new MediumidContainerTupel(Integer.parseInt(mediumID), contFile));
							}
							if(shouldTerminate()) throw new Exception("Archivsystem wurde beendet. Wiederherstellunglauf abgebrochen");
						}
					}
					finally {
						persMgr.leaveContainer(this);
						if(busyAreaEntered) din.exitBusyArea();
					}
				}
			}
			if(shouldTerminate()) throw new Exception("Archivsystem wurde beendet. Wiederherstellunglauf abgebrochen");
			_longTermTaskStatePublisherTask.setTaskProgress("Indizierung: %1$d von %2$d Datenspezifikationen untersucht, %3$d wiederherzustellende Container gefunden", i + 1, ads.length, medienIds.size());
		}
		return medienIds;
	}

	/** Holt einen Container aus dem Backup zurueck. */
	private void restoreFromBackup(int backupMediumID, String containerFileName, long objId, long atgId, long aspId, int sv, ArchiveDataKind archiveDataKind)
			throws Exception {

		// Zeit messen:
		long startTime = Util.startTimer();

		containerFileName = containerFileName.substring(containerFileName.lastIndexOf(File.separator) + 1);
		// InputStream oeffnen:
		InputStream is = null;
		try {
			is = BackupTask.getBackupModul().restoreContainer(backupMediumID, containerFileName);
			if(is == null) {
				_debug.warning("Container " + containerFileName + " nicht gefunden.");
			}
		}
		catch(BackupException e) {
			MessageSender.getInstance().sendMessage(
					"", MessageType.APPLICATION_DOMAIN, "", MessageGrade.ERROR, MessageState.MESSAGE, "Wiederherstellung fehlgeschlagen"
			);
			_debug.error("Wiederherstellung fehlgeschlagen.", e);
			throw e;
		}

		File file = new File(persMgr.getPathName(objId, atgId, aspId, sv, archiveDataKind), containerFileName);
		if(file.exists()) {
			file.delete();
		}
		file.createNewFile();
		FileOutputStream fos = new FileOutputStream(file);
		Util.copyStreams(is, fos);
		is.close();
		fos.close();

		persMgr.accessContainer(this, objId, atgId, aspId, sv, archiveDataKind, ContainerFile.getContID(file));
		persMgr.readContainerHeader(this);
		persMgr.setContainerHeaderParam(this, ContainerHdr.CHP_LOESCHUTZ, System.currentTimeMillis() + LOESSCHUTZ_VERL * 1000);
		persMgr.setContainerHeaderParam(this, ContainerHdr.CHP_RESTORED, "1");
		persMgr.writeContainerHeader(this);
		persMgr.leaveContainer(this);

		// Zeit messen & Log-Ausgabe:
		_debug.info("Container " + containerFileName + " erfolgreich wiederhergestellt (" + Util.stopTimer(startTime) + " sec.)");

		return;
	}

	/**
	 * Liest die Datei MEDIUM_ID_IGNORE_FILE aus und schreibt die darin enthaltenen MedienIds in das Set mediumIdIgnoreSet. Fehler werden großzügig behandelt -
	 * notfalls bleibt das Set leer.
	 */
	private void initIgnoreList() {
		File ignoreFile = new File(persMgr.getArchivePath(), UNAVAIL_MEDIUM_ID_FILE);
		if(!ignoreFile.exists()) {
			_debug.info("Datei " + UNAVAIL_MEDIUM_ID_FILE + " existiert nicht.");
			return;
		}
		BufferedReader br = null;
		try {
			br = new BufferedReader(new FileReader(ignoreFile));
			while(true) {
				String s = br.readLine();
				if(s != null) {
					try {
						int mediumID = Integer.parseInt(s.trim());
						mediumIdIgnoreSet.add(mediumID + "");
					}
					catch(NumberFormatException nfe) {
						// do nothing
					}
				}
				else {
					break;	//Dateiende erreicht
				}
				suspendTaskIfNecessary();	//Task ggf. anhalten
			}
		}
		catch(IOException e) {
			_debug.warning(
					"Exception trat beim Lesen der Datei " + UNAVAIL_MEDIUM_ID_FILE + " auf. Wiederherstellung verwendet keine MedienID-Ignore-Liste.", e
			);
		}
		finally {
			if(br != null) {
				try {
					br.close();
				}
				catch(IOException ioe) {
				}
			}
		}
	}

	private boolean isMediumIdToBeRestored(String mediumID) {
		if(mediumID.equals(ContainerHdr.MEDIUM_DESTROYED) || mediumIdIgnoreSet.contains(mediumID + "") || mediumID.equals(ContainerHdr.MEDIUM_UNSAVED)) {
			return false;
		}

		return true;
	}

	public static long getLOESSCHUTZ_VERL() {
		return LOESSCHUTZ_VERL;
	}

	public static void setLOESSCHUTZ_VERL(long loesschutz_verl) {
		LOESSCHUTZ_VERL = loesschutz_verl;
	}

	private class MediumidContainerTupel implements Comparable {

		private int mediumID;

		private File containerFile;

		public MediumidContainerTupel(int mediumid, File file) {
			containerFile = file;
			mediumID = mediumid;
		}

		public File getContainerFile() {
			return containerFile;
		}

		public int getMediumID() {
			return mediumID;
		}

		public int compareTo(Object o) {
			int otherMediumID = ((MediumidContainerTupel)o).getMediumID();
			return mediumID < otherMediumID ? -1 : (mediumID == otherMediumID ? 0 : 1);
		}
	}

}
