/*
 * Copyright 2017 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.sys.funclib.kappich.annotations.NotNull;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;

import java.util.concurrent.Executors;

/**
 * Abstrakte Basisklasse für die abstrakten Task-Basisklassen {@link SingleTask} und {@link
 * MultiTask}.
 *
 * Ein Task ist ein Thread, der laufend eine bestimmte Aktion ausführt. Er ähnelt einem {@link Executors#newSingleThreadExecutor()}, die einzelnen
 * Aufgaben werden aber (unter anderem) aus historischen Gründen nicht als Runnable o.ä. übergeben, sondern der Task holt sich selbst die
 * durchzuführenden Arbeitsschritte aus einer Queue oder einem Ringpuffer. Die Klasse, auf der ein Task arbeitet wird über den 
 * generischen Typ T definiert.
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schaefer
 * @author Alexander Schmidt
 * @author Kappich Systemberatung
 * @version $Revision$ / $Date$ / ($Author$)
 */
public abstract class RingBufferTask<T> extends RepeatingTask {

	private final RingBuffer<Object> _queue;

	protected RingBufferTask(ArchiveManager archiveMgr, RingBuffer<Object> queue) {
		super(archiveMgr);
		_queue = queue;
	}
	
	public boolean submit(@NotNull T step) throws InterruptedException {
		return _queue.push(step);
	}

	/**
	 * Terminiert den Task, indem die {@link #run()}-Methode unterbrochen wird, sofern sie im Wartezustand ist und nicht gerade einen Auftrag bearbeitet.
	 * Andernfalls wird der Task vor dem naechsten Zyklus beendet.
	 */
	@Override
	public void terminateTask() {
		super.terminateTask();
		if(_queue != null) _queue.push(TERMINATE_TASK);
	}

	@Override
	public void step() throws InterruptedException {
		Object take = _queue.pop();
		if(take == TERMINATE_TASK) {
			return;
		}
		
		work((T) take);	// synchronized, blockiert falls leer
	}

	protected abstract void work(final T step) throws InterruptedException;

	public RingBuffer<Object> getQueue() {
		return _queue;
	}
}
