/*
 * Copyright 2017 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTree;
import de.bsvrz.ars.ars.mgmt.tasks.tasklistener.TaskStepListener;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.sys.funclib.debug.Debug;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/**
 * Abstrakte Basisklasse für die abstrakten Task-Basisklassen {@link de.bsvrz.ars.ars.mgmt.tasks.SingleTask} und {@link
 * de.bsvrz.ars.ars.mgmt.tasks.MultiTask}.
 *
 * Ein Task ist ein Thread, eine bestimmte Aktion ausführt. Mit {@link #terminateTask()} kann diese Aufgabe abgebrochen werden.
 * Während der Ausführung kann der Task mit {@link #shouldTerminate()} prüfen, ob er terminiert wurde und dann die Aufgabe ggf. neu starten.
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schaefer
 * @author Alexander Schmidt
 * @author Kappich Systemberatung
 * @version $Revision$ / $Date$ / ($Author$)
 */
public abstract class Task implements SuspendableTask {
	
	/** Größe des Byte-Arrays für temporaeren Speicher beim Abfragen der Indexe. */
	protected static final int INDEX_MEMORY_SIZE = 16 * 1024;
	
	protected final Debug _debug = Debug.getLogger();
	
	/** Alle Objekte in dieser Liste werden benachrichtigt, wenn ein Zyklus der Hauptschleife durchlaufen wurde. */
	private final List<TaskStepListener> taskStepListeners = new ArrayList<TaskStepListener>();

	
	/** Zentrale Archiv-Verwaltung. */
	protected ArchiveManager archMgr;

	/**
	 * Der Name des Tasks bzw. Threads. kann frei gesetzt werden. Standardmäßig der Klassenname.
	 */
	private String _name;

	/**
	 * Der aktuelle Thread. Nachdem eine Exception auftritt wird ggf. ein neuer Thread gestartet.
	 */
	private volatile Thread _thread = null;

	/**
	 * Wurde der Task gestartet?
	 */
	private volatile boolean _isStarted = false;
	
	/** Datenindentifikationen */
	protected DataIdentTree didTree;
	
	/** PersistenceManager für den Zugriff auf die Daten des Archivsystems */
	protected PersistenceManager persMgr;

	/**
	 * Zeigt an ob der Task beendet werden soll. Der Task bricht dann möglichst schnell an der nächsten sicheren Stelle ab.
	 */
	protected volatile boolean _shouldTerminate = false;
	
	/**
	 * Wurde der Task schon komplett beendet?
	 */
	private volatile boolean _isTerminated = false;

	public Task(ArchiveManager archiveMgr) {
		archMgr = archiveMgr;
		_name = getClass().getSimpleName();
		if(archMgr != null) {
			didTree = archMgr.getDataIdentTree();
			persMgr = archMgr.getPersMgr();
			if(archiveMgr.getDavCon() != null) {
				getObjectsFromDav();
			}
		}
	}

	public final Thread.State getState() {
		Thread thread = _thread;
		if(thread != null) return thread.getState();
		return Thread.State.NEW;
	}

	public void start() {
		_thread = new Thread(this::execute, getName());
		_thread.start();
		_isStarted = true;
	}

	/**
	 * Diese Methode muss implementiert werden und definiert die Aktion, die dieser Task ausführt
	 */
	protected void execute(){
		_isTerminated = true;
	}

	/**
	 * Wird garantiert aufgerufen, nachdem der Task vollständig terminiert wurde, also die Step-Funktion das letzte mal ausgeführt wurde.
	 * <p>
	 * Tut standardmäßig nichts, kann überschrieben werden um Daten aufzuräumen, Abmeldungen durchzuführen usw.
	 */
	protected void cleanUp() {

	}

	/**
	 * Zu Testzwecken kann hiermit ein Listener eingefuegt werden, der am Ende eines jedem Schleifendurchlaufs aufgerufen wird. Auf diese Weise sind z.B. die
	 * Tasks einfacher zu testen. Wenn ein Task abgebrochen wird, kann nicht garantiert werden, dass für die letzte Schleife der {@link TaskListener} aufgerufen
	 * wird.
	 *
	 * @param tsl Callback-Interface
	 */
	public void addTaskStepListener(TaskStepListener tsl) {
		synchronized(taskStepListeners) {
			taskStepListeners.add(tsl);
		}
	}

	/**
	 * Entfernt den Listener.
	 *
	 * @param tsl Callback-Interface
	 */
	public void removeTaskStepListener(TaskStepListener tsl) {
		synchronized(taskStepListeners) {
			taskStepListeners.remove(tsl);
		}
	}

	/**
	 * Benachrichtigt darüber, dass ein Schleifendurchlauf abgearbeitet wurde.
	 */
	protected void taskStepDone() {
		synchronized(taskStepListeners) {
			for(TaskStepListener tsl : taskStepListeners) {
				tsl.taskStepDone(this);
			}
		}
	}

	/**
	 * Terminiert den Task, indem die {@link #run()}-Methode unterbrochen wird, sofern sie im Wartezustand ist und nicht gerade einen Auftrag bearbeitet.
	 * Andernfalls wird der Task vor dem naechsten Zyklus beendet.
	 */
	public void terminateTask() {
		_shouldTerminate = true;
	}

	/**
	 * Gibt an, ob der Task terminiert werden soll. Diese Methode sollte innerhalb des Tasks aufgerufen werden um festzustellen, ob der Vorgang abgebrochen werden soll.
	 * Sie gibt nicht an, ob der Task schon komplett beendet ist!
	 * @return Ob der Task sich beenden soll
	 */
	@Override
	public boolean shouldTerminate() {
		return _shouldTerminate;
	}

	/** 
	 * Gibt den Archivmanager zurück
	 * @return den Archivmanager
	 */
	public ArchiveManager getArchMgr() {
		return archMgr;
	}

	/**
	 * Hält den Task an, falls sich das Archivsystem in einem kritischen Zustand befindet.
	 */
	@Override
	public final void suspendTaskIfNecessary() {
		if(!shouldTerminate()) {        //nur anhalten, wenn das Archivsystem nicht herunterfahren soll.
			try {
				archMgr.getInQueuesMgr().checkForSuspendNonWriteTask();
			}
			catch(InterruptedException e) {
				_debug.warning("Thread wurde beim Warten unterbrochen", e);
			}
		}
	}

	/**
	 * Wird aufgerufen, sobald eine Datenverteilerverbindung verfügbar ist. kann benutzt werden, um Objekte zu laden und Anmeldungen durchzuführen.
	 * <p>
	 * Wichtig: Diese Methode wird im Konstruktor ausgeführt und sollte daher nicht erwarten, dass die Subklasse (`this`) vollständig initialisiert ist.
	 */
	public void getObjectsFromDav() {

	}

	/** 
	 * Gibt den Namen zurück
	 * @return den Namen
	 */
	public final String getName() {
		return _name;
	}

	/**
	 * Setzt den Namen, der auch den Namen des Threads definiert und in Debug-Ausgaben benutzt wird.
	 * @param name neuer Name (nicht null)
	 */
	public final void setName(final String name) {
		_name = Objects.requireNonNull(name);
		Thread thread = _thread;
		if(thread != null) {
			thread.setName(name);
		}
	}

	/**
	 * Wartet auf das Beenden des Threads nach der Terminierung.
	 * @throws InterruptedException
	 */
	public void join() throws InterruptedException {
		Thread thread = _thread;
		if(thread != null) {
			thread.join();
		}
	}

	/** 
	 * Gibt <tt>true</tt> zurück, wenn der Task aktuell läuft
	 * @return <tt>true</tt>, wenn der Task aktuell läuft, sonst <tt>false</tt>
	 */
	public final boolean isAlive() {
		return _isStarted && !_isTerminated;
	}

	/** 
	 * Gibt <tt>true</tt> zurück, wenn der Task schon gelaufen ist und komplett terminiert wurde
	 * @return <tt>true</tt>, wenn der Task schon gelaufen ist und komplett terminiert wurde, sonst <tt>false</tt>
	 */
	public boolean isTerminated() {
		return _isTerminated;
	}
}
