/*
 *
 * Copyright 2017 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.util.Arrays;
import java.util.zip.Inflater;

import static de.bsvrz.sys.funclib.losb.util.Util.dIdx2StrExt;

/**
 * TBD Dokumentation
 *
 * @author Kappich Systemberatung
 */
public class ContainerDataResult {
	private Inflater _inflater = null;
	private ArchiveDataKind dataKind = null;
	private byte[] _data = null;
	private boolean _isCompressed = false;
	private boolean _isStillCompressed = false;
	private long _archiveTime = 0L;
	private long _dataIndex = 0L;
	private long _dataTime = 0L;
	private boolean isNodata = false;
	private boolean isNoSource = false;
	private boolean isPotDataGap = false;
	private boolean isNoRights = false;
	private int dataSize = 0;
	private int dataUncSize = 0;
	private boolean useExtraBuffer = false;
	private long _containerID = 0L;

	public ContainerDataResult() {
	}

	public ContainerDataResult(final Inflater inflater) {
		_inflater = inflater;
	}

	public Inflater getInflater() {
		return _inflater;
	}

	private String getHeader() {
		int adkIdx = Util.getDataKindIndex(dataKind);
		return (adkIdx != -1 ? PersistenceManager.DATAKIND_SUFFIX[adkIdx] : "") + " " + ((isData())
				? ("(" + Util.leadBlank(getDataSize(), 5) + "/"
				+ (isCompressed() ? Util.leadBlank(getDataUncSize(), 5) : " --- ")
				+ ") ")
				: ((isNoData() ? "[keine Daten] " : "")
				+ (isNoSource() ? "[keine Quelle]" : "")
				+ (isNoRights() ? "[keine Rechte]" : "") + (isPotDataGap()
				? "[pot. Luecke] "
				: ""))) + " dx="
				+ dIdx2StrExt(
				getDataIndex()
		) + " (" + getDataIndex() + ")" + " dt=" + Util.timestrMillisFormatted(getDataTime()) + " at=" + Util.timestrMillisFormatted(getArchiveTime());
	}

	/**
	 * Liefert den serialisierten Datensatz
	 *
	 * @return Datensatz
	 */
	public byte[] getData() {
		return _data;
	}

	/** @return Ob der Datensatz im Header als komprimiert gekennzeichnet wurde */
	public boolean isCompressed() {
		return _isCompressed;
	}

	/**
	 * @return Ob der Datensatz nach dem Lesen mit {@link #getRawData()} immer noch komprimiert ist. Das ArS kann so konfiguriert werden, dass es beim Lesen die
	 * Datensätze dekomprimiert und sie unkomprimiert verschickt. Ist deutlich schneller als das Dekomprimieren durch das Stream-Interface.
	 */
	public boolean isStillCompressed() {
		return _isStillCompressed;
	}

	/**
	 * Liefert den Datenindex des Datensatzes dessen Header-Daten zuvor mittels {@link #readNextDataHdr()} eingelesen worden sind.
	 *
	 * @return Datenindex
	 */
	public long getDataIndex() {
		return _dataIndex;
	}

	/**
	 * Wird von der Persistenz benutzt, um den Datenindex des eingelesenen Datensatzes in diesem Iterator zu setzen.
	 *
	 * @param dataIndex Datenindex
	 */
	void setDataIndex(long dataIndex) {
		this._dataIndex = dataIndex;
	}

	/**
	 * @param tt Datenzeit, Archivzeit, Datenindex
	 * @return Datenzeit, Archivzeit, Datenindex (je nach Parameter)
	 */
	public long getTimeIndex(TimingType tt) {
		if(tt == TimingType.ARCHIVE_TIME) {
			return _archiveTime;
		}
		else if(tt == TimingType.DATA_TIME) {
			return _dataTime;
		}
		else {
			return _dataIndex;
		}
	}

	/** @return Laufende Nummer des Datenindex (ohne Anmeldezeit Quelle und Modifier-Bits) */
	public long getDataIndexLfdNr() {
		return Util.dIdxLfdnr(getDataIndex());
	}

	/**
	 * Liefert den Datenzeitstempel des Datensatzes dessen Header-Daten zuvor mittels {@link #readNextDataHdr()} eingelesen worden sind.
	 *
	 * @return Datenzeitstempel in Millisekunden seit 1970.
	 */
	public long getDataTime() {
		return _dataTime;
	}

	/**
	 * Wird von der Persistenz benutzt, um den Datenzeitstempel des eingelesenen Datensatzes in diesem Iterator zu setzen.
	 *
	 * @param dataTime Datenzeitstempel
	 */
	void setDataTime(long dataTime) {
		this._dataTime = dataTime;
	}

	/**
	 * Liefert die Markierung für diesen Datensatz.
	 *
	 * @return
	 */
	public DataState getDataState() {
		if(isNoData()) {
			return DataState.NO_DATA;
		}
		else if(isNoSource) {
			return DataState.NO_SOURCE;
		}
		else if(isPotDataGap) {
			return DataState.POSSIBLE_GAP;
		}
		else {
			return DataState.DATA;
		}
	}

	/** @return Wahr, falls der Datensatz keine spezielle Markierung besitzt */
	public boolean isData() {
		return !isNodata && !isNoSource && !isPotDataGap;
	}

	/**
	 * Nachdem {@link #readNextDataHdr()} aufgerufen wurde, gibt diese Methode an, ob der eingelesene Datensatz eine "keine Daten"-Markierung besitzt.
	 *
	 * @return Wahr, falls der Datensatz eine "keine Daten"-Markierung besitzt
	 */
	public boolean isNoData() {
		return isNodata;
	}

	/**
	 * Wird von der Persistenz benutzt, um die Markierung "keine Daten" des eingelesenen Datensatzes in diesem Iterator zu setzen.
	 *
	 * @param isNodata
	 */
	void setNodata(boolean isNodata) {
		this.isNodata = isNodata;
	}

	/**
	 * Nachdem {@link #readNextDataHdr()} aufgerufen wurde, gibt diese Methode an, ob der eingelesene Datensatz eine "keine Rechte"-Markierung besitzt.
	 *
	 * @return Wahr, falls der Datensatz eine "keine Rechte"-Markierung besitzt
	 */
	public boolean isNoRights() {
		return isNoRights;
	}

	/**
	 * Wird von der Persistenz benutzt, um die Markierung "keine Rechte" des eingelesenen Datensatzes in diesem Iterator zu setzen.
	 *
	 * @param isNoRights
	 */
	void setNoRights(boolean isNoRights) {
		this.isNoRights = isNoRights;
	}

	/**
	 * Nachdem {@link #readNextDataHdr()} aufgerufen wurde, gibt diese Methode an, ob der eingelesene Datensatz eine "keine Quelle"-Markierung besitzt.
	 *
	 * @return Wahr, falls der Datensatz eine "keine Quelle"-Markierung besitzt
	 */
	public boolean isNoSource() {
		return isNoSource;
	}

	/**
	 * Wird von der Persistenz benutzt, um die Markierung "keine Quelle" des eingelesenen Datensatzes in diesem Iterator zu setzen.
	 *
	 * @param isNoSource
	 */
	void setNoSource(boolean isNoSource) {
		this.isNoSource = isNoSource;
	}

	/**
	 * Nachdem {@link #readNextDataHdr()} aufgerufen wurde, gibt diese Methode an, ob der eingelesene Datensatz als "potentielle Datenluecke" gekennzeichnet
	 * wurde.
	 *
	 * @return Wahr, falls der Datensatz als "potentielle Datenluecke" gekennzeichnet wurde
	 */
	public boolean isPotDataGap() {
		return isPotDataGap;
	}

	/**
	 * Wird von der Persistenz benutzt, um die Markierung "potentielle Datenluecke" des eingelesenen Datensatzes in diesem Iterator zu setzen.
	 *
	 * @param isPotDataGap
	 */
	void setPotDataGap(boolean isPotDataGap) {
		this.isPotDataGap = isPotDataGap;
	}

	public int getDataSize() {
		return dataSize;
	}

	void setDataSize(int dataSize) {
		this.dataSize = dataSize;
	}

	public int getDataUncSize() {
		return dataUncSize;
	}

	void setDataUncSize(int dataUncSize) {
		this.dataUncSize = dataUncSize;
	}

	boolean mustUseExtraBuffer() {
		return useExtraBuffer;
	}

	void setUseExtraBuffer(boolean useExtraBuffer) {
		this.useExtraBuffer = useExtraBuffer;
	}

	public ArchiveDataKind getDataKind() {
		return dataKind;
	}

	public long getArchiveTime() {
		return _archiveTime;
	}

	public boolean isNodata() {
		return isNodata;
	}

	public boolean isUseExtraBuffer() {
		return useExtraBuffer;
	}

	public void setDataKind(final ArchiveDataKind dataKind) {
		this.dataKind = dataKind;
	}

	public void setData(final byte[] data) {
		_data = data;
	}

	public void setCompressed(final boolean compressed) {
		_isCompressed = compressed;
	}

	public void setStillCompressed(final boolean stillCompressed) {
		_isStillCompressed = stillCompressed;
	}

	public void setArchiveTime(final long archiveTime) {
		_archiveTime = archiveTime;
	}

	@Override
	public String toString() {
		return "ContainerDataResult{" +
				"dataKind=" + dataKind +
				", _data=" + Arrays.toString(_data) +
				", _isCompressed=" + _isCompressed +
				", _isStillCompressed=" + _isStillCompressed +
				", _archiveTime=" + _archiveTime +
				", _dataIndex=" + dIdx2StrExt(_dataIndex) +
				", _dataTime=" + _dataTime +
				", isNodata=" + isNodata +
				", isNoSource=" + isNoSource +
				", isPotDataGap=" + isPotDataGap +
				", isNoRights=" + isNoRights +
				", dataSize=" + dataSize +
				", dataUncSize=" + dataUncSize +
				", useExtraBuffer=" + useExtraBuffer +
				'}';
	}

	public long getContainerID() {
		return _containerID;
	}

	public void setContainerID(final long containerID) {
		_containerID = containerID;
	}

	public void copyTo(final ContainerDataResult result) {
		result.setUseExtraBuffer(useExtraBuffer);
		result.setStillCompressed(_isStillCompressed);
		result.setPotDataGap(isPotDataGap);
		result.setArchiveTime(_archiveTime);
		result.setNoSource(isNoSource);
		result.setNoRights(isNoRights);
		result.setNodata(isNodata);
		result.setDataUncSize(dataUncSize);
		result.setDataTime(_dataTime);
		result.setDataSize(dataSize);
		result.setDataIndex(_dataIndex);
		result.setDataKind(dataKind);
		result.setCompressed(_isCompressed);
		result.setContainerID(_containerID);
	}
}
