/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;


import de.bsvrz.ars.ars.mgmt.tasks.RepeatingTask;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.util.BufferedRandomAccessFile;
import de.bsvrz.dav.daf.util.CloseableRandomAccessFile;
import de.bsvrz.dav.daf.util.FileAccess;
import de.bsvrz.sys.funclib.dataSerializer.Serializer;
import de.bsvrz.sys.funclib.dataSerializer.SerializingFactory;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;
import de.bsvrz.sys.funclib.losb.util.ByteIO;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.DataFormatException;
import java.util.zip.Deflater;
import java.util.zip.Inflater;

/**
 * Diese Klasse enthaelt die Funktionalitaet zum Erzeugen, Lesen, Schreiben und Loeschen von Container-Dateien. Jeder Task, der mit Container-Dateien arbeitet,
 * enthaelt ein Objekt dieser Klasse. Die Sichtbarkeit von Konstruktor und Methoden ist beschraenkt, da diese Klasse nur durch den {@link
 * de.bsvrz.ars.ars.persistence.PersistenceManager} im gleichen Package benutzt wird. Dieser haelt für jede anfragende {@link RepeatingTask}
 * ein Objekt dieser Klasse und leitet Anfragen bzgl. Container-Dateien an dieses <code>ContainerFile</code>-Objekt weiter.
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schaefer
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ContainerFile {

	private static final Debug _debug = Debug.getLogger();

	private static final boolean DEBUG_CALLS = false;

	/** 4 Byte langes Versionskennzeichen. */
	private static final byte[] VERSION_STRING = new byte[]{'v', '1', '0', '0'};

	/** Laenge des Laengenbytes vor jedem Datensatz (4 Bytes, Integer). */
	private static final int DATALEN_LEN = ByteIO.INT4B_LEN;

	/** Laenge des Headers eines jeden Datensatzes. Archivzeitstempel (6B), Datenzeitstempel (6B), Datenindex (8B), Pruefsumme (4B). */
	private static final int DATAHDR_LEN = (2 * ByteIO.LONG6B_LEN) + ByteIO.LONG8B_LEN + ByteIO.INT4B_LEN;

	/** Laenge des Compress-Info-Feldes. Compress-Info enthaelt 0 bei unkomprimierten Datensatz, sonst dekomprimierte Laenge des komprimierten Datensatzes. */
	private static final int COMPRESS_LEN = ByteIO.INT4B_LEN;

	/** Wert des Compress-Info-Feldes, wenn DS nicht komprimiert ist */
	public static final int NOT_COMPRESSED = 0;

	private static int OPEN_FILE_LIMIT = 500;

	/** Falls wahr, werden Archivanfragen komprimiert beantwortet (ist langsam, weil das Stream-Interface der Kernsoftware langsam ist) */
	public static boolean DECOMPRESS_QUERY_RESULTS = false;

	/** Datensatzbytes, falls "keine Quelle" signalisiert wurde */
	private final static byte[] NO_SOURCE = new byte[]{'N', 'O', '_', 'S', 'R', 'C'};

	/** Datensatzbytes, falls "keine Daten" signalisiert wurde */
	private final static byte[] NO_DATA = new byte[]{'N', 'O', 'D', 'A', 'T', 'A'};

	/** Datensatzbytes, falls "keine Rechte" signalisiert wurde */
	private final static byte[] NO_RIGHTS = new byte[]{'N', 'O', 'R', 'I', 'G', 'H'};

	/** Datensatzbytes um eine potentielle Datenluecke zu kennzeichnen */
	private final static byte[] POT_GAP = new byte[]{'P', 'O', 'T', 'G', 'A', 'P'};

	/**
	 * Datensätze bis zu dieser Laenge bleiben immer unkomprimiert. <b>Muss mindestens so gross sein wie {@link #NO_SOURCE}, {@link #NO_DATA}, {@link #POT_GAP}
	 * lang sind!</b> <p>Auf <code>Integer.MAX_VALUE</code> setzen, um die Komprimierung zu deaktivieren.
	 */
	public static int MAX_UNCOMPRESSED = 64;

	/** Praefix eines Containerdateinamens. */
	static final String CONT_FILENAME_PREFIX = "dc";

	/** Dateiendung eines Containerdateinamens. */
	static final String CONT_FILENAME_SUFFIX = ".dat";

	/** Regulaerer Ausdruck zur Erkennung von Containerdateinamen. */
	static final Pattern CONT_FILENAME_PAT = Pattern.compile("dc(\\d+)\\.dat");

	/** Filter zur Erkennung von Containerdateien anhand des Dateinamens. */
	public static final FilenameFilter CONT_FILENAME_FILTER = new FilenameFilter() {
		public boolean accept(File dir, String name) {
			return CONT_FILENAME_PAT.matcher(name).matches();
		}
	};

	/**
	 * Größe der Daten-Puffer in Byte. Wenn eine Anfrage 10000 Unteranfragen enthaelt, werden auch 10000 ContainerFile-Objekte angelegt. Drum nicht zu gross
	 * waehlen. Bei Bedarf werden Größere Puffer extra angelegt.
	 */
	private static final int BUFFER_SIZE = 4 * 1024;


	/** Gibt an, ob gerade ein Performance-Test mit Dummy-Daten (um den Overhead durch Deserialisieren/Serialisieren zu vermeiden) laeuft. */
	private static boolean IS_PERFORMANCE_TEST = false;

	/** Dummy-Daten für den Performance-Test. */
	private static byte[] PERF_TESTDATA = null;

	/** Gibt an, ob ausgelagerte Container beim Nachfordern so behandelt werden, dass ein Iterator normal darüber iterieren kann */
	private boolean deletedSpecialTreatment = false;

	/** Gibt an, ob im aktuellen Container das "deleted"-Flag gesetzt ist */
	private boolean deletedHeaderFlag = false;

	/** Gibt den Seperator an, der für den Datenindexlistenaufbau beim Löschen eines Containers verwendet wird */
	byte[] indexListSep = new byte[]{'-'};

	//
	// ENDE STATISCHE FELDER
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	/** Puffer zum Einlesen von Laengenbytes etc. */
	private byte[] byte8Buf = new byte[8];

	/** Puffer zum Kopieren von Dateien oder Einlesen von Strings und Zahlen. Wird nur von ArchivTasks benutzt (und damit automatisch angelegt) */
	private byte[] writeBuf = null;

	/**
	 * Referenz auf den tatsaechlichen, von {@link ContainerFile#serializeData(ResultData,long)} verwendeten Puffer, der von {@link ContainerFile#writeBuf}
	 * abweicht, falls dieser zu klein ist). Notwendig, da Daten vorserialisiert werden müssen, um deren Länge zu bestimmen.
	 */
	private byte[] actualWriteBuf;

	/** Laenge des von {@link ContainerFile#serializeData(ResultData,long)} serialisierten Datensatzes. */
	private int totalWriteDataSize;

	/**
	 * Puffer zum Ein-/Auslesen des Container-Headers und anderer Daten. Muss ein extra Puffer sein, damit man zwischen dem Vorserialisieren und dem tatsächlichen
	 * Schreiben eines Datensatzes z.B. einen neuen Container anlegen kann (dazu muss man einen Header schreiben) ohne das vorserialisierte Datum zu beschaedigen.
	 */
	private byte[] readBuf = new byte[BUFFER_SIZE];

	/** Iterator für das Durchlaufen angefragter Datensätze aus einem Container. */
	private ContainerDataIterator dataIter;

	/** Puffer für den Serialisierer und zum Lesen eines Datensatzes. */
	private ByteArrayOutputStream bos = null;

	/** Datensatz-Serialisierer. */
	private Serializer serializer = null;

	/**
	 * Container-Header Parameter als Key/Value-Paare. Die Parameter sind in {@link KeyValParam} statisch deklariert. Als Key eines Parameters wird {@link
	 * KeyValParam#getKey()} benutzt.
	 */
	private ContainerHdr containerHdr = new ContainerHdr();

	/** Aktuelle Objekt-ID. */
	private long objId;

	/** Aktuelle Attributgruppen-ID. */
	private long atgId;

	/** Aktuelle Aspekt-ID. */
	private long aspId;

	/** Aktuelle Simulationsvariante. */
	private int sv;

	/** Aktuelle Datensatzart. */
	private ArchiveDataKind dataKind;

	/** Aktuelle Container-ID. */
	private long containerId;

	/** Aktuelle Container-Datei. */
	private File contFile;

	/** Container adressiert. */
	private boolean accessed = false; // Initial kein Container im Zugriff.

	/** Container read-only adressiert. */
	private boolean readOnly;

	/** Container-Header eingelesen. */
	private boolean headerRead = false; // Initial kein Header eingelesen.


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Variablen für den Dateniterator:
	/** Container-Datei. Ist ein RandomAccessFile, da Rueckspruenge noetig sind, wenn der letzte Datensatz nur teilweise in den Puffer eingelesen wurde. */
	private MyCloseableRandomAccessFile iterRaf;

	/** Wahr, falls der Iterator nur ueber Min- und Maxwert aus dem Container-Header iteriert. */
	private boolean iterIsHdrIterator = false;

	/** Position innerhalb des Puffers im Speicher. Bei {@link #iterIsHdrIterator} der Zaehler für gelieferten Min- und Maxwert. */
	private int iterBufPos = 0;

	/** Unkomprimierte Laenge des Datensatzes (oder 0 falls dieser nicht komprimiert ist). */
	private int iterUncompressedLen = 0;

	/** Zahl der beim letzten iterLoadDataChunk() gelesenen Bytes. */
	private int iterBytesRead;

	/** Laenge des naechsten Datensatzes (wird in iterReadNextData() besetzt) */
	private int iterNextDataLen;

	/** Marker, ob nach jedem iterReadNextData() genau einmal iterGetDataElem() aufgerufen wurde (sonst stimmt die readPos nicht) */
	private boolean iterDataRead = true;

	/** Wahr, wenn das Ende eines Containers erreicht ist */
	private boolean iterEOC = false;

	/** Für gelöschte Container beim Nachfordern: <br> Liste der Datenindizes im Container */
	private ArrayList<Long> iterlistOfDataIdx = new ArrayList<Long>();

	/** Für gelöschte Container beim Nachfordern: <br> aktuelle Position in der Liste der Datenindizes */
	private int iterlistOfDataIdxPos = -1;

	private static final ThreadLocal<LinkedHashSet<MyCloseableRandomAccessFile>> _openFilesInCurrentThread = ThreadLocal.withInitial(LinkedHashSet::new);

	private static final AtomicInteger _numOpenFiles = new AtomicInteger(0);

	private final CacheManager _cacheManager = CacheManager.getInstance();

	private CacheManager.Cache _cache;

	private static int HEADER_LENGTH = VERSION_STRING.length + DATALEN_LEN + ContainerHdr.HDR_TXT_LEN + ByteIO.SEPARATOR.length;

//	private Exception lastAccessed;


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	/** Konstruktor für ContainerFile. */
	ContainerFile() {
		assert byte8Buf.length == VERSION_STRING.length + ByteIO.INT4B_LEN;
		dataIter = new ContainerDataIterator(this);
	}

	public static int getOpenFileLimit() {
		return OPEN_FILE_LIMIT;
	}

	public static void setOpenFileLimit(final int openFileLimit) {
		OPEN_FILE_LIMIT = openFileLimit;
	}

	/**
	 * Wenn diese Methode aufgerufen wird, koennen keine Daten archiviert werden. Stattdessen werden Dummy-Daten mit bekannter Laenge archiviert, um die
	 * Performance des Speichersystems testen zu koennen.
	 *
	 * @param testDat Daten, die pro Datensatz in die Container geschrieben werden.
	 */
	public static void prepareForPerformanceTest(byte[] testDat) {
		IS_PERFORMANCE_TEST = true;
		PERF_TESTDATA = testDat;
	}

	/**
	 * Vergleicht die ID's von zwei ContainerFiles<br> grösser 0, wenn die ID dieses ContainerFiles grösser ist, als die zu Vergleichende<br> kleiner 0, wenn sie
	 * kleiner ist als die zu Vergleichende<br> 0, wenn sie gleich sind
	 *
	 * @param fileID2CompareWith
	 *
	 * @return das ergebnis
	 */
	public long compareContainerFile(long fileID2CompareWith) {
		return (this.containerId - fileID2CompareWith);
	}

	/**
	 * Vergleicht die ID's von zwei ContainerFiles:<br> - grösser 0, wenn die ID des ersten ContainerFiles grösser ist, als die Zweite;<br> - kleiner 0, wenn sie
	 * kleiner ist als die Zweite;<br> - 0, wenn sie gleich sind.
	 *
	 * @param file1 das erste File
	 * @param file2 das zweite File
	 *
	 * @return
	 *
	 * @throws PersistenceException falls die Dateien keine ContainerFiles sind
	 */
	public static long compareContainerFile(File file1, File file2) throws PersistenceException {
		if(isContainerFile(file1) && isContainerFile(file2)) {
			return getContID(file1) - getContID(file2);
		}
		else {
			throw new PersistenceException(
					"Fehler beim Vergleich der ID's von " + file1.getName() + " mit " + file2.getName()
					+ " (Mindestens eines der Files ist kein ContainerFile)."
			);
		}
	}

	/*
		 * @see java.lang.Object#toString()
		 */
	public String toString() {
		if(accessed) {
			if(readOnly) {
				return "[accessed read only: file='" + contFile.getAbsolutePath() + "']";
			}
			else {
				return "[accessed: cId=" + containerId + ",objId=" + objId + ",atgId=" + atgId + ",aspId=" + aspId + ",sv=" + sv + ",adk=" + dataKind
				       + ",file='" + contFile.getAbsolutePath() + "']";
			}
		}
		else {
			return "[unaccessed]";
		}
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Container: Hilfsmethoden Dateiname
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Liefert Kennzeichen, ob die angegebene Datei eine Container-Datei ist.
	 *
	 * @param f Datei
	 *
	 * @return Kennzeichen
	 */
	public static boolean isContainerFile(File f) {
		return !f.isDirectory() && isContainerFileName(f.getName());
	}

	/**
	 * Liefert Kennzeichen, ob der angegebene Dateiname eine moegliche Container-Datei ist.
	 *
	 * @param name Dateiname
	 *
	 * @return Kennzeichen
	 */
	public static boolean isContainerFileName(String name) {
		return CONT_FILENAME_PAT.matcher(name).matches();
	}

	/**
	 * Liefert die Container-ID zur angegebenen Container-Datei, indem die im Container-Dateinamen enthaltene Container-ID ausgelesen wird. Alternativ kann die
	 * Container-ID aus den Container-Header-Parametern gelesen werden; siehe dazu {@link #getContainerHeaderParam(KeyValParam)}.
	 *
	 * @param file Container-Datei
	 *
	 * @return Container-ID
	 */
	public static long getContID(File file) {
		return getContID(file.getName());
	}

	/**
	 * Liefert die Container-ID zum angegebenen Container-Datei-Namen, indem die im Container-Dateinamen enthaltene Container-ID ausgelesen wird. Alternativ kann
	 * die Container-ID aus den Container-Header-Parametern gelesen werden; siehe dazu {@link #getContainerHeaderParam(KeyValParam)}.
	 *
	 * @param file Container-Datei
	 *
	 * @return Container-ID
	 */
	public static long getContID(String fileName) {
		Matcher m = CONT_FILENAME_PAT.matcher(fileName);
		m.find();
		return new Long(m.group(1));
	}

	/**
	 * Liefert den Dateinamen des Daten-Containers mit der angegebenen Container-ID.
	 *
	 * @param containerId Container-ID
	 *
	 * @return Dateiname
	 */
	public static String getContainerFileName(long containerId) {
		// Container-ID ist 5-Byte-Ganzzahl, d.h. max 13-stellig dezimal
		final char[] chars = {'d','c','0','0','0','0','0','0','0','0','0','0','0','0','0','.','d','a','t'};
		final String containerIdString = Long.toString(containerId);
		final int containerIdStringLength = containerIdString.length();
		containerIdString.getChars(0, containerIdStringLength, chars, 15 - containerIdStringLength);
		return new String(chars);
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Container: Access/Create/Close
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Definiert auf welchem Container die folgenden Methoden-Aufrufe wirksam sein sollen. Muss immer aufgerufen werden bevor mit einem Container gearbeitet werden
	 * kann. Der Containerdateiname wird aus der Container-ID gebildet. Soll der Zugriff auf diesen Container beendet werden (um z.B. einen anderen Container zu
	 * adressieren), muss dazu die Methode {@link #leaveContainer()} aufgerufen werden.
	 *
	 * @param objId        Objekt-ID (muss >= 0 sein)
	 * @param atgId        Attributgruppen-ID (muss >= 0 sein)
	 * @param aspId        Aspekt-ID (muss >= 0 sein)
	 * @param sv           Simulationsvariante (zwischen 0 und 999)
	 * @param dataKind     Datensatzart
	 * @param contId       Container-ID (muss zwischen 0 und 2^40-1 liegen)
	 * @param containerDir Pfad des Verzeichnisses, das die Container-Datei enthaelt
	 *
	 * @throws PersistenceException
	 */
	void accessContainer(long objId, long atgId, long aspId, int sv, ArchiveDataKind dataKind, long contId, File containerDir) throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.accessContainer " + contId);
		if(accessed) {
//			lastAccessed.printStackTrace();
			throw new PersistenceException(
					"accessContainer(" + Util.did2Str(objId, atgId, aspId, sv) + ") nicht möglich: bereits Container im Zugriff: " + toString()
			);
		}

		// Parameterobjekte uebernehmen die Pruefung der Werte:
		ContainerHdr.CHP_CONT_ID.checkVal(contId);
		ContainerHdr.CHP_OBJ_ID.checkVal(objId);
		ContainerHdr.CHP_ATG_ID.checkVal(atgId);
		ContainerHdr.CHP_ASP_ID.checkVal(aspId);
		ContainerHdr.CHP_SIM_VAR.checkVal(sv);

		this.objId = objId;
		this.atgId = atgId;
		this.aspId = aspId;
		this.sv = sv;
		this.dataKind = dataKind;
		this.containerId = contId;
		this.contFile = new File(containerDir, getContainerFileName(contId));
		this.readOnly = false; // Lesen + Schreiben
		this.headerRead = false; // Header noch nicht eingelesen.
		this.accessed = true;  // Aktuellen Container adressiert.
		this._cache = _cacheManager.getCache(this);
//		this.lastAccessed = new Exception();
	}

	/**
	 * Definiert auf welchem Container die folgenden Methoden-Aufrufe wirksam sein sollen. Es koennen nur lesende Operationen auf die Container-Datei ausgeführt
	 * werden. Vorgesehen für den Zugriff auf Container, die bereits auf ein Medium vom Typ B ausgelagert sind. Soll der Zugriff auf diesen Container beendet
	 * werden (um z.B. einen anderen Container zu adressieren), muss dazu die Methode {@link #leaveContainer()} aufgerufen werden. Die Containerdatei kann einen
	 * beliebigen Namen haben.
	 *
	 * @param contFile Containerdatei (muss existieren)
	 *
	 * @throws PersistenceException
	 */
	void accessContainerReadOnly(File contFile) throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.accessContainerReadOnly " + contFile);
		if(accessed) {
			throw new PersistenceException("accessContainerReadOnly(" + contFile.getAbsolutePath() + ") nicht möglich: bereits Container im Zugriff\n" + this);
		}
		if(!contFile.exists()) {
			throw new PersistenceException("accessContainerReadOnly(" + contFile.getAbsolutePath() + ") nicht möglich: Datei existiert nicht\n" + this);
		}
		this.objId = -1;
		this.atgId = -1;
		this.aspId = -1;
		this.sv = -1;
		this.dataKind = PersistenceManager.getArchiveDataKindFromDir(contFile.getParentFile());
		this.containerId = getContID(contFile);
		this.contFile = contFile;
		this.readOnly = true;  // Nur lesen
		this.headerRead = false; // Header noch nicht eingelesen.
		this.accessed = true;  // Aktuellen Container adressiert.
		this._cache = _cacheManager.getCache(this);
		if(_cache != null) {
			flushCache();
			// Ein lesender Task braucht den Cache im weiteren Verlauf nicht mehr
			this._cache = null;
		}
	}

	private final void flushCache() throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.flushCache");
		_cache.flush();
	}

	private final void forgetCache() {
		if(DEBUG_CALLS) System.out.println("ContainerFile.forgetCache");
		_cacheManager.forgetCache(this);
		_cache = null;
	}


	/**
	 * Beendet das Arbeiten mit dem zuvor ueber eine der <code>accessContainer()</code>-Methoden definierten Container. Diese Methode muss aufgerufen werden, wenn
	 * danach auf einen anderen Container zugegriffen werden soll. Die Methode prueft nicht, ob ein Container vorher mit {@link #accessContainer(long,
	 *long,long,int,ArchiveDataKind,long,String)} definiert wurde. D.h., man kann diese Methode mehrfach aufrufen. Dadurch wird die Ausnahmebehandlung
	 * erleichtert.
	 *
	 * @throws PersistenceException
	 */
	void leaveContainer() throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.leaveContainer");
		_cache = null;
		objId = -1;
		atgId = -1;
		aspId = -1;
		sv = -1;
		dataKind = null;
		containerId = -1;
		contFile = null;
		accessed = false;	 // Kein Container mehr adressiert.
		containerHdr.clear();	// Header-Info loeschen
		headerRead = false;	 // Kein Header mehr eingelesen.
		iterClose();			// Container-Iterator schliessen
	}

	/**
	 * Liefert Kennzeichen, ob bereits eine Container-Datei für den mit {@link #accessContainer(long,long,long,int,ArchiveDataKind,long,String)} spezifizierten
	 * Container besteht.
	 *
	 * @return Kennzeichen
	 *
	 * @throws PersistenceException
	 */
	boolean existsContainer() throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.existsContainer");
		checkContainerAccessed();
		return _cache != null || contFile.exists();
	}

	/**
	 * Erzeugt eine neue Container-Datei für den mit accessContainer() spezifizierten Container. Schreibt den Container-Header und befuellt diesen mit den
	 * angegebenen Parametern.
	 *
	 * @param toSave Sichern
	 *
	 * @throws PersistenceException
	 */
	void createContainer(boolean toSave) throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.createContainer");
		checkContainerAccessedRW();
		if(contFile.exists()) throw new PersistenceException("Erzeugen der Container-Datei fehlgeschlagen: Datei existiert bereits\n" + toString());
		if(_cache != null) {
			forgetCache();
		}
		_cache = _cacheManager.createCache(this, true);
		createDefaultContainerHeader(toSave);           // Neuen Container-Header erzeugen
		writeInitialContainerHeader();                         // Neuen Container-Header in neue Datei schreiben
	}

	/**
	 * Schliesst die Container-Datei für den mit einer der <code>accessContainer()</code>-Methoden spezifizierten Container. Der Container-Header wird
	 * entsprechend aktualisiert.
	 *
	 * @param containerEntries Anzahl Datensätze im Container
	 * @param minArcTime       Minimaler Archivzeitstempel im Container
	 * @param maxArcTime       Maximaler Archivzeitstempel im Container
	 * @param minDataTime      Minimaler Datenzeitstempel im Container
	 * @param maxDataTime      Maximaler Datenzeitstempel im Container
	 * @param minDataIdx       Minimaler Datenindex im Container
	 * @param maxDataIdx       Maximaler Datenindex im Container
	 * @param vorhalteZeit     parametrierte Vorhaltezeit in Sekunden
	 *
	 * @throws PersistenceException
	 */
	void closeContainer(
			long containerEntries, long minArcTime, long maxArcTime, long minDataTime, long maxDataTime, long minDataIdx, long maxDataIdx, long vorhalteZeit)
			throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.closeContainer");
		checkContainerAccessedRW();
		if(_cache != null) {
			flushCache();
			forgetCache();
		}
		if(!headerRead) readContainerHeader();
		// Container-Header-Parameter im Speicher setzen:
		setContHdrParam(ContainerHdr.CHP_ANZ_DS, containerEntries);
		setContHdrParam(ContainerHdr.CHP_ARC_TIME_MIN, minArcTime);
		setContHdrParam(ContainerHdr.CHP_ARC_TIME_MAX, maxArcTime);
		setContHdrParam(ContainerHdr.CHP_DATA_TIME_MIN, minDataTime);
		setContHdrParam(ContainerHdr.CHP_DATA_TIME_MAX, maxDataTime);
		setContHdrParam(ContainerHdr.CHP_DATA_IDX_MIN, minDataIdx);
		setContHdrParam(ContainerHdr.CHP_DATA_IDX_MAX, maxDataIdx);
		setContHdrParam(ContainerHdr.CHP_LOESCHEN, maxArcTime + (vorhalteZeit * 1000));	// Im Container gibts nur Millisekunden
		// Alle geaenderten Parameter zusammen in Container-Header speichern:
		writeContainerHeader();

		// Falls eine ContainerBackupIndex-Datei existiert und der Container zu sichern ist, Container eintragen
		final boolean toSave = getContHdrParamAsBool(ContainerHdr.CHP_TO_SAVE);
		if(toSave) {
			try {

				final ContainerBackupIndex containerBackupIndex = ContainerBackupIndex.getContainerBackupIndex(contFile.getParentFile());
				if(containerBackupIndex != null) {
					try {
						containerBackupIndex.appendContainerID(containerId);
					}
					finally {
						containerBackupIndex.close();
					}
				}
			}
			catch(IOException e) {
				// Wenn das Speichern im ContainerBackupIndex fehlschlägt, versuchen den ContainerBackupIndex zu löschen, damit keine Inkonsistenz auftritt
				_debug.warning("Konnte ContainerBackupIndex nicht schreiben.", e);
				if(ContainerBackupIndex.existsIn(contFile.getParentFile())) {
					if(!ContainerBackupIndex.deleteContainerBackupIndex(contFile.getParentFile())){
						throw new PersistenceException("Konnte ContainerBackupIndex nicht löschen.");
					}
				}
			}
		}
	}


	/**
	 * Schliesst einen Container, indem jeder Datensatz-Header des Containers gelesen wird, um die Min-/Max-Werte und die Anzahl der Datensätze zu bestimmen. Die
	 * Information der Indexe reicht nicht aus, weil dort die Anzahl der Datensätze nicht verfuegbar ist.
	 *
	 * @param vorhalteZeit parametrierte Vorhaltezeit in Sekunden
	 *
	 * @throws PersistenceException
	 */
	public void closeContainerRecount(long vorhalteZeit) throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.closeContainerRecount");
		checkContainerAccessedRW();
		if(_cache != null) {
			flushCache();
			forgetCache();
		}
		long minArcTime = Long.MAX_VALUE, maxArcTime = 0, minDataTime = Long.MAX_VALUE, maxDataTime = 0, minDataIdx = Long.MAX_VALUE, maxDataIdx = 0;
		ContainerDataIterator cdi = iterGetDataIterator();	// Prüft bereits auf accessed
		int i = 0;
		while(cdi.readNextDataHdr()) {
			i++;
			minDataIdx = Math.min(minDataIdx, cdi.getDataIndex());
			maxDataIdx = Math.max(maxDataIdx, cdi.getDataIndex());

			minDataTime = Math.min(minDataTime, cdi.getDataTime());
			maxDataTime = Math.max(maxDataTime, cdi.getDataTime());

			minArcTime = Math.min(minArcTime, cdi.getArcTime());
			maxArcTime = Math.max(maxArcTime, cdi.getArcTime());
			cdi.skipData();
		}
		cdi.close();
		closeContainer(i, minArcTime, maxArcTime, minDataTime, maxDataTime, minDataIdx, maxDataIdx, vorhalteZeit);
	}

	/**
	 * Prueft ob der mit einer der <code>accessContainer()</code>-Methoden spezifizierte Container bereits abgeschlossen ist.
	 *
	 * @return Kennzeichen, ob Container abgeschlossen ist
	 *
	 * @throws PersistenceException
	 */
	boolean isContainerClosed() throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.isContainerClosed");
		checkContainerAccessed();
		if(_cache != null) return false;
		if(!headerRead) readContainerHeader();
		return (getContHdrParamAsInt(ContainerHdr.CHP_ANZ_DS) >= 0); // Abgeschlossener Container fuehrt die Anzahl seiner Datensätze im Header
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Container: Header
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Erzeugt Default-Header-Parameter für den Anfang einer neuen Container-Datei. Bevor mit {@link #writeContainerHeader()} der Header geschrieben werden kann,
	 * muss entweder mit dieser Methode ein neuer Header erzeugt oder mit {@link #readContainerHeader()} ein Header eingelesen werden.
	 *
	 * @param toSave Sichern
	 */
	private void createDefaultContainerHeader(boolean toSave) throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.createDefaultContainerHeader");
		containerHdr.clear();

		// zu setzende Werte:
		containerHdr.setVal(ContainerHdr.CHP_CONT_ID, containerId);
		containerHdr.setVal(ContainerHdr.CHP_OBJ_ID, objId);
		containerHdr.setVal(ContainerHdr.CHP_ATG_ID, atgId);
		containerHdr.setVal(ContainerHdr.CHP_ASP_ID, aspId);
		containerHdr.setVal(ContainerHdr.CHP_SIM_VAR, sv);
		containerHdr.setVal(ContainerHdr.CHP_DATA_KIND, dataKind);
		containerHdr.setVal(ContainerHdr.CHP_TO_SAVE, toSave);

		// Default-Werte:
		containerHdr.setDefaultVal(ContainerHdr.CHP_LOESCHEN);
		containerHdr.setDefaultVal(ContainerHdr.CHP_ANZ_DS);
		containerHdr.setDefaultVal(ContainerHdr.CHP_DATA_IDX_MIN);
		containerHdr.setDefaultVal(ContainerHdr.CHP_DATA_IDX_MAX);
		containerHdr.setDefaultVal(ContainerHdr.CHP_DATA_TIME_MIN);
		containerHdr.setDefaultVal(ContainerHdr.CHP_DATA_TIME_MAX);
		containerHdr.setDefaultVal(ContainerHdr.CHP_ARC_TIME_MIN);
		containerHdr.setDefaultVal(ContainerHdr.CHP_ARC_TIME_MAX);
		containerHdr.setDefaultVal(ContainerHdr.CHP_MEDIUM_ID);
		containerHdr.setDefaultVal(ContainerHdr.CHP_DELETED);
		containerHdr.setDefaultVal(ContainerHdr.CHP_LOESCHUTZ);
		containerHdr.setDefaultVal(ContainerHdr.CHP_RESTORED);
		headerRead = true;
	}

	/**
	 * Schreibt die aktuellen Header-Parameter an den Anfang der Container-Datei. Bei einer neuen leeren Datei werden die Parameter an den Anfang der Datei
	 * angehangen, bei einer zu aendernden Container-Datei werden die vorhandenen Parameter ueberschrieben. Bevor mit dieser Methode der Header geschrieben werden
	 * kann, muss entweder mit {@link #createDefaultContainerHeader(boolean,long) neuer Header erzeugt oder mit {@link #readContainerHeader()} ein Header
	 * eingelesen worden sein.
	 *
	 * @throws PersistenceException
	 */
	public void writeContainerHeader() throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.writeContainerHeader");
		checkContainerAccessedHeaderRead();
		String hdrTxt = containerHdr.writeContainerHdr();
		ByteIO.writeSignedInt4Bytes(byte8Buf, 0, hdrTxt.length() + ByteIO.SEPARATOR.length); // Header-Laenge (4 Bytes)
		BufferedRandomAccessFile raf = null;
		try {
			ByteArrayOutputStream out = new ByteArrayOutputStream(getHeaderLen());
			out.write(VERSION_STRING);	                 // Header-Version
			out.write(byte8Buf, 0, DATALEN_LEN);		// Header-Laenge schreiben (aus readBuf)
			out.write(hdrTxt.getBytes());		        // Header-Parameter schreiben
			out.write(ByteIO.SEPARATOR);                // Header-Seperator schreiben
			raf = new BufferedRandomAccessFile(contFile); // Nur ReadWrite, kein flush
			final BufferedRandomAccessFile finalRaf = raf;
			// Inhalt des Puffers ohne neue Kopie des Byte-Arrays schreiben
			out.writeTo(
					new OutputStream() {
						public void write(int b) throws IOException {
							// Wird nicht aufgerufen; nur der Vollständigkeit halber implementiert.
							finalRaf.write(b);
						}

						public void write(byte b[], int off, int len) throws IOException {
							finalRaf.write(b, off, len);
						}
					}
			);
		}
		catch(IOException e) {
			throw new PersistenceException("Container-Header-Parameter konnte nicht geaendert werden: " + e.getMessage() + "\n" + this);
		}
		finally {
			closeRandomAccessFile(raf);
		}
	}

	/**
	 * Schreibt die aktuellen Header-Parameter an den Anfang der Container-Datei. Bei einer neuen leeren Datei werden die Parameter an den Anfang der Datei
	 * angehangen, bei einer zu aendernden Container-Datei werden die vorhandenen Parameter ueberschrieben. Bevor mit dieser Methode der Header geschrieben werden
	 * kann, muss entweder mit {@link #createDefaultContainerHeader(boolean,long) neuer Header erzeugt oder mit {@link #readContainerHeader()} ein Header
	 * eingelesen worden sein.
	 *
	 * @throws PersistenceException
	 */
	public void writeInitialContainerHeader() throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.writeInitialContainerHeader");
		checkContainerAccessedHeaderRead();
		String hdrTxt = containerHdr.writeContainerHdr();
		ByteIO.writeSignedInt4Bytes(byte8Buf, 0, hdrTxt.length() + ByteIO.SEPARATOR.length); // Header-Laenge (4 Bytes)
		if(_cache != null) {
			_cache.cache(VERSION_STRING, VERSION_STRING.length);
			_cache.cache(byte8Buf, DATALEN_LEN);
			final byte[] hdrBytes = hdrTxt.getBytes();
			_cache.cache(hdrBytes, hdrBytes.length);
			_cache.cache(ByteIO.SEPARATOR, ByteIO.SEPARATOR.length);
			// Cache wird später nach der Zwischenspeicherung des ersten Datensatzes geschrieben und damit im Filesystem sichtbar, weil manche Klassen das
			// Filesystem als Index verwenden (z.B. MultiContainerDataIterator)
		}
		else {
			BufferedOutputStream out = null;
			try {
				out = new BufferedOutputStream(new FileOutputStream(contFile, false), getHeaderLen());
				out.write(VERSION_STRING);	                 // Header-Version
				out.write(byte8Buf, 0, DATALEN_LEN);		// Header-Laenge schreiben (aus readBuf)
				out.write(hdrTxt.getBytes());		        // Header-Parameter schreiben
				out.write(ByteIO.SEPARATOR);                // Header-Seperator schreiben
				out.flush();
			}
			catch(IOException e) {
				throw new PersistenceException("Container-Header-Parameter konnte nicht geaendert werden: " + e.getMessage() + "\n" + this);
			}
			finally {
				if(out != null) {
					try {
						out.close();
					}
					catch(IOException e) {
						throw new PersistenceException("Container-Datei konnte nicht geschlossen werden: " + e.getMessage() + "\n" + this);
					}
				}
			}
		}
	}

	void readContainerHeader(FileAccess raf) throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.readContainerHeader(raf)");
		try {
			raf.readFully(byte8Buf, 0, VERSION_STRING.length + ByteIO.INT4B_LEN);
			int headerLength = ByteIO.readSignedInt4Bytes(byte8Buf, VERSION_STRING.length); // Laenge Header und Seperator

			byte[] buf = getBuf(readBuf, headerLength);
			raf.readFully(buf, 0, headerLength);
			containerHdr.readContainerHdr(buf, headerLength);
			headerRead = true; // Header eingelesen
		}
		catch(Throwable e) {
			throw new PersistenceException("Container-Header konnte nicht gelesen werden: " + e.getMessage() + "\n" + this, e);
		}
	}

	/**
	 * Liest den Header der Container-Datei ein und springt im Eingabestrom an die Stelle hinter dem Header. Die Container-Header Parameter stehen dann als
	 * Key/Value-Paare zur Verfuegung und koennen mit {@link #getContainerHeaderParam(KeyValParam)} abgefragt werden. Bevor mit {@link #writeContainerHeader()} der
	 * Header geschrieben werden kann, muss entweder mit dieser Methode der Header eingelesen oder mit {@link #createDefaultContainerHeader(boolean,long)} ein
	 * neuer Header erzeugt werden.
	 *
	 * @throws PersistenceException
	 */
	void readContainerHeader() throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.readContainerHeader");
		if(_cache != null) flushCache();
		try {
			checkContainerAccessed();
			BufferedRandomAccessFile raf = new BufferedRandomAccessFile(contFile);
			readContainerHeader(raf);
			closeRandomAccessFile(raf);
		}
		catch(Exception e) {
			throw new PersistenceException("Container-Header konnte nicht gelesen werden: " + e.getMessage() + "\n" + this, e);
		}
	}

	/**
	 * Springt im uebergebenen Eingabestrom an die Stelle hinter dem Header der Container-Datei. Voraussetzung ist, dass das RandomAccessFile am Anfang der Datei
	 * steht.
	 *
	 * @param raf RandomAccessFile der Container-Datei
	 *
	 * @throws IOException
	 */
	private void skipContainerHeader(FileAccess raf) throws IOException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.skipContainerHeader(raf)");
		raf.readFully(byte8Buf, 0, VERSION_STRING.length + ByteIO.INT4B_LEN);
		raf.skipBytes(ByteIO.readSignedInt4Bytes(byte8Buf, VERSION_STRING.length));
	}

	/**
	 * Nachdem mit {@link #readContainerHeader()} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines Header-Parameters
	 * ermittelt werden. Wirft eine {@link PersistenceException}, falls der Parameter nicht gefunden wurde. Liefert den Wert des Parameters unabhaengig von dessen
	 * Typ immer als String zurueck.
	 *
	 * @param param Container-Header-Parameter
	 *
	 * @return Parameter-Wert als String
	 *
	 * @throws PersistenceException
	 */
	String getContHdrParamAsString(KeyValParam param) throws PersistenceException {
		checkContainerAccessedHeaderRead();
		String val = containerHdr.getValAsString(param);
		if(val != null) return val;
		throw new PersistenceException("ContainerHeaderParam nicht vorhanden: " + param + "\n" + toString());
	}

	/**
	 * Nachdem mit {@link #readContainerHeader()} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines Header-Parameters
	 * ermittelt werden. Liefert den Wert des Parameters als Datensatzart zurueck. Wirft eine {@link PersistenceException}, falls der Parameter nicht gefunden
	 * wurde oder nicht in den Typ {@link ArchiveDataKind} umgeformt werden kann.
	 *
	 * @param param Container-Header-Parameter
	 *
	 * @return Datensatzart
	 *
	 * @throws PersistenceException
	 */
	ArchiveDataKind getContHdrParamAsArchiveDataKind(KeyValParam param) throws PersistenceException {
		checkContainerAccessedHeaderRead();
		ArchiveDataKind adk = containerHdr.getValAsArchiveDataKind(param);
		if(adk != null) return adk;
		throw new PersistenceException("ContainerHeaderParam nicht besetzt: " + param + "\n" + toString());
	}

	/**
	 * Nachdem mit {@link #readContainerHeader()} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines Header-Parameters
	 * ermittelt werden. Liefert den Wert des Parameters als numerischen Wert vom Typ <code>long</code> zurueck. Wirfte eine {@link PersistenceException}, falls
	 * der Parameter nicht gefunden wurde oder nicht in den Typ <code>long</code> umgeformt werden kann.
	 *
	 * @param param Container-Header-Parameter
	 *
	 * @return Ganzzahl als <code>long</code>
	 *
	 * @throws PersistenceException
	 */
	long getContHdrParamAsLong(KeyValParam param) throws PersistenceException {
		checkContainerAccessedHeaderRead();
		Long l = containerHdr.getValAsLong(param);
		if(l != null) return l;
		throw new PersistenceException("ContainerHeaderParam nicht besetzt: " + param + "\n" + toString());
	}

	int getContHdrParamAsInt(KeyValParam param) throws PersistenceException {
		checkContainerAccessedHeaderRead();
		Integer i = containerHdr.getValAsInt(param);
		if(i != null) return i;
		throw new PersistenceException("ContainerHeaderParam nicht besetzt: " + param + "\n" + toString());
	}

	boolean getContHdrParamAsBool(KeyValParam param) throws PersistenceException {
		checkContainerAccessedHeaderRead();
		Boolean b = containerHdr.getValAsBool(param);
		if(b != null) return b;
		throw new PersistenceException("ContainerHeaderParam nicht besetzt: " + param + "\n" + toString());
	}

	/**
	 * Setzt einen Parameters im Container-Header. Der Header muss vorher entweder gelesen oder mit Default-Werten erzeugt worden sein. Wirft eine {@link
	 * PersistenceException}, wenn die Aenderung nicht durchgefuehrt werden kann.
	 *
	 * @param param Container-Header-Parameter
	 * @param val   Numerischer Wert des Container-Header-Parameters
	 *
	 * @throws PersistenceException
	 */
	void setContHdrParam(KeyValParam param, long val) throws PersistenceException {
		checkContainerAccessedHeaderRead();
		containerHdr.setVal(param, val);
	}

	/**
	 * Setzt einen Parameters im Container-Header. Der Header muss vorher entweder gelesen oder mit Default-Werten erzeugt worden sein. Wirft eine {@link
	 * PersistenceException}, wenn die Aenderung nicht durchgefuehrt werden kann.
	 *
	 * @param param Container-Header-Parameter
	 * @param val   Wahrheitswert des Container-Header-Parameters
	 *
	 * @throws PersistenceException
	 */
	void setContHdrParam(KeyValParam param, boolean val) throws PersistenceException {
		checkContainerAccessedHeaderRead();
		containerHdr.setVal(param, val);
	}

	/**
	 * Setzt einen Parameters im Container-Header. Der Header muss vorher entweder gelesen oder mit Default-Werten erzeugt worden sein. Wirft eine {@link
	 * PersistenceException}, wenn die Aenderung nicht durchgefuehrt werden kann.
	 *
	 * @param param Container-Header-Parameter
	 * @param val   Textwert des Container-Header-Parameters
	 *
	 * @throws PersistenceException
	 */
	void setContHdrParam(KeyValParam param, String val) throws PersistenceException {
		checkContainerAccessedHeaderRead();
		containerHdr.setVal(param, val);
	}

	/**
	 * Setzt einen Parameters im Container-Header. Der Header muss vorher entweder gelesen oder mit Default-Werten erzeugt worden sein. Wirft eine {@link
	 * PersistenceException}, wenn die Aenderung nicht durchgefuehrt werden kann.
	 *
	 * @param param Container-Header-Parameter
	 * @param val   Datensatzart des Container-Header-Parameters
	 *
	 * @throws PersistenceException
	 */
	void setContHdrParam(KeyValParam param, ArchiveDataKind val) throws PersistenceException {
		checkContainerAccessedHeaderRead();
		containerHdr.setVal(param, val);
	}

	/**
	 * Der Container-Header der Container-Datei, die mit der <code>accessContainer()</code>-Methode spezifiziert worden ist, wird mit dem Container-Header aus der
	 * angegebenen Container-Datei ueberschrieben.
	 *
	 * @param srcFile Container-Datei
	 *
	 * @throws PersistenceException
	 */
	void restoreHeader(File srcFile) throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.restoreHeader");
		checkContainerAccessedRW();
		if(_cache != null) {
			flushCache();
			forgetCache();
		}
		try {
			FileInputStream origContStr = new FileInputStream(contFile);
			origContStr.read(byte8Buf, 0, VERSION_STRING.length + ByteIO.INT4B_LEN);
			long oldHdrLenNoLenField = ByteIO.readSignedInt4Bytes(byte8Buf, VERSION_STRING.length); // Laenge Header und Längenbyte
			origContStr.close();

			FileInputStream srcContStr = new FileInputStream(srcFile);
			srcContStr.read(byte8Buf, 0, VERSION_STRING.length + ByteIO.INT4B_LEN);
			int newHdrLenNoLenField = ByteIO.readSignedInt4Bytes(byte8Buf, VERSION_STRING.length); // Laenge Header und Längenbyte

			byte[] newHdr = new byte[newHdrLenNoLenField];
			srcContStr.read(newHdr, 0, newHdrLenNoLenField);
			srcContStr.close();

			if(newHdrLenNoLenField == oldHdrLenNoLenField) {
				RandomAccessFile raf = new RandomAccessFile(contFile, "rw"); // Nur ReadWrite, kein flush
				raf.seek(0);
				raf.write(byte8Buf, 0, VERSION_STRING.length + ByteIO.INT4B_LEN);
				raf.write(newHdr, 0, newHdrLenNoLenField);  // Header drueberschreiben
				raf.close();
			}
			else {
				File tmpFile = Util.deleteCreateNewFile(contFile.getParentFile(), contFile.getName() + ".tmp");
				origContStr = new FileInputStream(contFile);
				FileOutputStream newContStr = new FileOutputStream(tmpFile);
				origContStr.read(byte8Buf, 0, VERSION_STRING.length);
				newContStr.write(byte8Buf, 0, VERSION_STRING.length);
				newContStr.write(newHdr, 0, newHdrLenNoLenField);
				origContStr.skip(oldHdrLenNoLenField);
				Util.copyStreams(origContStr, newContStr);
				origContStr.close();
				newContStr.close();
				contFile.delete();
				tmpFile.renameTo(contFile);
			}
			readContainerHeader();	// Header muss neu eingelesen werden
		}
		catch(IOException e) {
			throw new PersistenceException(
					"Container-Header konnte nicht wiederhergestellt werden. Quelle: '" + srcFile.getAbsolutePath() + "' Grund: " + e + "\n" + this
			);
		}
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Container: Datensätze
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/** Legt alle zum Schreiben benoetigte Puffer an. Wird nur von Schreib-Methoden aufgerufen. Lesende Tasks beanspruchen so erheblich weniger Speicher. */
	private void createWriteBuffers() {
		if(DEBUG_CALLS) System.out.println("ContainerFile.createWriteBuffers");
		// Nur anlegen, falls schreibender Task (ArchiveTask) zugreift
		if(bos == null) bos = new ByteArrayOutputStream(BUFFER_SIZE);
		if(writeBuf == null) writeBuf = new byte[BUFFER_SIZE];
		if(serializer == null) serializer = SerializingFactory.createSerializer(bos);
	}

	/**
	 * Serialisiert den uebergebenen Datensatz und speichert ihn in {@link ContainerFile#actualWriteBuf} und {@link ContainerFile#totalWriteDataSize}. Da der
	 * ArchivTask vor dem Archivieren wissen muss, ob der Container zu lang wird und abgeschlossen werden muss, ist das Anhaengen eines Datensatzes auf die
	 * Methoden {@link ContainerFile#serializeData(ResultData,long)} und {@link ContainerFile#appendSerializedData()} aufgeteilt. Wenn der Datensatz ein
	 * Spezialfall ist (keine Quelle, keine Daten, Datenluecke), wird ein speziell codierter Datensatz verwendet.
	 *
	 * @param rd      Zu archivierender Datensatz
	 * @param arcTime Archivzeitstempel
	 *
	 * @return Laenge des serialisierten Datensatzes
	 *
	 * @throws PersistenceException
	 */
	int serializeData(ResultData rd, long arcTime, Deflater compresser) throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.serializeData");
		createWriteBuffers();	// Puffer (nur für schreibende Tasks) lazy anlegen

		byte[] data;
		if(IS_PERFORMANCE_TEST) {
			data = PERF_TESTDATA;
		}
		else if(rd.hasData()) {
			// normaler Datensatz
			bos.reset();
			try {
				serializer.writeData(rd.getData());
			}
			catch(IOException e) {
				throw new PersistenceException("Datensatz konnte nicht serialisiert werden: " + e.getMessage() + "\n" + this + "\n" + rd);
			}
			data = bos.toByteArray();
		}
		else {
			// kuenstlicher Datensatz (keine Quelle, keine Daten)
			data = writeDataModifier(rd);
		}
		actualWriteBuf = getBuf(writeBuf, getTotalDataOverhead() + data.length);	// Nach Moeglichkeit bereits angelegten Puffer verwenden
		return embedSerialializedData(arcTime, rd.getDataTime(), rd.getDataIndex(), data, actualWriteBuf, compresser);
	}

	/**
	 * Serialisiert den uebergebenen Datensatz und speichert ihn in {@link ContainerFile#actualWriteBuf} und {@link ContainerFile#totalWriteDataSize}. Da der
	 * ArchivRequestedTask vor dem Archivieren wissen muss, ob der Container zu lang wird und abgeschlossen werden muss, ist das Anhaengen eines Datensatzes auf
	 * die Methoden {@link ContainerFile#serializeReqData(ArchivetData,long)} und {@link ContainerFile#appendSerializedData()} aufgeteilt. Wenn der Datensatz ein
	 * Spezialfall ist (keine Quelle, keine Daten), wird ein speziell codierter Datensatz verwendet. <p>Diese Methode wird zum serialisieren von nachgeforderten
	 * Datensätzen verwendet.
	 *
	 * @param ad      Zu archivierender Datensatz
	 * @param arcTime Archivzeitstempel
	 *
	 * @return Laenge des serialisierten Datensatzes
	 *
	 * @throws PersistenceException
	 */
	int serializeReqData(ArchiveData ad, long arcTime, Deflater compresser) throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.serializeReqData");
		createWriteBuffers();	// Puffer (nur für schreibende Tasks) lazy anlegen

		byte[] data;
		DataState dstate = ad.getDataType();

		// Nur die drei Faelle werden hier behandelt. Luecken werden ausserhalb erkannt (und
		// beim naechsten Archivsystem angefragt).

		if(dstate.equals(DataState.DATA)) {
			bos.reset();
			try {
				serializer.writeData(ad.getData());
			}
			catch(IOException e) {
				throw new PersistenceException("Datensatz konnte nicht serialisiert werden: " + e.getMessage() + "\n" + this + "\n" + Util.ad2Str(ad));
			}
			data = bos.toByteArray();
		}
		else if(dstate.equals(DataState.NO_DATA)) {
			data = NO_DATA;
		}
		else if(dstate.equals(DataState.NO_SOURCE)) {
			data = NO_SOURCE;
		}
		else {
			throw new PersistenceException("DataState eines nachgeforderten Datensatzes nicht DATA/NO_DATA/NO_SOURCE: " + dstate);
		}

		actualWriteBuf = getBuf(writeBuf, getTotalDataOverhead() + data.length);	// Nach Moeglichkeit bereits angelegten Puffer verwenden
		return embedSerialializedData(arcTime, ad.getDataTime(), ad.getDataIndex(), data, actualWriteBuf, compresser);
	}

	/**
	 * Serialisiert einen Datensatz der eine Datenluecke kennzeichnet und speichert ihn in {@link ContainerFile#actualWriteBuf} und {@link
	 * ContainerFile#totalWriteDataSize}. Da der ArchivTask vor dem Archivieren wissen muss, ob der Container zu lang wird und abgeschlossen werden muss, ist das
	 * Anhaengen eines Datensatzes auf die Methoden {@link ContainerFile#serializeData(ResultData,long)} und {@link ContainerFile#appendSerializedData()}
	 * aufgeteilt.
	 *
	 * @param dataIndex   Datenindex
	 * @param dataTime    Datenzeit
	 * @param archiveTime Archivzeit
	 *
	 * @return Laenge des serialisierten Datensatzes
	 *
	 * @throws PersistenceException
	 */
	int serializeDataGap(long dataIndex, long dataTime, long archiveTime, Deflater compresser) throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.serializeDataGap");
		createWriteBuffers();	// Puffer (nur für schreibende Tasks) lazy anlegen
		actualWriteBuf = getBuf(writeBuf, getTotalDataOverhead() + POT_GAP.length);	// Nach Moeglichkeit bereits angelegten Puffer verwenden
		return embedSerialializedData(archiveTime, dataTime, dataIndex, POT_GAP, actualWriteBuf, compresser);
	}

	/**
	 * Schreibt das Bytefeld mit allen Datensatz-Headern, so dass es mit {@link ContainerFile#appendSerializedData()} geschrieben werden kann. Von aussen wird
	 * diese Methode nur für Testzwecke verwendet. Normalerweise sind die Methoden {@link ContainerFile#serializeData(ResultData,long)} bzw. {@link
	 * ContainerFile#serializeReqData(ResultData,long)} zu verwenden. Die Daten werden komprimiert abgelegt, wenn sie laenger als {@link #MAX_UNCOMPRESSED} sind
	 * und die Kompression eine Verkuerzung ergeben hat.
	 *
	 * @param aTime   Archivzeit
	 * @param dTime   Datenzeit
	 * @param dIdx    Datenindex
	 * @param serData zu schreibendes Bytefeld
	 * @param dest    Ziel
	 *
	 * @return Laenge des serialisierten Datensatzes
	 *
	 * @throws PersistenceException
	 */
	int embedSerialializedData(long arcTime, long dataTime, long dataIdx, byte[] serData, byte[] dest, Deflater compresser) throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.embedSerialializedData");
		int p = ByteIO.INT4B_LEN;									// [1. Platzhalter f. Laenge]
		p += ByteIO.writeUnsignedLong6Byte(dest, p, arcTime);		// 2. Archivzeit
		p += ByteIO.writeUnsignedLong6Byte(dest, p, dataTime);    // 3. Datenzeit
		p += ByteIO.writeSignedLong8Byte(dest, p, dataIdx);		// 4. Datenindex
		int chkSumPos = p;											// [5. Platzhalter für Pruefsumme]
		p += ByteIO.INT4B_LEN;

		int netDataLen = 0;
		boolean uncompressed = true;

		if(serData.length > MAX_UNCOMPRESSED) {
			int compressInfoPos = p;				// [6. Platzhalter f. Kompressions-Info]
			p += ByteIO.INT4B_LEN;

			// Nur schreibende Tasks landen hier. Alle anderen lesen nur und brauchen keinen Deflater.
			// Notwendig, weil das Teil extrem Speicher braucht.
			if(compresser == null) compresser = new Deflater();

			compresser.reset();
			compresser.setInput(serData);
			compresser.finish();
			netDataLen = compresser.deflate(dest, p, dest.length - p - ByteIO.SEPARATOR.length);// 7. Daten

			// Hat das Komprimieren was gebracht? Wenn nicht, unkomprimierte Daten drueberschreiben
			if(netDataLen < serData.length && compresser.finished()) {
				ByteIO.writeSignedInt4Bytes(dest, compressInfoPos, serData.length);   // 6. Kompressions-Info (orig. Datenlaenge)
				p += netDataLen;
				uncompressed = false;
			}
			else {
				p = compressInfoPos;
			}
		}

		if(uncompressed) {
			p += ByteIO.writeSignedInt4Bytes(dest, p, NOT_COMPRESSED);	// 6. Kompressions-Info (=0)
			p += ByteIO.writeBytes(dest, p, serData);		 // 7. Daten
			netDataLen = serData.length;
		}
		totalWriteDataSize = getTotalDataOverhead() + netDataLen;	 // tatsaechliche Laenge ueber alles
		int lengthBytes = totalWriteDataSize - DATALEN_LEN;
		int checkSum = (int)(lengthBytes + arcTime + dataTime + dataIdx);

		ByteIO.writeSignedInt4Bytes(dest, 0, lengthBytes);		   // 1. Laenge
		ByteIO.writeSignedInt4Bytes(dest, chkSumPos, checkSum);   // 5. Pruefsumme
		ByteIO.writeSeparator(dest, p);						// 8. Separator
		return totalWriteDataSize;
	}

	/**
	 * Haengt den bereits mit {@link ContainerFile#preserializeData(ResultData,long)} uebergebenen und serialisierten Datensatz an die aktuelle Container-Datei
	 * an.
	 *
	 * @throws PersistenceException
	 */
	void appendSerializedDataRaf() throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.appendSerializedDataRaf");
		checkContainerAccessedRW();
		BufferedRandomAccessFile raf = null;
		try {
			raf = new BufferedRandomAccessFile(contFile);
			long le = raf.length();
			if(le == 0) throw new PersistenceException("Container enthaelt keinen Header:\n" + this);
			raf.seek(le);
			raf.write(actualWriteBuf, 0, totalWriteDataSize);
		}
		catch(IOException e) {
			throw new PersistenceException("Archivdatensatz konnte nicht in Container-Datei geschrieben werden: " + e.getMessage() + "\n" + this);
		}
		finally {
			closeRandomAccessFile(raf);
		}
	}

	/**
	 * Haengt den bereits uebergebenen und serialisierten Datensatz an die aktuelle Container-Datei an.
	 *
	 * @throws PersistenceException
	 */
	void appendSerializedData() throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.appendSerializedData");
		checkContainerAccessedRW();
		if(_cache == null) {
			_cache = _cacheManager.createCache(this, false);
		}
		if(_cache != null) {
			boolean isFirstDataInContainer = (_cache.getContainerSize() == getHeaderLen());
			_cache.cache(actualWriteBuf, totalWriteDataSize);
			// Nach dem cachen des ersten Datensatzes muss der Cache geflusht werden, damit die Datei vorhanden ist und gefunden werden kann.
			// Ein flushen nach dem cachen des Headers wird unterdrückt, damit keine Container-Dateien ohne Daten entstehen
			if(isFirstDataInContainer) {
				_cache.flush();
			}
		}
		else {
			FileOutputStream out = null;
			try {
				out = new FileOutputStream(contFile, true);
				out.write(actualWriteBuf, 0, totalWriteDataSize);
			}
			catch(FileNotFoundException e) {
				throw new PersistenceException("Container Datei zum Anhängen eines Datensatzes nicht gefunden:\n" + this);
			}
			catch(IOException e) {
				throw new PersistenceException("Archivdatensatz konnte nicht in Container-Datei geschrieben werden: " + e.getMessage() + "\n" + this);
			}
			finally {
				if(out != null) {
					try {
						out.close();
					}
					catch(IOException e) {
						throw new PersistenceException(
								"Container-Datei konnte beim Anhängen eines Datensatzes nicht geschlossen werden: " + e.getMessage() + "\n" + this
						);
					}
				}
			}
		}
	}

	/**
	 * Wenn der Originaldatensatz keine Daten enthaelt, liefert diese Methode die Daten, die stattdessen im Container (zur Markierung) gespeichert werden sollen.
	 * Derartige Daten werden nur vom DAV erzeugt und muessen deswegen immer ONLINE sein.
	 *
	 * @param rd Datensatz
	 *
	 * @return Daten bzw. null, falls ein Fehler vorliegt
	 *
	 * @throws PersistenceException
	 */
	private byte[] writeDataModifier(ResultData rd) throws PersistenceException {
		if(!rd.getDataKind().equals(ArchiveDataKind.ONLINE)) {
			throw new PersistenceException("Fehler beim Serialisieren: illegalen Datensatz mit DAV-Bit erhalten: " + Util.rd2Str(rd));
		}
		if(rd.isNoSourceAvailable()) {
			return NO_SOURCE;
		}
		else if(rd.isNoDataAvailable()) {
			return NO_DATA;
		}
		else if(rd.isNoRightsAvailable()) {
			return NO_RIGHTS;
		}
		else {
			return null;
		}
	}

	/**
	 * Zaehlt die Anzahl der Datensätze in dem mit einer <code>accessContainer()</code>-Methode im Zugriff befindlichen Container durch Iterieren ueber die
	 * Datensätze in der Container-Datei. Diese Methode ist nur notwendig, wenn ein noch nicht abgeschlossener Container wieder befuellt wird, weil dann die
	 * Anzahl nicht im Header steht. Im Normalbetrieb wird die Anzahl im DataIdentNode mitgefuehrt.
	 *
	 * @return Anzahl Datensätze im Container
	 *
	 * @throws PersistenceException
	 */
	int countContainerEntries() throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.countContainerEntries");
		ContainerDataIterator cdi = iterGetDataIterator();	// Prüft bereits auf accessed
		int i = 0;
		try {
			while(cdi.readNextDataHdr()) {
				i++;
				cdi.skipData();
			}
		}
		finally {
			cdi.close();
		}
		return i;
	}

	/**
	 * Ermittelt die Gesamtlaenge des mit <code>accessContainer()</code> im Zugriff befindlichen Containers.
	 *
	 * @return Laenge der Container-Datei
	 *
	 * @throws PersistenceException
	 */
	long getContainerSize() throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.getContainerSize");
		checkContainerAccessed();
		if(_cache != null) {
			return _cache.getContainerSize();
		}
		return contFile.length();
	}

	/**
	 * Ermittelt die Gesamtlaenge des mit <code>accessContainer()</code> im Zugriff befindlichen Containers.
	 *
	 * @return Laenge der Container-Datei
	 *
	 * @throws PersistenceException
	 */
	public static int getHeaderLen() {
		return HEADER_LENGTH;
	}

	/**
	 * Ermittelt den vollstaendigen Overhead für einen Datensatz. Inklusive Laengenbyte, Datensatz-Header, Compress-Feld und abschliessendem Separator.
	 *
	 * @return
	 */
	public static int getTotalDataOverhead() {
		return DATALEN_LEN + DATAHDR_LEN + COMPRESS_LEN + ByteIO.SEPARATOR.length;
	}

	/**
	 * Loescht alle Datensätze eines Datencontainers, so dass lediglich der Header mit der Verwaltungs-Information in der Datei erhalten bleibt.<br><br> Das
	 * Header-Flag darf keinesfalls vor dem Loeschen der Daten auf "geloescht" gesetzt werden!<br><br> Update 03/05/2006: <br> Schreibt eine optimierte Liste der
	 * Datenindizes (siehe {@link #createDataIndexBuffer()}) in den Body des Datencontainers, damit Datensätze aus geloeschten Containern beim Nachfordern
	 * ignoriert werden koennen.
	 *
	 * @throws PersistenceException
	 */
	void deleteContainerData() throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.deleteContainerData");
		checkContainerAccessedRW();
		if(_cache != null) {
			flushCache();
			forgetCache();
		}
		BufferedRandomAccessFile raf = null;
		try {
			// Header überspringen
			raf = new BufferedRandomAccessFile(contFile);  // Nur ReadWrite, kein flush
			// Eine Liste der Datensätze (Indizes) aufbauen
			byte[] buffer = createDataIndexBuffer();
			// Nach dem Header die verkürzte Liste der Indizes schreiben
			skipContainerHeader(raf);
			raf.write(buffer);
			// Alles danach abschneiden
			raf.setLength(raf.getFilePointer());
		}
		catch(Exception e) {
			throw new PersistenceException("Datensaetze konnten nicht aus Container geloescht werden.", e);
		}
		finally {
			closeRandomAccessFile(raf);
		}
	}

	/**
	 * Erzeugt eine verkürzte Liste von Datenindizes, die im aktuellen Datencontainer archiviert sind und schreibt sie in ein Byte-Array (falls z.B.
	 * 0,4,8,16,24,28,32 die Indizes der Datensätze sind, wird daraus: 0-8,16,24-32)
	 *
	 * @return die Liste als Byte-Array
	 *
	 * @throws PersistenceException
	 */
	byte[] createDataIndexBuffer() throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.createDataIndexBuffer");
		// Maximalgrösse des Containerbodys nach dem Löschen bestimmen
		int entries = this.countContainerEntries();
		int sepLen = indexListSep.length;
		int maxSize = entries * (2 * ByteIO.LONG8B_LEN + sepLen);

		byte[] buf = new byte[maxSize];
		dataIter = iterGetDataIterator();
		try {
			long last = Long.MIN_VALUE;
			long current = Long.MIN_VALUE;
			int pos = 0;

			while(dataIter.readNextDataHdr()) {
				current = dataIter.getDataIndex();
				if(pos == 0) {
					pos += ByteIO.writeSignedLong8Byte(buf, pos, current);
					pos += ByteIO.writeBytes(buf, pos, indexListSep);
					last = current;
				}
				else {
					if(isGap(last, current)) {
						pos += ByteIO.writeSignedLong8Byte(buf, pos, last);
						pos += ByteIO.writeSignedLong8Byte(buf, pos, current);
						pos += ByteIO.writeBytes(buf, pos, indexListSep);
						last = current;
					}
				}
				if(!dataIter.isPotDataGap()) last = current;
				dataIter.skipData();
			}
			pos += ByteIO.writeSignedLong8Byte(buf, pos, last);
			// Buffer beschneiden, falls er zu gross dimensioniert war
			if(pos < maxSize) {
				byte[] tmp = new byte[pos];
				System.arraycopy(buf, 0, tmp, 0, tmp.length);
				return tmp;
			}
			// Buffer zurückgeben, falls er genau Maximalgrösse hat
			else {
				return buf;
			}
		}
		finally {
			dataIter.close();
		}
	}

	/**
	 * Prüft ob zwischen dem Index start und end eine Lücke ist.
	 *
	 * @param start
	 * @param end
	 *
	 * @return
	 */
	private static boolean isGap(long start, long end) {
		return start == -1 ? false : noModBits(end) - noModBits(start) > 1;
	}

	private static long noModBits(long dataIndex) {
		return Util.dIdxNoModBits(dataIndex);
	}


	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Daten-Iterator
	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	/**
	 * Liefert einen Iterator zum Durchlaufen aller Datensätze des aktuellen Containers.
	 *
	 * @param atg             Attributgruppe für das Deserialisieren der Datensätze. Kann null sein, falls die Datensätze nur uebersprungen werden sollen.
	 *                        <p>
	 *                        Vorsicht: Der ContainerDataIterator verwendet den Lesepuffer der Klasse {@link ContainerFile}. Andere Leseoperationen koennen den
	 *                        Lesepuffer ueberschreiben, so dass der ContainerDataIterator nicht mehr funktioniert.
	 * @param considerDeleted Gibt an, ob ausgelagerte Container besonders behandelt werden sollen.
	 *
	 * @return ContainerDataIterator für das Durchlaufen der Datensätze
	 *
	 * @throws PersistenceException
	 */
	synchronized ContainerDataIterator iterGetDataIterator(boolean considerDel) throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.iterGetDataIterator");
		if(_cache != null) flushCache();
		// Flag für Sonderbehandlung für ausgelagerte Container setzen
		deletedSpecialTreatment = considerDel;
		checkContainerAccessed();
		if(iterRaf != null) throw new PersistenceException("Container-Iterator bereits in Verwendung.\n" + toString());
		iterIsHdrIterator = false;
		iterBufPos = 0;
		iterBytesRead = 0;
		iterNextDataLen = 0;
		iterDataRead = true;
		iterEOC = false;
		dataIter.setCounter(-1);
		dataIter.setArchiveDataKind(dataKind);
		try {
			iterRaf = new MyCloseableRandomAccessFile();
			int numOpenFiles = _numOpenFiles.get();
			while(numOpenFiles > OPEN_FILE_LIMIT) {
				MyCloseableRandomAccessFile first = removeOldest(_openFilesInCurrentThread.get());
				if(first != null) {
					// Braucht nicht aus Queue entfernt zu werden, da wir pollFirst() gemacht haben
					first.closeFile(false);
				}
				else {
					break;
				}
			}
			if(deletedSpecialTreatment) {
				readContainerHeader(iterRaf);
				deletedHeaderFlag = containerHdr.getValAsBool(ContainerHdr.CHP_DELETED);
				if(deletedHeaderFlag) {
					setListOfIndexesDeleted(iterRaf);
				}
			}
			else {
				skipContainerHeader(iterRaf);
			}
		}
		catch(IOException e) {
			iterClose();
			throw new PersistenceException("Container-Datei konnte nicht zum Iterieren geoeffnet werden: " + e.getMessage() + "\n" + this);
		}
		return dataIter;
	}

	/**
	 * Entfernt den ältestem Eintrag aus dem LinkedHashSet und gibt ihn zurück. Falls das Set leer ist wird es nicht verändert und null zurückgegeben.
	 */
	@Nullable
	private static <T> T removeOldest(final LinkedHashSet<T> set) {
		if(set.isEmpty()) return null;
		Iterator<T> iterator = set.iterator();
		T next = iterator.next();
		iterator.remove();
		return next;
	}

	synchronized ContainerDataIterator iterGetHeaderIterator() throws PersistenceException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.iterGetHeaderIterator");
		checkContainerAccessed();
		if(iterRaf != null) throw new PersistenceException("Container-Iterator bereits in Verwendung.\n" + toString());
		iterIsHdrIterator = true;
		iterBufPos = 0;	   // hierdrin wird bis 2 gezaehlt (Min- und Max-Wert)
		iterDataRead = true;
		iterEOC = false;
		dataIter.setCounter(-1);
		dataIter.setArchiveDataKind(dataKind);
		if(!headerRead) readContainerHeader();
		return dataIter;
	}

	/**
	 * Erzeugt eine Liste der Datensatz-Indizes, die im durch raf spezifizierten Container vorkommen (sortiert)
	 *
	 * @param raf
	 *
	 * @throws IOException
	 */
	void setListOfIndexesDeleted(FileAccess raf) throws IOException {
		if(DEBUG_CALLS) System.out.println("ContainerFile.setListOfIndexesDeleted(raf)");
		iterlistOfDataIdx = new ArrayList<Long>();
		iterlistOfDataIdxPos = 0;
		byte[] first = new byte[ByteIO.LONG8B_LEN];
		byte[] last = new byte[ByteIO.LONG8B_LEN];

		while(raf.position() != raf.length()) {
			raf.readFully(first);
			raf.skip(1); // Separator
			raf.readFully(last);
			long firstIdx = ByteIO.readSignedLong8Byte(first, 0);
			long lastIdx = ByteIO.readSignedLong8Byte(last, 0);
			for(long i = firstIdx; i <= lastIdx;) {
				iterlistOfDataIdx.add(i);
				i += 4;
			}
		}
	}

	/**
	 * Für Testfälle
	 * @return
	 */
	ArrayList<Long> getListOfIndexesDeleted() {
		return iterlistOfDataIdx;
	}


	/**
	 * Liefert einen Iterator zum Durchlaufen aller Datensätze des aktuellen Containers.
	 *
	 * @param atg Attributgruppe für das Deserialisieren der Datensätze. Kann null sein, falls die Datensätze nur uebersprungen werden sollen.
	 *            <p>
	 *            Vorsicht: Der ContainerDataIterator verwendet den Lesepuffer der Klasse {@link ContainerFile}. Andere Leseoperationen koennen den Lesepuffer
	 *            ueberschreiben, so dass der ContainerDataIterator nicht mehr funktioniert.
	 *
	 * @return ContainerDataIterator für das Durchlaufen der Datensätze
	 *
	 * @throws PersistenceException
	 */
	synchronized ContainerDataIterator iterGetDataIterator() throws PersistenceException {
		return iterGetDataIterator(false);
	}

	/**
	 * Prueft, ob ein weiterer Datensatz zur letzten Abfrage verfuegbar ist. Wird vom ContainerDataIterator benutzt, um ueber Datensätze zu iterieren.
	 *
	 * @return Kennzeichen
	 *
	 * @throws PersistenceException
	 */
	boolean iterReadNextDataHdr() throws PersistenceException {
		checkContainerAccessed();
		// Spezialbehandlung für gelöschte Container beim Nachfordern
		if(deletedSpecialTreatment && iterlistOfDataIdx.size() > 0) {
			if(iterlistOfDataIdxPos < iterlistOfDataIdx.size()) {
				dataIter.setDataIndex(iterlistOfDataIdx.get(iterlistOfDataIdxPos));
				return true;
			}
			else {
				iterlistOfDataIdx = new ArrayList<Long>();
				iterlistOfDataIdxPos = -1;
				return false;
			}
		}
		// Keine Spezialbehandlung in allen anderen Fällen
		if(!iterDataRead) throw new PersistenceException("Fehler: Datensatz wurde nicht gelesen oder übersprungen.\n" + toString());

		if(iterIsHdrIterator) {
			if(iterBufPos == 0) {
				dataIter.setHdrMinVal(true);
				dataIter.setHdrMaxVal(false);
				dataIter.setArcTime(getContHdrParamAsLong(ContainerHdr.CHP_ARC_TIME_MIN));
				dataIter.setDataTime(getContHdrParamAsLong(ContainerHdr.CHP_DATA_TIME_MIN));
				dataIter.setDataIndex(getContHdrParamAsLong(ContainerHdr.CHP_DATA_IDX_MIN));
				dataIter.setCounter(dataIter.getCounter() + 1);
				iterBufPos = 1;
			}
			else if(iterBufPos == 1) {
				dataIter.setHdrMinVal(false);
				dataIter.setHdrMaxVal(true);
				dataIter.setArcTime(getContHdrParamAsLong(ContainerHdr.CHP_ARC_TIME_MAX));
				dataIter.setDataTime(getContHdrParamAsLong(ContainerHdr.CHP_DATA_TIME_MAX));
				dataIter.setDataIndex(getContHdrParamAsLong(ContainerHdr.CHP_DATA_IDX_MAX));
				dataIter.setCounter(dataIter.getCounter() + 1);
				iterBufPos = 2;
			}
			else {
				iterEOC = true;
				return false;
			}
		}
		else {
			if(iterBufPos + DATAHDR_LEN > iterBytesRead) {
				// Kein Platz für Header: Um den teilgelesenen Header zurueck, neu laden
				iterLoadDataChunk(iterBytesRead - iterBufPos);
				if(iterBufPos + DATAHDR_LEN > iterBytesRead) {
					// immer noch? -> Datei zuende
					iterEOC = true;
					return false;
				}
			}
			iterNextDataLen = ByteIO.readSignedInt4Bytes(readBuf, iterBufPos);
			iterBufPos += DATALEN_LEN;

			if(iterBufPos + iterNextDataLen > iterBytesRead) {
				// Platz für kompletten DS? Wenn nicht, Puffer zuruecksetzen und dadurch voll ausnutzen
				iterLoadDataChunk(iterBytesRead - iterBufPos);
			}

			dataIter.setArcTime(ByteIO.readUnsignedLong6Byte(readBuf, iterBufPos));
			dataIter.setDataTime(ByteIO.readUnsignedLong6Byte(readBuf, iterBufPos + ByteIO.LONG6B_LEN));
			dataIter.setDataIndex(ByteIO.readSignedLong8Byte(readBuf, iterBufPos + 2 * ByteIO.LONG6B_LEN));
			int checkSum = ByteIO.readSignedInt4Bytes(readBuf, iterBufPos + 2 * ByteIO.LONG6B_LEN + ByteIO.LONG8B_LEN);
			iterBufPos += DATAHDR_LEN;

			int calcCheckSum = (int)(iterNextDataLen + dataIter.getArcTime() + dataIter.getDataTime() + dataIter.getDataIndex());
			if(checkSum != calcCheckSum) {
				throw new PersistenceException(
						"Prüfsumme des Datensatzes an Pos. " + (dataIter.getCounter() + 1) + " fehlerhaft: " + checkSum + " != dx=" + dataIter.getDataIndex()
						+ "+dt=" + dataIter.getDataTime() + "+at=" + dataIter.getArcTime() + "=" + calcCheckSum + "\n" + toString()
				);
			}

			dataIter.setHdrMinVal(false);
			dataIter.setHdrMaxVal(false);
			dataIter.setDataSize(iterNextDataLen - COMPRESS_LEN - DATAHDR_LEN - ByteIO.SEPARATOR.length);
			dataIter.setDataUncSize(ByteIO.readSignedInt4Bytes(readBuf, iterBufPos));
			dataIter.setNodata(Util.cmpBytes(readBuf, iterBufPos + ByteIO.INT4B_LEN, NO_DATA));
			dataIter.setNoSource(Util.cmpBytes(readBuf, iterBufPos + ByteIO.INT4B_LEN, NO_SOURCE));
			dataIter.setNoRights(Util.cmpBytes(readBuf, iterBufPos + ByteIO.INT4B_LEN, NO_RIGHTS));
			dataIter.setPotDataGap(Util.cmpBytes(readBuf, iterBufPos + ByteIO.INT4B_LEN, POT_GAP));
			dataIter.setUseExtraBuffer(iterBufPos + (iterNextDataLen - DATAHDR_LEN) > iterBytesRead);
		}

		dataIter.setCounter(dataIter.getCounter() + 1);
		iterDataRead = false;
		return true;
	}

	/**
	 * Springt stepBack Bytes in der Datei zurueck und liest Datensätze aus in den Lese-Puffer bis dieser voll ist oder das Ende der Datei erreicht ist. Der
	 * letzte Datensatz wird ggf. dabei nur teilweise gelesen. Dies muss die aufrufende Methode erkennen und bei naechsten Aufruf den Parameter stepBack
	 * entsprechend setzen.
	 *
	 * @param stepBack Zahl der Bytes, um die von der momentanen Position des Dateizeigers zurueckgesprungen wird.
	 *
	 * @return ContainerDataIterator zum Durchlaufen der gefundenen Datensätze
	 *
	 * @throws PersistenceException
	 */
	private void iterLoadDataChunk(long stepBack) throws PersistenceException {
		try {
			iterRaf.seek(iterRaf.getFilePointer() - stepBack);
			iterBytesRead = iterRaf.read(readBuf);
			iterBufPos = 0;
		}
		catch(IOException e) {
			iterClose();
			throw new PersistenceException("Iterator: Fehler beim Zugriff auf die Containerdatei: " + e.getMessage() + "\n" + this);
		}
	}

	/**
	 * Laedt den Datensatz in einen Extra-Puffer, weil er zu gross für den Standard-Buffer ist.
	 *
	 * @param stepBack    Zahl der Bytes, um die von der momentanen Position des Dateizeigers zurueckgesprungen wird.
	 * @param extraBuffer Extra-Puffer
	 *
	 * @throws PersistenceException
	 */
	private void iterLoadExtraBuffer(long stepBack, byte[] extraBuffer) throws PersistenceException {
		try {
			long oldPos = iterRaf.getFilePointer();
			iterRaf.seek(iterRaf.getFilePointer() - (iterBytesRead - iterBufPos));
			iterRaf.readFully(extraBuffer);
			iterRaf.seek(oldPos);
		}
		catch(IOException e) {
			iterClose();
			throw new PersistenceException("Iterator: Fehler beim Zugriff auf die Containerdatei: " + e.getMessage() + "\n" + this);
		}
	}

	/**
	 * Ueberspringt den naechsten Datensatz
	 *
	 * @throws PersistenceException
	 */
	void iterSkipData() throws PersistenceException {
		checkContainerAccessed();
		// Spezialbehandlung für ausgelagerte Container
		if(deletedSpecialTreatment && deletedHeaderFlag) {
			iterlistOfDataIdxPos++;
		}
		// Keine Spezialbehandlung für ausgelagerte Container
		else {
			if(iterEOC) throw new PersistenceException("Ende des Containers erreicht: " + toString());
			if(iterDataRead) throw new PersistenceException("Datensatz wurde bereits gelesen oder übersprungen: " + toString());
			iterDataRead = true;
			if(!iterIsHdrIterator) iterBufPos += iterNextDataLen - DATAHDR_LEN;
		}
	}

	/**
	 * @return Ob der Datensatz auch nach {@link #iterGetRawData()} nocj komprimiert ist. Wenn {@link #DECOMPRESS_QUERY_RESULTS} gesetzt ist, werden aus
	 *         Performancegruenden die Ergebnisdaten bereits im Archivsystem dekomprimiert (das Stream-Interface der Kernsoftware ist etwas langsam).
	 */
	boolean iterStillCompressed() {
		return iterUncompressedLen != NOT_COMPRESSED;
	}

	/**
	 * Liefert den naechsten Datensatz so wie er im Container gespeichert wurde.
	 *
	 * @throws PersistenceException
	 */
	byte[] iterGetRawData(Inflater decompresser) throws PersistenceException {
		checkContainerAccessed();

		if(iterIsHdrIterator) return null;

		if(iterDataRead) throw new PersistenceException("Datensatz wurde bereits gelesen oder übersprungen: " + toString());
		iterDataRead = true;

		iterUncompressedLen = ByteIO.readSignedInt4Bytes(readBuf, iterBufPos);
		iterBufPos += COMPRESS_LEN;
		int netDataLen = iterNextDataLen - COMPRESS_LEN - DATAHDR_LEN;

		if(!dataIter.isData()) {
			iterBufPos += netDataLen;
			return null;
		}
		else {
			byte[] srcBuffer;
			int srcBufferPos;
			if(dataIter.mustUseExtraBuffer()) {
				srcBuffer = new byte[netDataLen];
				iterLoadExtraBuffer(iterBytesRead - iterBufPos, srcBuffer);
				srcBufferPos = 0;
			}
			else {
				srcBuffer = readBuf;
				srcBufferPos = iterBufPos;
			}

			if(DECOMPRESS_QUERY_RESULTS && iterUncompressedLen != 0) {
				byte[] tmpBuf = new byte[iterUncompressedLen];
				decompresser.reset();
				decompresser.setInput(srcBuffer, srcBufferPos, netDataLen);
				try {
					int realUncomprLen = decompresser.inflate(tmpBuf);
					if(iterUncompressedLen != realUncomprLen) {
						throw new PersistenceException(
								"Unkomprimierte Laenge des Datensatzes fehlerhaft: " + realUncomprLen + " ermittelt, " + iterUncompressedLen + " erwartet\n"
								+ this
						);
					}
				}
				catch(DataFormatException e) {
					throw new PersistenceException("Datensatz konnte nicht dekomprimiert werden: " + e.getMessage() + "\n" + this);
				}
				iterBufPos += netDataLen;
				iterUncompressedLen = 0;
				return tmpBuf;
			}
			else {
				byte[] buf = new byte[netDataLen];
				System.arraycopy(srcBuffer, srcBufferPos, buf, 0, netDataLen);	// schnellste Methode so
				iterBufPos += netDataLen;
				return buf;
			}
		}
	}

	/**
	 * Schliesst das RandomAccessFile, das zum Iterieren verwendet wurde.
	 *
	 * @throws PersistenceException
	 */
	void iterClose() throws PersistenceException {
		try {
			closeRandomAccessFile(iterRaf);
		}
		catch(PersistenceException e) {
			throw e;
		}
		finally {
			iterRaf = null;
		}
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Private Methoden
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/** Prueft, ob ein Container zum Lesen im Zugriff ist. Wirft eine {@link PersistenceException} falls nicht. */
	private void checkContainerAccessed() throws PersistenceException {
		if(!accessed) throw new PersistenceException("Container zuerst mit accessContainer() spezifizieren");
	}

	/** Prueft, ob ein Container zum Lesen und Schreiben im Zugriff ist. Wirft eine {@link PersistenceException} falls nicht. */
	private void checkContainerAccessedRW() throws PersistenceException {
		checkContainerAccessed();
		if(readOnly) throw new PersistenceException("Container nur zum Lesen adressiert\n" + toString());
	}

	/** Prueft, ob ein Container zum Lesen im Zugriff ist und dessen Container-Header bereits eingelesen ist. Wirft eine {@link PersistenceException} falls nicht. */
	private void checkContainerAccessedHeaderRead() throws PersistenceException {
		checkContainerAccessed();
		if(!headerRead) throw new PersistenceException("Container-Header weder eingelesen noch neu erzeugt\n" + toString());
	}

	/**
	 * Liefert den uebergebenen Puffer falls dessen Größe ausreichend ist, andernfalls einen neu angelegten. Zweck: sowenig Arbeit für die Speicherverwaltung wie
	 * moeglich.
	 *
	 * @param defaultBuffer Vorhandener Puffer
	 * @param desiredSize   Gewuenschte PufferGröße
	 *
	 * @return Byte-Array-Puffer
	 */
	private static byte[] getBuf(byte[] defaultBuffer, int desiredSize) {
		return (desiredSize <= defaultBuffer.length) ? defaultBuffer : new byte[desiredSize];
	}

	/**
	 * Schliesst ein RandomAccessFile und faengt die Exception ab. (Nur zur Abkuerzung verwendet).
	 *
	 * @param raf Zu schliessendes RandomAccessFile
	 *
	 * @throws PersistenceException
	 */
	private void closeRandomAccessFile(FileAccess raf) throws PersistenceException {
		if(raf != null) {
			try {
				raf.close();
			}
			catch(IOException e) {
				throw new PersistenceException("Container-Datei konnte nicht geschlossen werden: " + e.getMessage() + "\n" + this);
			}
		}
	}

	long getContainerId() {
		return containerId;
	}

	File getContFile() {
		return contFile;
	}

	long getObjId() {
		return objId;
	}

	long getAtgId() {
		return atgId;
	}

	long getAspId() {
		return aspId;
	}

	int getSv() {
		return sv;
	}

	private class MyCloseableRandomAccessFile extends CloseableRandomAccessFile {
		public MyCloseableRandomAccessFile() {
			super(ContainerFile.this.contFile);
		}

		@Override
		protected void openFile() throws IOException {
			// Offene Dateien zählen
			_numOpenFiles.incrementAndGet();
			_openFilesInCurrentThread.get().add(this);
			super.openFile();
		}

		@Override
		public void closeFile() throws IOException {
			closeFile(true);
		}

		public void closeFile(final boolean removeFromQueue) throws IOException {
			// Datei nicht mehr zählen, da geschlossen
			_numOpenFiles.decrementAndGet();
			if(removeFromQueue) {
				_openFilesInCurrentThread.get().remove(this);
			}
			super.closeFile();
		}
	}
}
