/*
 *
 * Copyright 2017 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.config.SystemObject;

/**
 * Klasse, die die IDs der Objekte einer Datenidentifikation aufnehmen kann. Objekte der Klasse werden als Schlüssel für das HashSet mit nicht zu cachenden
 * Datenidentifikationen verwendet.
 */
public class IdDataIdentification {

	/** Id des Systemobjekts dieser Datenidentifikation */
	private final long _objectId;

	/** Id der Attributgruppe dieser Datenidentifikation */
	private final long _atgId;

	/** Id des Aspekts dieser Datenidentifikation */
	private final long _aspectId;

	/** Simulationsvariante dieser Datenidentifikation */
	private final int _simVariant;

	/**
	 * Erzeugt ein neues Objekt mit den angegebenen Werten
	 *
	 * @param objectId          Id des Systemobjekts dieser Datenidentifikation
	 * @param atgId             Id der Attributgruppe dieser Datenidentifikation
	 * @param aspectId          Id des Aspekts dieser Datenidentifikation
	 * @param simulationVariant Simulationsvariante dieser Datenidentifikation
	 */
	public IdDataIdentification(final SystemObject object, DataDescription dataDescription) {
		_objectId = object.getId();
		_atgId = dataDescription.getAttributeGroup().getId();
		_aspectId = dataDescription.getAspect().getId();
		_simVariant = dataDescription.getSimulationVariant();
	}

	/**
	 * Erzeugt ein neues Objekt mit den angegebenen Werten
	 *
	 * @param objectId          Id des Systemobjekts dieser Datenidentifikation
	 * @param atgId             Id der Attributgruppe dieser Datenidentifikation
	 * @param aspectId          Id des Aspekts dieser Datenidentifikation
	 * @param simulationVariant Simulationsvariante dieser Datenidentifikation
	 */
	public IdDataIdentification(final long objectId, final long atgId, final long aspectId, final int simulationVariant) {
		_objectId = objectId;
		_atgId = atgId;
		_aspectId = aspectId;
		_simVariant = simulationVariant;
	}

	public IdDataIdentification(final ArchiveDataSpecification ads) {
		_objectId = ads.getObject().getId();
		_atgId = ads.getDataDescription().getAttributeGroup().getId();
		_aspectId = ads.getDataDescription().getAspect().getId();
		_simVariant = ads.getDataDescription().getSimulationVariant();
	}

	public boolean equals(Object otherObject) {
		if(this == otherObject) return true;
		if(otherObject instanceof IdDataIdentification) {
			IdDataIdentification other = (IdDataIdentification)otherObject;
			if(_objectId == other._objectId && _atgId == other._atgId && _aspectId == other._aspectId && _simVariant == other._simVariant) return true;
		}
		return false;
	}

	public int hashCode() {
		return (int)(_objectId ^ (_objectId >>> 32) ^ (_atgId << 8) ^ (_atgId >>> 24) ^ (_atgId >>> 56) ^ (_aspectId << 16) ^ (_aspectId >>> 16) ^ (
				_aspectId >>> 48) | (((long)_simVariant) << 22));
	}

	public long getObjectId() {
		return _objectId;
	}

	public long getAtgId() {
		return _atgId;
	}

	public long getAspectId() {
		return _aspectId;
	}

	public int getSimVariant() {
		return _simVariant;
	}

	public String toString() {
		return "IdDataIdentification{obj: " + _objectId + " atg: " + _atgId + " asp: " + _aspectId + " sv: " + _simVariant;
	}
}
