/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import com.google.common.collect.ImmutableSet;
import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DIdNodeNotFoundException;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTree;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTreeIterator;
import de.bsvrz.ars.ars.mgmt.tasks.ArchiveTask;
import de.bsvrz.ars.ars.mgmt.tasks.BackupTask;
import de.bsvrz.ars.ars.mgmt.tasks.RestoreTask;
import de.bsvrz.ars.ars.mgmt.tasks.Task;
import de.bsvrz.ars.ars.persistence.index.*;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.Deflater;

/**
 * Zentrale Persistenz-Verwaltung für die Archivierung von Daten auf einem Speichermedium vom Typ A (Festplatte) und die entsprechende Meta-Daten-Verwaltung.
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schaefer
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class PersistenceManager {

	private Debug logger = Debug.getLogger();

	private static final String OBJ_ID_PREFIX = "obj";

	private static final String ATG_ID_PREFIX = "atg";

	private static final String ASP_ID_PREFIX = "asp";

	private static final String SV_PREFIX = "sv";

	/** Verzeichnisnamen der Datensatzarten. Die ersten Indexe entsprechen {@link Util#getDataKindIndex(ArchiveDataKind)}. */
	public static final String[] DATAKIND_SUFFIX = new String[]{"oa", "on", "na", "nn"};

	private static final Set<String> VALID_SUFFIXES = ImmutableSet.copyOf(DATAKIND_SUFFIX);

	private static final Pattern OBJ_DIRNAME_PAT = Pattern.compile("^" + OBJ_ID_PREFIX + "(\\d+)$");

	private static final Pattern ATG_DIRNAME_PAT = Pattern.compile("^" + ATG_ID_PREFIX + "(\\d+)$");

	private static final Pattern ASP_DIRNAME_PAT = Pattern.compile("^" + ASP_ID_PREFIX + "(\\d+)$");

	private static final Pattern SV_DIRNAME_PAT = Pattern.compile("^" + SV_PREFIX + "(\\d+)$");

	private static final Pattern NUM_DIR_PAT = Pattern.compile("^\\w+?(\\d+)$");

	/**
	 * Dateiname des Lock-Files des Archivsystems. Die Datei wird angelegt, wenn das Archivsystem gestartet wird. Die wird gelöscht, wenn das Archivsystem
	 * heruntergefahren wird.
	 */
	public static final String ACTIVITY_FLAG_FILE_NAME = "_isActive.flag";

	/**
	 * Eine Datei dieses Namens wird pro Verzeichnis einer Datenidentifikation/Datensatzart angelegt, wenn ein neuer Datensatz archiviert wurde und der Index
	 * moeglicherweise (wg. Caching) noch nicht auf die Platte durchgeschrieben wurde. Beim Herunterfahren des Systems werden alle Indexe durchgeschrieben und die
	 * Flag-Dateien geloescht. Existiert diese Datei beim Systemstart noch, ist dies ein Hinweis auf eine unkorrekte Beendigung des Archivsystems. Für jede DId,
	 * für die die Datei existiert, werden die Indexe neu aufgebaut.
	 */
	public static String REBUILDINDEX_FILE_FLAG_NAME = "_rebuildIndex.flag";

	/** Zum Wiederaufbau der Indexe werden temporaer Indexe dieser Größe angelegt */
	private static int REBUILDINDEX_IDX_SIZE = 2048;

	/** Zaehlt die besuchten Verzeichnisse beim Pruefen des Persistenzverzeichnisses */
	private static int directoryCount = 0;


	/** Enthaelt die zuletzt verwendete ContainerID */
	private AtomicLong nextContainerID = new AtomicLong(1);

	private ArchiveManager archivMgr;

	private StartupProperties startupProps;

	private String archivePath;

	/** Die Datei zeigt an, ob das Persistenzverzeichnis von einem Archivsystem gerade benutzt wird. */
	private File isActive;

	/**
	 * Jede Task richtet ihre Persistanz-Anfragen an den Persistenz-Manager. In dieser Map haelt der Persistenz-Manager pro Task eine weitere Map mit
	 * ContainerFile-Objekten. Alle Anfragen bzgl. Container-Dateien leitet der PersistenceManager an ein ContainerFile-Objekt weiter, das in der Map des Tasks
	 * enthalten ist. Gibt der Task einen Index an, so wird das ContainerFile-Objekt mit diesem Map-Schluessel benutzt, ansonsten das Objekt mit dem Map-Schluessel
	 * 0.
	 */
	private Map<Task, Map<Long, ContainerFile>> containerFileMap = new HashMap<Task, Map<Long, ContainerFile>>();

	private boolean _preventWriteStartupInfo = false;

	/**
	 * Erzeugt den Persistenz-Manager.
	 *
	 * @param archMgr  Archiv-Manager
	 * @param archPath Archivierungs-Verzeichnis
	 */
	public PersistenceManager(ArchiveManager archMgr, String archPath) {
		archivMgr = archMgr;
		setPersistenceDir(archPath);
		startupProps = new StartupProperties(archivePath);
	}

	public void setPersistenceDir(String archPath) {
		archivePath = archPath;
		File archiveDir = new File(archivePath);
		boolean ok = archiveDir.exists();
		if(!ok) {
			logger.info("Persistenzverzeichnis '" + archPath + "' existiert nicht, wird neu angelegt.");
			ok = archiveDir.mkdir();
		}
		if(!ok) archivMgr.quitError("Persistenzverzeichnis '" + archPath + "' existiert nicht und konnte nicht angelegt werden.");
	}

	/**
	 * Liefert Kennzeichen, ob die angegebene Datei eine Container-Datei ist.
	 *
	 * @param file zu pruefende Datei
	 *
	 * @return Kennzeichen
	 */
	public static boolean isContainerFile(File file) {
		return ContainerFile.isContainerFile(file);
	}

	/**
	 * @param name Name einer Datei
	 *
	 * @return Wahr, wenn es der Name einer Indexdatei ist
	 */
	public static boolean isIndexFile(String name) {
		return name.equals(DataTimeIndex.IDX_FILENAME) || name.equals(ATimeMonotoneIndex.IDX_FILENAME) || name.equals(DIdxATimeMonotoneIndex.IDX_FILENAME)
		       || name.equals(DataIndexIndex.IDX_FILENAME);
	}

	/**
	 * @param name Name einer Datei
	 *
	 * @return Wahr, wenn es der Name einer _rebuildIndex.flag-Datei ist
	 */
	public static boolean isRebuildIndexFlagFile(String name) {
		return name.equals(REBUILDINDEX_FILE_FLAG_NAME);
	}

	/**
	 * @param name Name einer Datei
	 *
	 * @return Wahr, wenn es der Name einer _isActive.flag-Datei ist
	 */
	public static boolean isActivityFlagFile(String name) {
		return name.equals(ACTIVITY_FLAG_FILE_NAME);
	}

	/**
	 * @param name Name einer Datei
	 *
	 * @return Wahr, wenn es ein zulaessiger Dateiname des Archivsystems ist
	 */
	public static boolean isValidStorageArSFile(String name) {
		return isValidIDFolderName(name) || ContainerFile.isContainerFileName(name) || isIndexFile(name) || isRebuildIndexFlagFile(name);
	}

	/**
	 * @param name Name einer Datei
	 *
	 * @return Wahr, wenn es ein zulaessiger Dateiname im obersten Persistenzverzeichnis des Archivsystems ist
	 */
	public static boolean isValidTopLevelArSFile(String name) {
		return isValidIDFolderName(name) || RestoreTask.isUnavailMediumIDsFile(name) || StartupProperties.isStartupInfoFile(name) || isActivityFlagFile(name);
	}


	/**
	 * Liefert die Container-ID zur angegebenen Container-Datei, indem die im Container-Dateinamen enthaltene Container-ID ausgelesen wird. Alternativ kann die
	 * Container-ID aus den Container-Header-Parametern gelesen werden.
	 *
	 * @param file Container-Datei
	 *
	 * @return Container-ID
	 */
	public static long getContID(File file) {
		return ContainerFile.getContID(file);
	}

	/**
	 * Liefert die Container-ID zum Namen der angegebenen Container-Datei, indem die im Container-Dateinamen enthaltene Container-ID ausgelesen wird. Alternativ
	 * kann die Container-ID aus den Container-Header-Parametern gelesen werden.
	 *
	 * @param fileName Container-Datei
	 *
	 * @return Container-ID
	 */
	public static long getContID(String fileName) {
		return ContainerFile.getContID(fileName);
	}

	/**
	 * Liefert den Dateinamen des Daten-Containers mit der angegebenen Container-ID.
	 *
	 * @param containerId Container-ID
	 *
	 * @return Dateiname
	 */
	public static String getContFileName(long containerId) {
		return ContainerFile.getContainerFileName(containerId);
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Container: Verzeichnis
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation. Der Pfad wird gebildet indem die IDs und die SV in Dreiergruppen zerlegt und mit einem
	 * Praefix versehen werden. Pro Dreiergruppe und pro SimVarwird ein Verzeichnis verwendet.
	 *
	 * @param objId Objekt-ID
	 * @param atgId Attributgruppen-ID
	 * @param aspId Aspekt-ID
	 * @param sv    Simulationsvariante
	 *
	 * @return Archiv-Pfad
	 */
	public String getPathName(long objId, long atgId, long aspId, int sv) {
		return getPathName(archivePath, objId, atgId, aspId, sv);
	}


	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation und den Wurzelpfad des Archivsystems. Der Pfad wird gebildet indem die IDs und die SV in
	 * Dreiergruppen zerlegt und mit einem Praefix versehen werden. Pro Dreiergruppe und pro SimVarwird ein Verzeichnis verwendet.
	 *
	 * @param basePath Wurzelpfad des Archivsystems
	 * @param objId    Objekt-ID
	 * @param atgId    Attributgruppen-ID
	 * @param aspId    Aspekt-ID
	 * @param sv       Simulationsvariante
	 *
	 * @return Archiv-Pfad
	 */
	public static String getPathName(String basePath, long objId, long atgId, long aspId, int sv) {
		// (48 + 48 + 48 + 6) ist die maximale String-Länge zur Darstellung von Objekt- Atg- und Aspekt-ID und Simulationsvariante
		StringBuilder pathName = new StringBuilder(basePath.length() + (48 + 48 + 48 + 6));
		appendPathName(pathName, basePath, objId, atgId, aspId, sv);
		return pathName.toString();
	}

	/**
	 * Ergänzt das übergebene StringBuilder-Objekt um den Archiv-Pfad für die angegebene Datenidentifikation und den Wurzelpfad des Archivsystems.
	 * Der Pfad wird gebildet indem die IDs und die SV in Dreiergruppen zerlegt und mit einem Praefix versehen werden.
	 * Pro Dreiergruppe und pro SimVar wird ein Verzeichnis verwendet.
	 *
	 * @param pathName StringBuilder-Objekt, an das der ermittelte Pfad ergänzt werden soll.
	 * @param basePath Wurzelpfad des Archivsystems
	 * @param objId    Objekt-ID
	 * @param atgId    Attributgruppen-ID
	 * @param aspId    Aspekt-ID
	 * @param sv       Simulationsvariante
	 */
	public static void appendPathName(
			final StringBuilder pathName, final String basePath, final long objId, final long atgId, final long aspId, final int sv) {
		pathName.append(basePath);
		Util.appendIdElements(pathName, File.separator + OBJ_ID_PREFIX, objId);
		Util.appendIdElements(pathName, File.separator + ATG_ID_PREFIX, atgId);
		Util.appendIdElements(pathName, File.separator + ASP_ID_PREFIX, aspId);
		pathName.append(File.separator).append(SV_PREFIX).append(sv);
	}

	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation und Datensatzart. Der Pfad wird gebildet indem die IDs und die SV in Dreiergruppen zerlegt
	 * und mit einem Praefix versehen werden. Pro Dreiergruppe, pro SimVar und pro Datensatzart wird ein Verzeichnis verwendet.
	 *
	 * @param objId    Objekt-ID
	 * @param atgId    Attributgruppen-ID
	 * @param aspId    Aspekt-ID
	 * @param sv       Simulationsvariante
	 * @param dataKind Datensatzart
	 *
	 * @return Archiv-Pfad
	 */
	public String getPathName(long objId, long atgId, long aspId, int sv, ArchiveDataKind dataKind) {
		return getPathName(objId, atgId, aspId, sv, Util.getDataKindIndex(dataKind));
	}

	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation und Datensatzart. Der Pfad wird gebildet indem die IDs und die SV in Dreiergruppen zerlegt
	 * und mit einem Praefix versehen werden. Pro Dreiergruppe, pro SimVar und pro Datensatzart wird ein Verzeichnis verwendet.
	 *
	 * @param objId    Objekt-ID
	 * @param atgId    Attributgruppen-ID
	 * @param aspId    Aspekt-ID
	 * @param sv       Simulationsvariante
	 * @param dataKind Datensatzart, 0 für OA, 1 für ON, 2 für NA und 3 für NN
	 *
	 * @return Archiv-Pfad
	 */
	public String getPathName(long objId, long atgId, long aspId, int sv, int dataKind) {
		// (48 + 48 + 48 + 6 + 3) ist die maximale String-Länge zur Darstellung von Objekt- Atg- und Aspekt-ID, Simulationsvariante und Datenart
		StringBuilder pathName = new StringBuilder(archivePath.length() + (48 + 48 + 48 + 6 + 3));
		appendPathName(pathName, archivePath, objId, atgId, aspId, sv);
		pathName.append(File.separator).append(DATAKIND_SUFFIX[dataKind]);
		return pathName.toString();
	}

	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation und Datensatzart. Der Pfad wird gebildet indem die IDs und die SV in Dreiergruppen zerlegt
	 * und mit einem Praefix versehen werden. Pro Dreiergruppe, pro SimVar und pro Datensatzart wird ein Verzeichnis verwendet.
	 *
	 * @param ads Spezifikation des Archivdatensatzes (enthaelt objID, atgID, aspID, sv)
	 * @param archiveDataKind Datensatzart
	 *
	 * @return Name des Archiv-Pfads
	 */
	public String getPathName(ArchiveDataSpecification ads, ArchiveDataKind archiveDataKind) {
		DataDescription dd = ads.getDataDescription();
		return getPathName(ads.getObject().getId(), dd.getAttributeGroup().getId(), dd.getAspect().getId(), dd.getSimulationVariant(), archiveDataKind);
	}

	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation und Datensatzart. Der Pfad wird gebildet indem die IDs und die SV in Dreiergruppen zerlegt
	 * und mit einem Praefix versehen werden. Pro Dreiergruppe, pro SimVar und pro Datensatzart wird ein Verzeichnis verwendet.
	 *
	 * @param dataIdentification objID, atgID, aspID, sv
	 * @param archiveDataKind Datensatzart
	 *
	 * @return Name des Archiv-Pfads
	 */
	public String getPathName(IdDataIdentification dataIdentification, ArchiveDataKind archiveDataKind) {
		return getPathName(dataIdentification.getObjectId(), dataIdentification.getAtgId(), dataIdentification.getAspectId(), dataIdentification.getSimVariant(), archiveDataKind);
	}

	public static ArchiveDataKind getArchiveDataKindFromDir(File dir) {
		for(int i = 0; i < DATAKIND_SUFFIX.length; i++) {
			if(dir.getName().equals(DATAKIND_SUFFIX[i])) return Util.getDataKindFromIndex(i);
		}
		return null;
	}

	/**
	 * Liefert von allen Containerdateien im angegebenen Verzeichnis die hoechste Container-ID.
	 *
	 * @param path Zu durchsuchendes Verzeichnis
	 *
	 * @return Die hoechste vorhandene Container-Id, -1 falls das Container-Verzeichnis nicht vorhanden ist oder keine Container darin vorhanden sind.
	 */
	public long getLargestContainerID(String path) {
		String[] fNames = new File(path).list(ContainerFile.CONT_FILENAME_FILTER);
		if(fNames == null) return -1;
		if(fNames.length <= 0) {
			return -1;
		}
		else {
			return ContainerFile.getContID(Util.getLastInAlphabet(fNames));
		}
	}

	/**
	 * Liefert die ContainerIDs von allen Containerdateien im angegebenen Verzeichnis, aufsteigend sortiert.
	 *
	 * @param path Zu durchsuchendes Verzeichnis
	 *
	 * @return Ein aufsteigend sortiertes Array mit allen ContainerIDs im angegebenen Verzeichnis
	 */
	public static long[] getSortedContainerIDs(String path) {
		String[] fNames = new File(path).list(ContainerFile.CONT_FILENAME_FILTER);
		if(fNames == null) return new long[]{};
		final long[] result = new long[fNames.length];
		for(int i = 0; i < fNames.length; i++) {
			result[i] = ContainerFile.getContID(fNames[i]);
		}
		Arrays.sort(result);
		return result;
	}

	/**
	 * Diese Methode loescht das komplette Verzeichnis eines {@link DataIdentNode} mit allen Unterverzeichnissen der Datensatzarten und aller darin befindlichen
	 * Datencontainern; damit werden alle archivierten Datensätze einer Simulationsvariante geloescht. Der {@link DataIdentNode} muss in der Verwaltung aus dem
	 * {@link DataIdentTree}geloescht werden. Falls die Simulationsvariante gleich Null ist, oder das Verzeichnis nicht geloescht werden konnte, wird eine {@link
	 * PersistenceException} geworfen.
	 *
	 * @param objId
	 * @param atgId
	 * @param aspId
	 * @param simVar
	 *
	 * @throws PersistenceException
	 */
	public void deleteSimVar(long objId, long atgId, long aspId, int simVar) throws PersistenceException {

		if(simVar == 0) throw new PersistenceException("Simulations-Variante '0' (Echtdaten) kann nicht geloescht werden!");

		String path = getPathName(objId, atgId, aspId, simVar);
		File dir = new File(path);
		if(dir.exists() && !Util.deleteDir(dir)) throw new PersistenceException("Simulations-Varianten-Verzeichnis konnte nicht geloescht werden: " + path);
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Container: Access/Create/Close
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Assoziiert den angegebenen Task mit dem spezifizierten Container, so dass alle anschliessend von diesem Task aufgerufenen Methoden auf diesen Container
	 * wirken. Soll der Zugriff auf diesen Container beendet werden (um z.B. einen anderen Container zu adressieren), muss dazu die Methode {@link
	 * #leaveContainer(Task)} aufgerufen werden. Falls das Verzeichnis der Datenidentifikation nicht existiert, wird es angelegt.
	 *
	 * @param task        Anfragender Task
	 * @param didPath     Pfad bis zur Datensatzart
	 * @param containerId Container-ID
	 *
	 * @throws PersistenceException
	 */
	public void accessContainer(Task task, String didPath, long containerId) throws PersistenceException {
		File dir = new File(didPath);
		long ids[] = extractDIdFromPath(dir);
		accessContainer(task, ids[0], ids[1], ids[2], (int)ids[3], getArchiveDataKindFromDir(dir), containerId);
	}

	/**
	 * Assoziiert den angegebenen Task mit dem spezifizierten Container, so dass alle anschliessend von diesem Task aufgerufenen Methoden auf diesen Container
	 * wirken. Soll der Zugriff auf diesen Container beendet werden (um z.B. einen anderen Container zu adressieren), muss dazu die Methode {@link
	 * #leaveContainer(Task)} aufgerufen werden. Falls das Verzeichnis der Datenidentifikation nicht existiert, wird es angelegt.
	 *
	 * @param task        Anfragender Task
	 * @param objId       Objekt-ID
	 * @param atgId       Attributgruppen-ID
	 * @param aspId       Aspekt-ID
	 * @param sv          Simulationsvariante
	 * @param dataKind    Datensatzart
	 * @param containerId Container-ID
	 *
	 * @throws PersistenceException
	 */
	public void accessContainer(Task task, long objId, long atgId, long aspId, int sv, ArchiveDataKind dataKind, long containerId) throws PersistenceException {
		accessContainer(task, 0, objId, atgId, aspId, sv, dataKind, containerId);
	}

	/**
	 * Assoziiert den angegebenen Task mit dem spezifizierten Container, so dass alle anschliessend von diesem Task aufgerufenen Methoden auf diesen Container
	 * wirken. Soll der Zugriff auf diesen Container beendet werden (um z.B. einen anderen Container zu adressieren), muss dazu die Methode {@link
	 * #leaveContainer(Task)} aufgerufen werden. Falls das Verzeichnis der Datenidentifikation nicht existiert, wird es angelegt.
	 *
	 * @param task        Anfragender Task
	 * @param objId       Objekt-ID
	 * @param atgId       Attributgruppen-ID
	 * @param aspId       Aspekt-ID
	 * @param sv          Simulationsvariante
	 * @param dataKind    Datensatzart
	 * @param pathName        Pfadname des Verzeichnisses der Containerdatei
	 * @param containerId Container-ID
	 *
	 * @throws PersistenceException
	 */
	public void accessContainer(Task task, long objId, long atgId, long aspId, int sv, ArchiveDataKind dataKind, String pathName, boolean createDir, long containerId) throws PersistenceException {
		accessContainer(task, 0, objId, atgId, aspId, sv, dataKind, pathName, createDir, containerId);
	}

	/**
	 * Assoziiert den angegebenen Task mit dem spezifizierten Container, so dass alle anschliessend von diesem Task aufgerufenen Methoden auf diesen Container
	 * wirken. Soll der Zugriff auf diesen Container beendet werden (um z.B. einen anderen Container zu adressieren), muss dazu die Methode {@link
	 * #leaveContainer(Task)} aufgerufen werden. Damit ein Task mehrere Container gleichzeitig parallel offen halten kann, wird der cfIdx zur Unterscheidung der
	 * ContainerFile-Objekte benutzt. Falls das Verzeichnis der Datenidentifikation nicht existiert, wird es angelegt.
	 *
	 * @param task        Anfragender Task
	 * @param cfIdx       ContainerFile-Index
	 * @param objId       Objekt-ID
	 * @param atgId       Attributgruppen-ID
	 * @param aspId       Aspekt-ID
	 * @param sv          Simulationsvariante
	 * @param dataKind    Datensatzart
	 * @param containerId Container-ID
	 *
	 * @throws PersistenceException
	 */
	public void accessContainer(Task task, int cfIdx, long objId, long atgId, long aspId, int sv, ArchiveDataKind dataKind, long containerId)
			throws PersistenceException {
		File contDir = getCreatePersDir(objId, atgId, aspId, sv, dataKind);
		getContainerFile(task, cfIdx).accessContainer(objId, atgId, aspId, sv, dataKind, containerId, contDir);
	}

	/**
	 * Assoziiert den angegebenen Task mit dem spezifizierten Container, so dass alle anschliessend von diesem Task aufgerufenen Methoden auf diesen Container
	 * wirken. Soll der Zugriff auf diesen Container beendet werden (um z.B. einen anderen Container zu adressieren), muss dazu die Methode {@link
	 * #leaveContainer(Task)} aufgerufen werden. Damit ein Task mehrere Container gleichzeitig parallel offen halten kann, wird der cfIdx zur Unterscheidung der
	 * ContainerFile-Objekte benutzt. Falls das Verzeichnis der Datenidentifikation nicht existiert, wird es angelegt.
	 *
	 * @param task        Anfragender Task
	 * @param cfIdx       ContainerFile-Index
	 * @param did         Datenidentifikation
	 * @param dataKind    Datensatzart
	 * @param containerId Container-ID
	 *
	 * @throws PersistenceException
	 */
	public void accessContainer(Task task, long cfIdx, IdDataIdentification did, ArchiveDataKind dataKind, long containerId)
			throws PersistenceException {
		File contDir = getCreatePersDir(did.getObjectId(), did.getAtgId(), did.getAspectId(), did.getSimVariant(), dataKind);
		getContainerFile(task, cfIdx).accessContainer(did.getObjectId(), did.getAtgId(), did.getAspectId(), did.getSimVariant(), dataKind, containerId, contDir);
	}


	/**
	 * Assoziiert den angegebenen Task mit dem spezifizierten Container, so dass alle anschliessend von diesem Task aufgerufenen Methoden auf diesen Container
	 * wirken. Soll der Zugriff auf diesen Container beendet werden (um z.B. einen anderen Container zu adressieren), muss dazu die Methode {@link
	 * #leaveContainer(Task)} aufgerufen werden. Damit ein Task mehrere Container gleichzeitig parallel offen halten kann, wird der cfIdx zur Unterscheidung der
	 * ContainerFile-Objekte benutzt. Falls das Verzeichnis der Datenidentifikation nicht existiert, wird es angelegt.
	 *
	 * @param task        Anfragender Task
	 * @param cfIdx       ContainerFile-Index
	 * @param objId       Objekt-ID
	 * @param atgId       Attributgruppen-ID
	 * @param aspId       Aspekt-ID
	 * @param sv          Simulationsvariante
	 * @param dataKind    Datensatzart
	 * @param pathName        Pfadname des Verzeichnisses der Containerdatei
	 * @param containerId Container-ID
	 *
	 * @throws PersistenceException
	 */
	public void accessContainer(Task task, int cfIdx, long objId, long atgId, long aspId, int sv, ArchiveDataKind dataKind, String pathName, boolean createDir, long containerId)
			throws PersistenceException {
		File contDir = createDir ? getCreatePersDir(pathName) : new File(pathName);
		getContainerFile(task, cfIdx).accessContainer(objId, atgId, aspId, sv, dataKind, containerId, contDir);
	}

	/**
	 * Assoziiert den angegebenen Task mit der angegebenen Containerdatei, so dass alle anschliessend von diesem Task aufgerufenen Methoden auf diesen Container
	 * wirken. Es koennen nur lesende Operationen auf die Container-Datei ausgeführt werden. Vorgesehen für den Zugriff auf Containerdateien, die bereits auf ein
	 * Medium vom Typ B ausgelagert sind. Soll der Zugriff auf diesen Container beendet werden (um z.B. einen anderen Container zu adressieren), muss dazu die
	 * Methode {@link #leaveContainer(Task)} aufgerufen werden.
	 *
	 * @param task     Anfragender Task
	 * @param contFile Containerdatei
	 *
	 * @throws PersistenceException
	 */
	public void accessContainerReadOnly(Task task, File contFile) throws PersistenceException {
		accessContainerReadOnly(task, 0, contFile);
	}

	/**
	 * Assoziiert den angegebenen Task mit der angegebenen Containerdatei, so dass alle anschliessend von diesem Task aufgerufenen Methoden auf diesen Container
	 * wirken. Es koennen nur lesende Operationen auf die Container-Datei ausgeführt werden. Vorgesehen für den Zugriff auf Containerdateien, die bereits auf ein
	 * Medium vom Typ B ausgelagert sind. Soll der Zugriff auf diesen Container beendet werden (um z.B. einen anderen Container zu adressieren), muss dazu die
	 * Methode {@link #leaveContainer(Task)} aufgerufen werden. Damit ein Task mehrere Container gleichzeitig parallel offen halten kann, wird der cfIdx zur
	 * Unterscheidung der ContainerFile-Objekte benutzt.
	 *
	 * @param task     Anfragender Task
	 * @param cfIdx    ContainerFile-Index
	 * @param contFile Containerdatei
	 *
	 * @throws PersistenceException
	 */
	public void accessContainerReadOnly(Task task, int cfIdx, File contFile) throws PersistenceException {
		getContainerFile(task, cfIdx).accessContainerReadOnly(contFile);
	}

	/**
	 * Beendet das Arbeiten mit dem zuvor ueber eine der <code>accessContainer()</code>-Methoden definierten Container. Diese Methode muss aufgerufen werden, wenn
	 * danach auf einen anderen Container zugegriffen werden soll.
	 *
	 * @param task Anfragender Task
	 *
	 * @throws PersistenceException
	 */
	public void leaveContainer(Task task) throws PersistenceException {
		leaveContainer(task, 0);
	}

	/**
	 * Beendet das Arbeiten mit dem zuvor ueber eine der <code>accessContainer()</code>-Methoden definierten Container. Diese Methode muss aufgerufen werden, wenn
	 * danach auf einen anderen Container zugegriffen werden soll. Damit ein Task mehrere Container gleichzeitig parallel offen halten kann, wird der cfIdx zur
	 * Unterscheidung der ContainerFile-Objekte benutzt.
	 *
	 * @param task  Anfragender Task
	 * @param cfIdx ContainerFile-Index
	 *
	 * @throws PersistenceException
	 */
	public void leaveContainer(Task task, long cfIdx) throws PersistenceException {
		getContainerFile(task, cfIdx).leaveContainer();
	}

	/**
	 * Prueft ob bereits eine Container-Datei für den mit {@link #accessContainer(Task,long,long,long,int,ArchiveDataKind,long)} spezifizierten Container
	 * existiert.
	 *
	 * @param task Anfragende Task
	 *
	 * @return Kennzeichen
	 *
	 * @throws PersistenceException
	 */
	public boolean existsContainer(Task task) throws PersistenceException {
		return getContainerFile(task).existsContainer();
	}

	/**
	 * Erzeugt eine neue Container-Datei für den mit {@link #accessContainer(Task,long,long,long,int,ArchiveDataKind,long)} spezifizierten Container. Schreibt den
	 * Container-Header und befuellt diesen mit den angegebenen Parametern.
	 *
	 * @param task   Anfragender Task
	 * @param toSave Sichern
	 *
	 * @throws PersistenceException
	 */
	public void createContainer(Task task, boolean toSave) throws PersistenceException {
		getContainerFile(task).createContainer(toSave);
	}

	/**
	 * Legt im Verzeichnis der Datenidentifikation ein {@link #REBUILDINDEX_FILE_FLAG_NAME}-Datei an.
	 *
	 * @param didPath Pfad der Datenidentifikation
	 *
	 * @throws PersistenceException
	 * @see #REBUILDINDEX_FILE_FLAG_NAME
	 */
	public void createRebuildIxdFlagFile(String didPath) throws PersistenceException {
		try {
			if(!new File(didPath, REBUILDINDEX_FILE_FLAG_NAME).createNewFile()) {
				logger.fine(REBUILDINDEX_FILE_FLAG_NAME + "-Datei existiert bereits in '" + didPath + "'");
			}
		}
		catch(IOException e) {
			throw new PersistenceException("Fehler beim Anlegen der " + REBUILDINDEX_FILE_FLAG_NAME + "-Datei in '" + didPath + "'");
		}
	}

	/**
	 * Loescht die {@link #REBUILDINDEX_FILE_FLAG_NAME}-Datei.
	 *
	 * @param didPath
	 */
	private void removeRebuildIxdFlagFile(String didPath) {
		new File(didPath, REBUILDINDEX_FILE_FLAG_NAME).delete();
	}

	/**
	 * Schliesst die Container-Datei für den mit {@link #accessContainer(Task,long,long,long,int,ArchiveDataKind,long)} spezifizierten Container. Der
	 * Container-Header wird entsprechend aktualisiert. Die Min/Max-Werte und die Anzahl der Datensätze werden dem DataIdentNode und dessen Indexen entnommen.
	 *
	 * @param task Anfragender Task
	 * @param din  DataIdentNode des aktuellen Containers
	 * @param adk  Datensatzart des aktuellen Containers
	 *
	 * @throws PersistenceException
	 * @throws IndexException
	 */
	public void closeContainer(Task task, DataIdentNode din, ArchiveDataKind adk) throws PersistenceException, IndexException {
		ContainerFile cf = getContainerFile(task);
		cf.closeContainer(
				din.getCurContEntries(adk),
				din.minATime(adk),
				din.maxATime(adk),
				din.minDTime(adk),
				din.maxDTime(adk),
				din.minDIdx(adk),
				din.maxDIdx(adk),
				din.arSParamGetVorhalten()
		);
		din.setCurContEntries(Util.getDataKindIndex(adk), DataIdentNode.INVALID);
	}

	/**
	 * Schliesst einen Container, indem jeder Datensatz-Header des Containers gelesen wird, um die Min-/Max-Werte und die Anzahl der Datensätze zu bestimmen. Die
	 * Information der Indexe reicht nicht aus, weil dort die Anzahl der Datensätze nicht verfuegbar ist.
	 *
	 * @param task Anfragender Task
	 * @param din  DataIdentNode des aktuellen Containers
	 * @param adk  Datensatzart des aktuellen Containers
	 *
	 * @throws PersistenceException
	 */
	public void closeContainerRecount(Task task, DataIdentNode din, ArchiveDataKind adk) throws PersistenceException {
		getContainerFile(task).closeContainerRecount(din.arSParamGetVorhalten());
		din.setCurContEntries(Util.getDataKindIndex(adk), DataIdentNode.INVALID);
	}

	/**
	 * Prueft ob der mit {@link #accessContainer(Task,long,long,long,int,ArchiveDataKind,long)} spezifizierte Container bereits abgeschlossen ist.
	 *
	 * @param task Anfragender Task
	 *
	 * @return Kennzeichen, ob Container abgeschlossen ist
	 *
	 * @throws PersistenceException
	 */
	public boolean isContainerClosed(Task task) throws PersistenceException {
		return isContainerClosed(task, 0);
	}

	/**
	 * Prueft ob der mit {@link #accessContainer(Task,int,long,long,long,int,ArchiveDataKind,long)} spezifizierte Container bereits abgeschlossen ist.
	 *
	 * @param task Anfragender Task
	 *
	 * @return Kennzeichen, ob Container abgeschlossen ist
	 *
	 * @throws PersistenceException
	 */
	public boolean isContainerClosed(Task task, int cfIdx) throws PersistenceException {
		return getContainerFile(task, cfIdx).isContainerClosed();
	}

	/**
	 * Loescht alle für den angegebenen Task angelegten ContainerFile-Objekte bis auf das angegebene Minimum. Es werden die ContainerFiles mit den groessten
	 * ContainerFile-Indexen geloescht. Wenn das Minimum 0 ist, wird der Index-Map ebenfalls geloescht. Diese Methode ist notwendig, damit die ContainerFiles nicht
	 * stetig wachsen.
	 *
	 * @param task   Anfragender Task
	 * @param minCFs Minimale Anzahl ContainerFiles, die nicht geloescht werden.
	 */
	public void removeContainerFileObjects(Task task, int minCFs) {
		synchronized(containerFileMap) {
			Map<Long, ContainerFile> cfMap = containerFileMap.get(task);
			if(cfMap == null) return; // Fertig, da keine ContainerFiles für diesen Task angelegt sind.

			List<Long> cfIdxsToRemove = new ArrayList<Long>(); // Liste der zu loeschenden ContainerFile-Indexe
			List<Long> cfIdxs = new ArrayList<Long>(cfMap.keySet());
			Collections.sort(cfIdxs); // ContainerFile-Indexe aufsteigend sortieren
			for(int i = minCFs; i < cfIdxs.size(); i++) {
				cfIdxsToRemove.add(cfIdxs.get(i)); // Index zum Loeschen vormerken, wenn oberhalb minCFs
			}
			for(Long idx : cfIdxsToRemove) {
				cfMap.remove(idx);
			}
			if(minCFs == 0) containerFileMap.remove(task);
		}
	}

	/**
	 * Loescht alle für den angegebenen Task angelegten ContainerFile-Objekte und den Index-Map. Diese Methode ist notwendig, damit die ContainerFiles nicht
	 * stetig wachsen.
	 *
	 * @param task Anfragender Task
	 */
	public void removeContainerFileObjects(Task task) {
		removeContainerFileObjects(task, 0);
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Container: Indexe
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Weist, falls nicht bereits geschehen, dem DataIdentNode seine Indexe zu.
	 *
	 * @param dit     Baum bestehend aus {@link DataIdentNode}
	 * @param ads     Spezifikation des Archivdatensatzes (enthaelt objID, atgID, aspID, sv)
	 * @param adk     Datensatzart
	 * @param task    Anfragender Task
	 * @param tempMem Temporaerer Speicher zum EInlesen der Indexe
	 *
	 * @throws DIdNodeNotFoundException
	 * @throws IndexException
	 * @throws PersistenceException
	 */
	public void assignIndexes(DataIdentTree dit, ArchiveDataSpecification ads, ArchiveDataKind adk, Task task, byte[] tempMem)
			throws DIdNodeNotFoundException, IndexException, PersistenceException {
		DataDescription dd = ads.getDataDescription();
		assignIndexes(dit, ads.getObject().getId(), dd.getAttributeGroup().getId(), dd.getAspect().getId(), dd.getSimulationVariant(), adk, task, tempMem);
	}

	public void assignIndexes(DataIdentTree dit, IdDataIdentification ads, ArchiveDataKind adk, Task task, byte[] tempMem)
			throws DIdNodeNotFoundException, IndexException, PersistenceException {
		assignIndexes(dit, ads.getObjectId(), ads.getAtgId(), ads.getAspectId(), ads.getSimVariant(), adk, task, tempMem);
	}

	/**
	 * Weist, falls nicht bereits geschehen, dem DataIdentNode seine Indexe zu. Prueft dabei, ob der letzte Eintrag eines Index fortgefuehrt werden muss (wenn der
	 * letzte Container angebrochen ist und nicht abgeschlossen werden muss). Die Container werden nicht verändert.
	 * <p>
	 * Wenn der Index nicht fortgefuehrt wird, werden der enstprechende {@link DataIdentNode#containerEntries} auf -1 gesetzt. Wenn der Index fortgefuehrt wird,
	 * wird im letzten Container nachgezaehlt (aufwendig) und {@link DataIdentNode#containerEntries} auf das Ergebnis gesetzt.
	 *
	 * @param dit     Baum bestehend aus {@link DataIdentNode}
	 * @param objID   Objekt-ID
	 * @param atgID   Attributgruppen-ID
	 * @param aspID   Aspekt-ID
	 * @param sv      Simulationsvariante
	 * @param archiveDataKind     Datensatzart
	 * @param task    Anfragender Task
	 * @param tempMem Temporaerer Speicher zum Einlesen der Indexe
	 *
	 * @throws DIdNodeNotFoundException
	 * @throws IndexException
	 * @throws PersistenceException
	 */
	public void assignIndexes(DataIdentTree dit, long objID, long atgID, long aspID, int sv, ArchiveDataKind archiveDataKind, Task task, byte[] tempMem)
			throws DIdNodeNotFoundException, IndexException, PersistenceException {
		int archiveDataKindIndex = Util.getDataKindIndex(archiveDataKind);
		DataIdentNode din = dit.getDataIdentNode(objID, atgID, aspID, sv);
		String didPath = getPathName(objID, atgID, aspID, sv, archiveDataKindIndex);
		long lastDIDContId = getLargestContainerID(didPath);

		if(lastDIDContId < 0) {
			// Diese DID wurde überhaupt noch nie archiviert, daher letzten Eintrag nicht weiterverwenden (welchen auch?)
			din.assignNewIndexes(archiveDataKindIndex);
		}
		else {
			try {
				// es existiert bereits mindestens ein Container, auf Abschluss pruefen
				accessContainer(task, objID, atgID, aspID, sv, archiveDataKind, lastDIDContId);
				boolean containerClosed = isContainerClosed(task);

				if(containerClosed) {
					din.assignNewIndexes(archiveDataKindIndex);	// Container abgeschlossen, letzten Eintrag nicht weiterverwenden
					din.setCurContEntries(archiveDataKindIndex, DataIdentNode.INVALID);
				}
				else {
					// Anzahl im Container nachzaehlen, dabei lastNoSrc-Flag mitnehmen
					int anz = 0;
					boolean lastDataIsNoSrc = false;
					ContainerDataIterator iter = getArchiveData(task);
					while(iter.readNextDataHdr()) {
						anz++;
						lastDataIsNoSrc = iter.isNoSource();
						iter.skipData();
					}
					iter.close();
					if(archiveDataKind.equals(ArchiveDataKind.ONLINE)) din.setLastOAWasNoSource(lastDataIsNoSrc);

					try {
						// Container nicht abgeschlossen, letzten Eintrag weiterverwenden:
						din.assignContinuedIndexes(archiveDataKindIndex, didPath, tempMem, anz);

						if(lastDIDContId != din.getOpenContID(archiveDataKind)) {
							throw new CorruptIndexException(
									"Letzte ContainerID nicht eindeutig: " + didPath + "/" + ContainerFile.getContainerFileName(lastDIDContId) + ", Index:"
									+ din.getOpenContID(archiveDataKind)
							);
						}
					}
					catch(CorruptIndexException e) {
						leaveContainer(task);
						rebuildIndex(task, tempMem, objID, atgID, aspID, sv, archiveDataKind);
						logger.warning("Indexe neu aufgebaut: " + didPath + Debug.NEWLINE + e.getMessage());
						din.assignContinuedIndexes(archiveDataKindIndex, didPath, tempMem, anz);
						accessContainer(task, objID, atgID, aspID, sv, archiveDataKind, lastDIDContId);
					}
				}
			}
			finally {
				leaveContainer(task);
			}
		}
	}

	public void prepareShutdown(DataIdentTree dit) throws InterruptedException {
		CacheManager.getInstance().close();
		if(dit != null) {
			DataIdentTreeIterator it = dit.createIterator();
			ExecutorService threadPool = Executors.newFixedThreadPool(archivMgr.getNumCloseIndexThreads());
			final AtomicInteger anzOA = new AtomicInteger(0);
			final AtomicInteger anzON = new AtomicInteger(0);
			final AtomicInteger anzNA = new AtomicInteger(0);
			final AtomicInteger anzNN = new AtomicInteger(0);

			// Indexdateien in mehreren Threads schließen
			while(it.getNext()) {
				DataIdentNode din = it.getDataIdentNode();
				IdDataIdentification dataIdentification = it.getDataIdentification();
				threadPool.submit(new IndexCloserWorker(anzOA, anzON, anzNA, anzNN, din, dataIdentification));
			}

			// Abwarten, bis alle Threads fertig sind.
			threadPool.shutdown();
			while(!threadPool.awaitTermination(10, TimeUnit.SECONDS)) {
				logger.info("Indexe werden geschlossen: " + anzOA.get() + " OA, " + anzON.get() + " ON, " + anzNA.get() + " NA, " + anzNN.get() + " NN");
			}
			logger.info("Alle Indexe geschlossen: " + anzOA.get() + " OA, " + anzON.get() + " ON, " + anzNA.get() + " NA, " + anzNN.get() + " NN");
		}
		if(_preventWriteStartupInfo) {
			logger.warning("Herunterfahren: Startup-Info wird unterdrückt, weil beim Wiederherstellungslauf ein Fehler aufgetreten ist");
			return;
		}
		try {
			startupProps.setVal(StartupProperties.STUP_MAX_CONT_ID, getLastContainerID());
			startupProps.setVal(StartupProperties.STUP_LAST_ATIME, ArchiveTask.getLastArchiveTime());
			startupProps.setVal(StartupProperties.STUP_LAST_USED_MEDIUM_ID, BackupTask.getLastUsedMediumID());
			startupProps.writeStartUpProperties();
		}
		catch(PersistenceException e) {
			logger.error("Herunterfahren: Fehler beim Erzeugen der Startup-Info: " + e.getMessage());
		}
	}

	/**
	 * Schliesst die Indexe der Datensatzart des DataIndentNode.
	 *
	 * @param din
	 * @param objID
	 * @param atgID
	 * @param aspID
	 * @param sv
	 * @param archiveDataKind    Datensatzart
	 * @param tmpMem temporaerer Speicher, der zum Schliessen des Index benoetigt wird
	 *
	 * @return Wahr, falls der Index gescjhlossen werden musste, falsch, wenn der Index unveraendert war.
	 *
	 * @throws IndexException
	 */
	public boolean closeIndex(DataIdentNode din, IdDataIdentification dataIdentification, ArchiveDataKind archiveDataKind, byte[] tmpMem) throws IndexException {
		if(din.isRebuildIndexFileFlagCreated(archiveDataKind)) {
			String didPath = getPathName(dataIdentification, archiveDataKind);
			din.closeIndexes(archiveDataKind, didPath, tmpMem);
			removeRebuildIxdFlagFile(didPath);
			return true;
		}
		return false;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Container: Header
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Liest den Header der Container-Datei ein und springt im Eingabestrom an die Stelle hinter dem Header. Die Container-Header Parameter stehen dann als
	 * Key/Value-Paare zur Verfuegung und koennen mit {@link #getContainerHeaderParamAsString(Task,KeyValParam)} abgefragt werden.
	 *
	 * @param task Anfragender Task
	 *
	 * @throws PersistenceException
	 */
	public void readContainerHeader(Task task) throws PersistenceException {
		getContainerFile(task).readContainerHeader();
	}

	/**
	 * Liest den Header der Container-Datei ein und springt im Eingabestrom an die Stelle hinter dem Header. Die Container-Header Parameter stehen dann als
	 * Key/Value-Paare zur Verfuegung und koennen mit {@link #getContainerHeaderParamAsString(Task,KeyValParam)} abgefragt werden.
	 *
	 * @param task  Anfragender Task
	 * @param cfIdx ContainerFile-Index
	 *
	 * @throws PersistenceException
	 */
	public void readContainerHeader(Task task, long cfIdx) throws PersistenceException {
		getContainerFile(task, cfIdx).readContainerHeader();
	}

	/**
	 * Nachdem mit {@link #readContainerHeader(Task)} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines
	 * Header-Parameters ermittelt werden. Wirfte eine {@link PersistenceException}, falls der Parameter nicht gefunden wurde. Liefert den Wert des Parameters
	 * unabhaengig von dessen Typ immer als String zurueck.
	 *
	 * @param task  Anfragende Task
	 * @param param Container-Header-Parameter
	 *
	 * @return Parameter-Wert als String
	 *
	 * @see KeyValParam
	 */
	public String getContainerHeaderParamAsString(Task task, KeyValParam param) throws PersistenceException {
		return getContainerFile(task).getContHdrParamAsString(param);
	}

	/**
	 * Nachdem mit {@link #readContainerHeader(Task)} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines
	 * Header-Parameters ermittelt werden. Wirfte eine {@link PersistenceException}, falls der Parameter nicht gefunden wurde. Liefert den Wert des Parameters
	 * unabhaengig von dessen Typ immer als String zurueck.
	 *
	 * @param task  Anfragende Task
	 * @param cfIdx ContainerFile-Index
	 * @param param Container-Header-Parameter
	 *
	 * @return Parameter-Wert als String
	 *
	 * @see KeyValParam
	 */
	public String getContainerHeaderParamAsString(Task task, long cfIdx, KeyValParam param) throws PersistenceException {
		return getContainerFile(task, cfIdx).getContHdrParamAsString(param);
	}

	/**
	 * Nachdem mit {@link #readContainerHeader(Task)} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines
	 * Header-Parameters ermittelt werden. Liefert den Wert des Parameters als Datensatzart zurueck. Wirfte eine {@link PersistenceException}, falls der Parameter
	 * nicht gefunden wurde oder nicht in den Typ {@link ArchiveDataKind} umgeformt werden kann.
	 *
	 * @param task  Anfragende Task
	 * @param param Container-Header-Parameter
	 *
	 * @return Datensatzart
	 *
	 * @throws PersistenceException
	 */
	public ArchiveDataKind getContainerHeaderParamAsArchiveDataKind(Task task, KeyValParam param) throws PersistenceException {
		return getContainerFile(task).getContHdrParamAsArchiveDataKind(param);
	}

	/**
	 * Nachdem mit {@link #readContainerHeader(Task)} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines
	 * Header-Parameters ermittelt werden. Liefert den Wert des Parameters als boolschen Wert zurueck. Wirfte eine {@link PersistenceException}, falls der
	 * Parameter nicht gefunden wurde oder nicht in den Typ <code>boolean</code> umgeformt werden kann.
	 *
	 * @param task  Anfragende Task
	 * @param param Container-Header-Parameter
	 *
	 * @return <code>true</code> oder <code>false</code>
	 *
	 * @throws PersistenceException
	 */
	public boolean getContainerHeaderParamAsBoolean(Task task, KeyValParam param) throws PersistenceException {
		return getContainerFile(task).getContHdrParamAsBool(param);
	}

	/**
	 * Nachdem mit {@link #readContainerHeader(Task)} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines
	 * Header-Parameters ermittelt werden. Liefert den Wert des Parameters als boolschen Wert zurueck. Wirfte eine {@link PersistenceException}, falls der
	 * Parameter nicht gefunden wurde oder nicht in den Typ <code>boolean</code> umgeformt werden kann.
	 *
	 * @param task  Anfragende Task
	 * @param cfIdx ContainerFile-Index
	 * @param param Container-Header-Parameter
	 *
	 * @return <code>true</code> oder <code>false</code>
	 *
	 * @throws PersistenceException
	 */
	public boolean getContainerHeaderParamAsBoolean(Task task, long cfIdx, KeyValParam param) throws PersistenceException {
		return getContainerFile(task, cfIdx).getContHdrParamAsBool(param);
	}

	/**
	 * Nachdem mit {@link #readContainerHeader(Task)} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines
	 * Header-Parameters ermittelt werden. Liefert den Wert des Parameters als numerischen Wert vom Typ <code>long</code> zurueck. Wirfte eine {@link
	 * PersistenceException}, falls der Parameter nicht gefunden wurde oder nicht in den Typ <code>long</code> umgeformt werden kann.
	 *
	 * @param task  Anfragende Task
	 * @param param Container-Header-Parameter
	 *
	 * @return Ganzzahl als <code>long</code>
	 *
	 * @throws PersistenceException
	 */
	public long getContainerHeaderParamAsLong(Task task, KeyValParam param) throws PersistenceException {
		return getContainerFile(task).getContHdrParamAsLong(param);
	}

	/**
	 * Nachdem mit {@link #readContainerHeader(Task)} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines
	 * Header-Parameters ermittelt werden. Liefert den Wert des Parameters als numerischen Wert vom Typ <code>long</code> zurueck. Wirfte eine {@link
	 * PersistenceException}, falls der Parameter nicht gefunden wurde oder nicht in den Typ <code>long</code> umgeformt werden kann.
	 *
	 * @param task  Anfragende Task
	 * @param cfIdx ContainerFile-Index
	 * @param param Container-Header-Parameter
	 *
	 * @return Ganzzahl als <code>long</code>
	 *
	 * @throws PersistenceException
	 */
	public long getContainerHeaderParamAsLong(Task task, int cfIdx, KeyValParam param) throws PersistenceException {
		return getContainerFile(task, cfIdx).getContHdrParamAsLong(param);
	}

	/**
	 * Nachdem mit {@link #readContainerHeader(Task)} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines
	 * Header-Parameters ermittelt werden. Liefert den Wert des Parameters als numerischen Wert vom Typ <code>long</code> zurueck. Wirfte eine {@link
	 * PersistenceException}, falls der Parameter nicht gefunden wurde oder nicht in den Typ <code>long</code> umgeformt werden kann.
	 *
	 * @param task  Anfragende Task
	 * @param param Container-Header-Parameter
	 *
	 * @return Ganzzahl als <code>long</code>
	 *
	 * @throws PersistenceException
	 */
	public int getContainerHeaderParamAsInt(Task task, KeyValParam param) throws PersistenceException {
		return getContainerFile(task).getContHdrParamAsInt(param);
	}

	/**
	 * Nachdem mit {@link #readContainerHeader(Task)} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines Parameters im
	 * Container-Header geaendert werden. Wirft eine {@link PersistenceException}, wenn die Aenderung nicht durchgefuehrt werden kann.
	 *
	 * @param task  Anfragender Task
	 * @param param Container-Header-Parameter
	 * @param val   Wert des Container-Header-Parameters
	 *
	 * @throws PersistenceException
	 */
	public void setContainerHeaderParam(Task task, KeyValParam param, String val) throws PersistenceException {
		getContainerFile(task).setContHdrParam(param, val);
	}

	/**
	 * Nachdem mit {@link #readContainerHeader(Task)} der Header einer Container-Datei eingelesen worden ist, kann mit dieser Methode der Wert eines numerischen
	 * Parameters im Container-Header geaendert werden. Wirft eine {@link PersistenceException}, wenn die Aenderung nicht durchgefuehrt werden kann.
	 *
	 * @param task  Anfragender Task
	 * @param param Container-Header-Parameter
	 * @param val   Numerischer Wert des Container-Header-Parameters
	 *
	 * @throws PersistenceException
	 */
	public void setContainerHeaderParam(Task task, KeyValParam param, long val) throws PersistenceException {
		getContainerFile(task).setContHdrParam(param, val);
	}

	/**
	 * Der Container-Header der Container-Datei, die mit dem Aufruf {@link #accessContainer(Task,long,long,long,int,ArchiveDataKind,long)} spezifiziert worden ist,
	 * wird mit dem Container-Header aus der angegebenen Container-Datei ueberschrieben.
	 *
	 * @param task Anfragender Task
	 * @param file Container-Datei
	 *
	 * @throws PersistenceException
	 */
	public void restoreHeader(Task task, File file) throws PersistenceException {
		getContainerFile(task).restoreHeader(file);
	}

	/**
	 * Die mit {@link #setContainerHeaderParam(Task task,KeyValParam param,String val)} gesetzten Parameter werden im Container-Header persistiert.
	 *
	 * @param task Anfragender Task
	 *
	 * @throws PersistenceException
	 */
	public void writeContainerHeader(Task task) throws PersistenceException {
		writeContainerHeader(task, 0);
	}

	/**
	 * Die mit {@link #setContainerHeaderParam(Task task,KeyValParam param,String val)} gesetzten Parameter werden im Container-Header persistiert.
	 *
	 * @param task Anfragender Task
	 *
	 * @throws PersistenceException
	 */
	public void writeContainerHeader(Task task, int cfIdx) throws PersistenceException {
		getContainerFile(task, cfIdx).writeContainerHeader();
	}

	/**
	 * Versucht, die StartUp-Properties-Datei einzulesen. Wenn das nicht gelingt, wird das gesamte Persistenzverzeichnis ({@link RestorePersDirTsk}) durchlaufen
	 * und versucht, einen gueltigen Startpunkt wiederherzustellen. Das Persistenzverzeichnis wird in <code>InQueuesMgr.NUM_OF_ARCH_QUEUES_ONLINE</code> Teile
	 * aufgeteilt, die jeweils von einem Thread bearbeitet werden (zwengs Performance). Am Schluss wird die StartUp-Properties-Datei geloescht. Beim Herunterfahren
	 * wird in {@link #prepareShutdown(DataIdentTree,byte[])} die Datei neu geschrieben. Daran kann das ArS beim naechsten Start erkennen, ob es ordnungsgemaess
	 * heruntergefahren wurde oder ob ein Wiederherstellungslauf erforderlich ist.
	 *
	 * @return Wahr, falls der Durchlauf des Persistenzverzeichnisses erfolgreich war und das Archivsystem starten kann, falsch sonst.
	 */
	public boolean startupProcedure() {
		String logMsg = "";
		long mxAT = 0;
		int counter = 0, lastUsedMediumID = 0;
		try {
			startupProps.readStartUpProperties();
			nextContainerID.set(startupProps.getValAsLong(StartupProperties.STUP_MAX_CONT_ID));
			mxAT = startupProps.getValAsLong(StartupProperties.STUP_LAST_ATIME);
			lastUsedMediumID = startupProps.getValAsInt(StartupProperties.STUP_LAST_USED_MEDIUM_ID);
		}
		catch(PersistenceException e) {
			_preventWriteStartupInfo = true;
			logger.info(e.getMessage() + Debug.NEWLINE + "Starte Wiederherstellungslauf...");
			RestorePersDirTsk.startRestore(this, archivMgr.getNumOfArchQueuesOnline());
			while(!RestorePersDirTsk.restoreDone()) {
				try {
					if(archivMgr.wasTerminated()) {		// Archivsystem per Telnet beendet?
						RestorePersDirTsk.terminateRestore();
						throw new InterruptedException();
					}
					Thread.sleep(100);
					if(++counter % 50 == 0) logger.info(RestorePersDirTsk.status());
				}
				catch(InterruptedException ie) {
					logger.info("Wiederherstellungslauf abgebrochen.");
					return false;
				}
			}
			if(!RestorePersDirTsk.success()) {
				logger.error("Wiederherstellungslauf fehlgeschlagen.\n" + RestorePersDirTsk.summary());
				return false;
			}
			_preventWriteStartupInfo = false;
			nextContainerID = new AtomicLong(RestorePersDirTsk.getMaxContID());
			mxAT = RestorePersDirTsk.getMaxArchiveTime();
			lastUsedMediumID = RestorePersDirTsk.getMaxMedienID();
			logMsg = RestorePersDirTsk.summary() + "\n";
			RestorePersDirTsk.cleanUpTask();	// Speicher freigeben
		}

		long now = ArchiveManager.archiveTime();
		ArchiveTask.setLastArchiveTime(mxAT > 0 ? mxAT : now);	// gemaess TAnfArS 5.1.2.4.3.3
		BackupTask.setLastUsedMediumID(lastUsedMediumID);

		logger.info(
				logMsg + "StartupInfo: max. ContainerID=" + nextContainerID + ", letzte MedienID = " + lastUsedMediumID + ", max. Archivzeit="
				+ Util.timestrMillisFormatted(mxAT) + (mxAT <= 0 ? " (nicht ermittelbar, wurde auf " + Util.timestrMillisFormatted(now) + " gesetzt)" : "")
		);

		if(!startupProps.deleteStartupPropsFile())
		// Gibt zwar ein Problem beim Beenden des Archivsystems, sollte aber den Start nicht verhindern
		{
			logger.warning("StartUp-Properties konnten nicht geloescht werden: " + startupProps);
		}

		return true;
	}


	/**
	 * Baut Datenindex-, Archivzeit- und Datenzeitindexe der Datenidentifikation und Datensatzart unter dem angegebenen Verzeichnis neu auf. Eventuell vorhandene
	 * Indexdateien werden vorher geloescht.
	 *
	 * @param task    Anfragender Task
	 * @param tmpBuf  Temporaerer Speicherbereich, der zum Neuaufbau der Indexe verwendet werden kann.
	 * @param didPath Vollstaendiger Pfad bis zur Datensatzart
	 *
	 * @throws PersistenceException
	 * @throws IndexException
	 * @see #rebuildIndex(Task,byte[],long,long,long,int,ArchiveDataKind)
	 */
	public int[] rebuildIndex(Task task, byte[] tmpBuf, String didPath) throws PersistenceException, IndexException {
		File dir = new File(didPath);
		long ids[] = extractDIdFromPath(dir);
		return rebuildIndex(task, tmpBuf, ids[0], ids[1], ids[2], (int)ids[3], getArchiveDataKindFromDir(dir));
	}


	/**
	 * Baut Datenindex-, Archivzeit- und Datenzeitindexe der angegebenen Datenidentifikation und Datensatzart neu auf. Eventuell vorhandene Indexdateien werden
	 * vorher geloescht. Nach Moeglichkeit werden die Min/Max-Werte aus den Container-Headern verwendet. Falls diese nicht vorhanden sind (weil der Container nicht
	 * abgeschlossen ist), werden alle Datensatz-Header gelesen. Sollte ein Container eigentlich die Index-Info im Header stehen haben und sie steht dort nicht,
	 * wird der Header ergänzt. <p>Ruft {@link #accessContainer(de.bsvrz.ars.ars.mgmt.tasks.Task, long, long, long, int,
	 * de.bsvrz.dav.daf.main.archive.ArchiveDataKind, long)} auf.
	 *
	 * @param task   Anfragender Task
	 * @param tmpBuf Temporaerer Speicherbereich, der zum Neuaufbau der Indexe verwendet werden kann.
	 * @param objId  Objekt-ID
	 * @param atgId  Attributgruppen-ID
	 * @param aspId  Aspekt-ID
	 * @param sv     Simulationsvariante
	 * @param archiveDataKind    Datensatzart
	 *
	 * @return Ein zweielementiges Feld: der 1. Wert ist die Zahl der besuchten Container, der 2. ist die Zahl der Container, bei denen die Index-Info nicht aus
	 *         dem Header gelesen werden konnte, sondern durch Durchlaufen aller Datensätze ermittelt werden musste.
	 *
	 * @throws PersistenceException
	 * @throws IndexException
	 * @throws PersistenceException
	 */
	public int[] rebuildIndex(Task task, byte[] tmpBuf, long objId, long atgId, long aspId, int sv, ArchiveDataKind archiveDataKind)
			throws IndexException, PersistenceException {
		int[] cont_visited_fullrebuilt = new int[2];
		String didPath = getPathName(objId, atgId, aspId, sv, archiveDataKind);
//		System.out.println("didPath = " + didPath);
//		Thread.dumpStack();
		File monoIdxIdxFile = new File(didPath, ATimeMonotoneIndex.IDX_FILENAME);
		File dataTimeIdxFile = new File(didPath, DataTimeIndex.IDX_FILENAME);

		final boolean dataTimeIndexFileDeleted = Util.deleteIfExists(dataTimeIdxFile);
		final boolean monoIdxIdxFileDeleted = Util.deleteIfExists(monoIdxIdxFile);

		if(!monoIdxIdxFileDeleted || !dataTimeIndexFileDeleted) {
			throw new PersistenceException("Loeschen der alten Indexdateien fehlgeschlagen: " + didPath);
		}
		else {
			String[] fNames = new File(didPath).list(ContainerFile.CONT_FILENAME_FILTER);
			if(fNames != null && fNames.length > 0) {
				Arrays.sort(fNames);
				DataTimeIndex dtidx = new DataTimeIndex(REBUILDINDEX_IDX_SIZE, didPath, tmpBuf, DataTimeIndex.ERASE_OLD);

				DIdxATimeMonotoneIndex dIdxATIdx = null;
				ATimeMonotoneIndex aTIdx = null;
				DataIndexIndex dIdxIdx = null;

				if(archiveDataKind.isRequested()) {
					// Bei nachgelieferten Daten kann es Datenindex-Rueckspruenge geben, deshalb muessen
					// Archivzeit (monoton) und Datenindex (nicht monoton) in getrennten Indexen gehalten werden:
					aTIdx = new ATimeMonotoneIndex(REBUILDINDEX_IDX_SIZE, didPath, ATimeMonotoneIndex.ERASE_OLD);
					dIdxIdx = new DataIndexIndex(REBUILDINDEX_IDX_SIZE, didPath, tmpBuf, DataIndexIndex.ERASE_OLD);
				}
				else
				// Bei aktuellen Daten sind Archivzeit und Datenindex stets monoton und koennen im selben Index
				// gehalten werden:
				{
					dIdxATIdx = new DIdxATimeMonotoneIndex(REBUILDINDEX_IDX_SIZE, didPath, DIdxATimeMonotoneIndex.ERASE_OLD);
				}

				for(String contFile : fNames) {
					try {
						cont_visited_fullrebuilt[0]++;
						long containerID = ContainerFile.getContID(contFile);
						accessContainer(task, objId, atgId, aspId, sv, archiveDataKind, containerID);
						readContainerHeader(task);
						long minDidx = getContainerHeaderParamAsLong(task, ContainerHdr.CHP_DATA_IDX_MIN);
						long maxDidx = getContainerHeaderParamAsLong(task, ContainerHdr.CHP_DATA_IDX_MAX);
						long minDtime = getContainerHeaderParamAsLong(task, ContainerHdr.CHP_DATA_TIME_MIN);
						long maxDtime = getContainerHeaderParamAsLong(task, ContainerHdr.CHP_DATA_TIME_MAX);
						long minAtime = getContainerHeaderParamAsLong(task, ContainerHdr.CHP_ARC_TIME_MIN);
						long maxAtime = getContainerHeaderParamAsLong(task, ContainerHdr.CHP_ARC_TIME_MAX);

						boolean rebuild = new Long(ContainerHdr.CHP_DATA_IDX_MIN.getDefaultVal()) == minDidx
						                  || new Long(ContainerHdr.CHP_DATA_IDX_MAX.getDefaultVal()) == maxDidx
						                  || new Long(ContainerHdr.CHP_DATA_TIME_MIN.getDefaultVal()) == minDtime
						                  || new Long(ContainerHdr.CHP_DATA_TIME_MAX.getDefaultVal()) == maxDtime
						                  || new Long(ContainerHdr.CHP_ARC_TIME_MIN.getDefaultVal()) == minAtime
						                  || new Long(ContainerHdr.CHP_ARC_TIME_MAX.getDefaultVal()) == maxAtime;

						if(!rebuild) {
							try {
								// Sämtliche Prüfungen der verschiedenen Indizes müssen durchgeführt werden, bevor irgendetwas verändert wird.
								// Zuerst wird geprüft, ob das Hinzufügen zu allen Indizes funktionieren würde
								dtidx.checkForAdd(didPath, tmpBuf, minDtime, maxDtime, minDidx, maxDidx, containerID);
								if(dIdxATIdx != null) {
									dIdxATIdx.checkForAdd(didPath, minDidx, maxDidx, minAtime, maxAtime, minDtime, maxDtime, containerID);
								}
								else {
									aTIdx.checkForAdd(didPath, minDidx, maxDidx, minAtime, maxAtime, minDtime, maxDtime, containerID);
									dIdxIdx.checkForAdd(didPath, tmpBuf, minDidx, maxDidx, containerID);
								}

								// Erst wenn sicher ist, dass das Hinzufügen zu allen Indizes funktionieren würde, werden die Indizes aktualisiert
								dtidx.add(didPath, tmpBuf, minDtime, maxDtime, minDidx, maxDidx, containerID);
								if(dIdxATIdx != null) {
									dIdxATIdx.add(didPath, minDidx, maxDidx, minAtime, maxAtime, minDtime, maxDtime, containerID);
								}
								else {
									aTIdx.add(didPath, minDidx, maxDidx, minAtime, maxAtime, minDtime, maxDtime, containerID);
									dIdxIdx.add(didPath, tmpBuf, minDidx, maxDidx, containerID);
								}
							}
							catch(IndexException e) {
								logger.error(
										"Container-Header ergibt Indexfehler und wird deshalb wieder hergestellt: " + didPath + " Container: " + contFile, e
								);
								rebuild = true;
							}
						}
						if(rebuild) {
							logger.fine("Vollstaendige Wiederherstellung ueber alle Datensaetze notwendig: " + didPath + " Container: " + contFile);
							cont_visited_fullrebuilt[1]++;
							minDidx = Long.MAX_VALUE;
							maxDidx = Long.MIN_VALUE;
							minAtime = Long.MAX_VALUE;
							maxAtime = Long.MIN_VALUE;
							minDtime = Long.MAX_VALUE;
							maxDtime = Long.MIN_VALUE;
							ContainerDataIterator iter = getArchiveData(task);
							while(iter.readNextDataHdr()) {
								long val = iter.getDataIndex();
								minDidx = val < minDidx ? val : minDidx;
								maxDidx = val > maxDidx ? val : maxDidx;
								val = iter.getArcTime();
								minAtime = val < minAtime ? val : minAtime;
								maxAtime = val > maxAtime ? val : maxAtime;
								val = iter.getDataTime();
								minDtime = val < minDtime ? val : minDtime;
								maxDtime = val > maxDtime ? val : maxDtime;
								iter.skipData();
							}
							iter.close();

							// Sämtliche Prüfungen der verschiedenen Indizes müssen durchgeführt werden, bevor irgendetwas verändert wird.
							boolean indexModificationPossible = false;
							try {
								// Zuerst wird geprüft, ob das Hinzufügen zu allen Indizes funktionieren würde
								dtidx.checkForAdd(didPath, tmpBuf, minDtime, maxDtime, minDidx, maxDidx, containerID);
								if(dIdxATIdx != null) {
									dIdxATIdx.checkForAdd(didPath, minDidx, maxDidx, minAtime, maxAtime, minDtime, maxDtime, containerID);
								}
								else {
									aTIdx.checkForAdd(didPath, minDidx, maxDidx, minAtime, maxAtime, minDtime, maxDtime, containerID);
									dIdxIdx.checkForAdd(didPath, tmpBuf, minDidx, maxDidx, containerID);
								}
								indexModificationPossible = true;
							}
							catch(IndexException e) {
								logger.error(
										"Wiederherstellung des Containerheaders hat das Indexproblem des Containers nicht behoben. Container wird ignoriert: "
										+ didPath + " Container: " + contFile, e
								);
							}

							if(indexModificationPossible) {
								// Erst wenn sicher ist, dass das Hinzufügen zu allen Indizes funktionieren würde, werden die Indizes aktualisiert
								dtidx.add(didPath, tmpBuf, minDtime, maxDtime, minDidx, maxDidx, containerID);
								if(dIdxATIdx != null) {
									dIdxATIdx.add(didPath, minDidx, maxDidx, minAtime, maxAtime, minDtime, maxDtime, containerID);
								}
								else {
									aTIdx.add(didPath, minDidx, maxDidx, minAtime, maxAtime, minDtime, maxDtime, containerID);
									dIdxIdx.add(didPath, tmpBuf, minDidx, maxDidx, containerID);
								}
								if(isContainerClosed(task)) {
									// das sollte nicht vorkommen
									setContainerHeaderParam(task, ContainerHdr.CHP_DATA_IDX_MIN, minDidx);
									setContainerHeaderParam(task, ContainerHdr.CHP_DATA_IDX_MAX, maxDidx);
									setContainerHeaderParam(task, ContainerHdr.CHP_DATA_TIME_MIN, minDtime);
									setContainerHeaderParam(task, ContainerHdr.CHP_DATA_TIME_MAX, maxDtime);
									setContainerHeaderParam(task, ContainerHdr.CHP_ARC_TIME_MIN, minAtime);
									setContainerHeaderParam(task, ContainerHdr.CHP_ARC_TIME_MAX, maxAtime);
									logger.warning("Container-Header musste wiederhergestellt werden: " + didPath + " Container: " + contFile);
								}
							}
						}
					}
					catch(PersistenceException e) {
						throw e;
					}
					finally {
						leaveContainer(task);
					}
				}
				dtidx.closeIndex(didPath, tmpBuf);
				if(dIdxATIdx != null) {
					dIdxATIdx.closeIndex(didPath);
				}
				else {
					aTIdx.closeIndex(didPath);
					dIdxIdx.closeIndex(didPath, tmpBuf);
				}
			}
		}
		return cont_visited_fullrebuilt;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Container: Datensätze
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Serialisiert den uebergebenen Datensatz und speichert ihn in {@link ContainerFile#actualDataBuf} und {@link ContainerFile#actualDataSize}. Da der ArchivTask
	 * vor dem Archivieren wissen muss, ob der Container zu lang wird und abgeschlossen werden muss, ist das Anhaengen eines Datensatzes auf die Methoden {@link
	 * ContainerFile#preserializeData()} und {@link ContainerFile#appendPreserializedData()} aufgeteilt.
	 *
	 * @param task    Anfragender Task
	 * @param newData Zu archivierender Datensatz
	 * @param arcTime Archivzeitstempel
	 *
	 * @return Laenge des serialisierten Datensatzes
	 *
	 * @throws PersistenceException
	 */
	public int serializeData(Task task, ResultData newData, long arcTime, Deflater compresser) throws PersistenceException {
		return getContainerFile(task).serializeData(newData, arcTime, compresser);
	}

	/**
	 * Serialisiert eine Datenluecke anstelle eines Datensatzes. Ansonsten Funktion indentisch zu {@link #serializeData(Task,ResultData,long)}.
	 *
	 * @param task        Anfragender Task
	 * @param dataIndex   Datenindex der Luecke (= letzter echter Index mit gesetztem Archivbit)
	 * @param dataTime    Datenzeit
	 * @param archiveTime Archivzeit
	 *
	 * @return Laenge des serialisierten Datensatzes
	 *
	 * @throws PersistenceException
	 */
	public int serializeDataGap(Task task, long dataIndex, long dataTime, long archiveTime, Deflater compresser) throws PersistenceException {
		return getContainerFile(task).serializeDataGap(dataIndex, dataTime, archiveTime, compresser);
	}

	/**
	 * Serialisiert den uebergebenen nachgeforderten Datensatz und speichert ihn in {@link ContainerFile#actualDataBuf} und {@link ContainerFile#actualDataSize}.
	 * Da der ArchivReqTask vor dem Archivieren wissen muss, ob der Container zu lang wird und abgeschlossen werden muss, ist das Anhaengen eines Datensatzes auf
	 * die Methoden {@link ContainerFile#preserializeData()} und {@link ContainerFile#appendPreserializedData()} aufgeteilt.
	 *
	 * @param task    Anfragender Task
	 * @param newData Zu archivierender Datensatz
	 * @param arcTime Archivzeitstempel
	 *
	 * @return Laenge des serialisierten Datensatzes
	 *
	 * @throws PersistenceException
	 */
	public int serializeReqData(Task task, ArchiveData newData, long arcTime, Deflater compresser) throws PersistenceException {
		return getContainerFile(task).serializeReqData(newData, arcTime, compresser);
	}

	/**
	 * Haengt den uebergebenen, serialisierten Archivdatensatz an den mit {@link #accessContainer(Task,long,long,long,int,ArchiveDataKind,long)} spezifizierten
	 * Container an.
	 *
	 * @param task Anfragender Task
	 *
	 * @throws PersistenceException
	 */
	public void appendSerializedData(Task task) throws PersistenceException {
		getContainerFile(task).appendSerializedData();
	}

	/**
	 * Liefert einen Iterator zum Durchlaufen aller Datensätze des aktuellen Containers.
	 *
	 * @param task Anfragender Task
	 *
	 * @return ContainerDataIterator für das Durchlaufen der Datensätze
	 *
	 * @throws PersistenceException
	 */
	public ContainerDataIterator getArchiveData(Task task) throws PersistenceException {
		return getContainerFile(task).iterGetDataIterator();
	}

	/**
	 * Liefert einen Iterator zum Durchlaufen aller Datensätze des aktuellen Containers.
	 *
	 * @param task            Anfragender Task
	 * @param considerDeleted gibt an, ob Sonderbehandlung für ausgelagerte Container nötig ist (wichtig beim Nachfordern).
	 *
	 * @return ContainerDataIterator für das Durchlaufen der Datensätze
	 *
	 * @throws PersistenceException
	 */
	public ContainerDataIterator getArchiveData(Task task, boolean considerDeleted) throws PersistenceException {
		return getContainerFile(task).iterGetDataIterator(considerDeleted);
	}

	/**
	 * Liefert einen Iterator zum Durchlaufen aller Datensätze des aktuellen Containers. (Evtl. mit Sonderbehandlung für ausgelagerte Container)
	 *
	 * @param task            Anfragender Task
	 * @param cfIdx           ContainerFile-Index
	 * @param considerDeleted Gibt an, ob Sonderbehandlung ausgelagerter Container nötig ist (wichtig beim Nachfordern).
	 *
	 * @return ContainerDataIterator für das Durchlaufen der Datensätze
	 *
	 * @throws PersistenceException
	 */
	public ContainerDataIterator getArchiveData(Task task, long cfIdx, boolean considerDeleted) throws PersistenceException {
		return getContainerFile(task, cfIdx).iterGetDataIterator(considerDeleted);
	}

	/**
	 * Liefert einen Iterator zum Durchlaufen aller Datensätze des aktuellen Containers. Damit ein Task mehrere Container gleichzeitig parallel offen halten kann,
	 * wird der cfIdx zur Unterscheidung der ContainerFile-Objekte benutzt.
	 *
	 * @param task  Anfragender Task
	 * @param cfIdx ContainerFile-Index
	 *
	 * @return ContainerDataIterator für das Durchlaufen der Datensätze
	 *
	 * @throws PersistenceException
	 */
	public ContainerDataIterator getArchiveData(Task task, long cfIdx) throws PersistenceException {
		return getContainerFile(task, cfIdx).iterGetDataIterator();
	}

	/**
	 * Liefert einen Iterator zum Durchlaufen der Min-/Maxwerte des Container-Headers Damit ein Task mehrere Container gleichzeitig parallel offen halten kann,
	 * wird der cfIdx zur Unterscheidung der ContainerFile-Objekte benutzt.
	 *
	 * @param task  Anfragender Task
	 * @param cfIdx ContainerFile-Index
	 *
	 * @return ContainerDataIterator für das Durchlaufen der Min-/Maxwerte des Container-Headers
	 *
	 * @throws PersistenceException
	 */
	public ContainerDataIterator getHdrData(Task task, long cfIdx) throws PersistenceException {
		return getContainerFile(task, cfIdx).iterGetHeaderIterator();
	}

	/**
	 * Zaehlt die Anzahl der Datensätze in dem mit einer <code>accessContainer()</code>-Methode im Zugriff befindlichen Container durch Iterieren ueber die
	 * Datensätze in der Container-Datei. Diese Methode ist nur notwendig, wenn ein noch nicht abgeschlossener Container wieder befuellt wird, weil dann die
	 * Anzahl nicht im Header steht. Im Normalbetrieb wird die Anzahl im DataIdentNode mitgefuehrt.
	 *
	 * @param task Anfragender Task
	 *
	 * @return Anzahl Datensätze im Container
	 *
	 * @throws PersistenceException
	 */
	public int countContainerEntries(Task task) throws PersistenceException {
		return getContainerFile(task).countContainerEntries();
	}

	/**
	 * Ermittelt die Gesamtlänge des mit <code>accessContainer()</code>- im Zugriff befindlichen Containers.
	 *
	 * @return Länge der Container-Datei
	 *
	 * @throws PersistenceException
	 */
	public long getContainerSize(Task task) throws PersistenceException {
		return getContainerFile(task).getContainerSize();
	}

	/**
	 * Loescht alle Datensätze eines Datencontainers, so dass lediglich der Header mit der Verwaltungs-Information in der Datei erhalten bleibt.<br> ACHTUNG:<br>
	 * Das Setzen des Header-Flags auf "geloescht" muss NACH dem Loeschen der Daten aus dem Container erfolgen!
	 *
	 * @param task Anfragender Task
	 *
	 * @throws PersistenceException
	 */
	public void deleteContainerData(Task task) throws PersistenceException {
		try {
			getContainerFile(task).deleteContainerData();
		}
		catch(Exception e) {
			logger.warning("Container konnte nicht gelöscht werden", e);
		}
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Prüfprozedur für das Persistenzverzeichnis
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Das Logfile {@link #isActive} anlegen. Wird verwendet, um einem Archivsystem anzuzeigen, ob ein anderes Archivsystem dasselbe Persistenzverzeichnis
	 * verwendet
	 *
	 * @return Fehlermeldung, wenn das Erstellen nicht erfolgreich war, null sonst
	 */
	public String createLockFile() {
		isActive = new File(archivePath, ACTIVITY_FLAG_FILE_NAME);
		if(isActive.exists()) {
			return "Die Sperrdatei \"" + isActive.getAbsolutePath() + "\" zeigt an, dass bereits ein Archivsystem im Persistenzverzeichnis aktiv ist.";
		}
		try {
			if(isActive.createNewFile()) {
				return null;
			}
			else {
				return "Sperrdatei \"" + isActive.getAbsolutePath() + "\" konnte nicht erstellt werden.";
			}
		}
		catch(Exception e) {
			return "Fehler beim Erstellen der Sperrdatei \"" + isActive.getAbsolutePath() + "\": " + e.getMessage();
		}
	}

	/**
	 * Löscht das LogFile {@link #isActive}, falls es von diesem ArS angelegt wurde und es existiert und gibt damit das Persistenzverzeichnis wieder frei
	 *
	 * @return true, wenn das File erfolgreich gelöscht werden konnte
	 */
	public boolean deleteLockFile() {
		if(isActive == null) {
			return true;
		}
		else {
			if(isActive.delete()) {
				return true;
			}
			else {
				logger.warning(
						"Sperrdatei \"" + isActive.getAbsolutePath() + "\" konnte zur Freigabe des " + "Persistenzverzeichnisses nicht gelöscht werden."
				);
				return false;
			}
		}
	}

	/**
	 * Prueft das Persistenzverzeichnis auf fremde Dateien und Verzeichnisse und falsche Reihenfolgen von ID-Verzeichnissen. Die Ergebnisse werden ueber den Logger
	 * protokolliert. Kann durch einen Kommandozeilenparameter beim Systemstart aufgerufen werden.
	 */
	public void checkPersistenceDir(String mode) throws InterruptedException {
		AtomicBoolean allValid = new AtomicBoolean(true);
		AtomicInteger numVisited = new AtomicInteger(0);
		directoryCount = 0;
		DataIdentTree countingDit = new DataIdentTree();
		File archiveDir = new File(archivePath);

		ExecutorService threadPool = Executors.newFixedThreadPool(archivMgr.getNumCheckPersistenceThreads());

		File[] files = archiveDir.listFiles();
		if(files == null) files = new File[0];
		for(File dir : files) {
			boolean valid = isValidTopLevelArSFile(dir.getName());
			if(!valid) logger.fine("Fremde Datei/Verzeichnis im Persistenzverzeichnis: '" + dir.getAbsolutePath() + "'");
			if(!valid) allValid.set(false);
			if(dir.isDirectory()) {
				checkPersDirNode(dir, countingDit, mode, threadPool, allValid, numVisited);
			}
		}

		threadPool.shutdown();

		while(!threadPool.awaitTermination(10, TimeUnit.SECONDS)) {
			logger.info(numVisited + " Datenidentifikationen analysiert");
		}

		logger.info(
				"# " + directoryCount + " durchsuchte Verzeichisse" + Debug.NEWLINE + "# " + countingDit.countDataIdentNodes()
				+ " Datenidentifikationen gefunden." + Debug.NEWLINE + "# Ergebnis: " + (allValid.get()
				                                                                         ? "Das Archiv ist in einem konsistenten Zustand."
				                                                                         : "Es sind Fehler aufgetreten (siehe LogLevel FINE, FINER und FINEST für Details).")
		);
	}

	private void checkPersDirNode(File curDir, DataIdentTree dit, String mode, final ExecutorService threadPool, final AtomicBoolean allValid, final AtomicInteger numVisited) {
		if(++directoryCount % 5000 == 0) logger.info(directoryCount + " Verzeichisse indiziert");

		if(!isValidStorageArSFile(curDir.getName())) {
			logger.fine("Fremdes Verzeichnis im Persistenzverzeichnis: '" + curDir.getAbsolutePath() + "'");
			allValid.set(false);
		}

		if(curDir.isDirectory()) {
			if(isValidSvFolderName(curDir.getName())) {
				threadPool.submit(new ConsistencyCheckWorker(curDir, dit, mode, allValid, numVisited));
			}
			else {
				File[] content = curDir.listFiles();
				if(content != null) {
					for(File dir : content) {
						checkPersDirNode(dir, dit, mode, threadPool, allValid, numVisited);
					}
				}
			}
		}
	}

	/**
	 * Extrahiert Objekt-ID, Attributgruppen-ID, Aspekt-ID und SimVar von unten beginnend aus einem Verzeichnispfad, der mindestens die Laenge 4 haben muss.
	 * Voraussetzung ist, dass alle Verzeichnisnamen entsprechend den Konventionen des PersistenceManagers gebildet sind. Das unterste Verzeichnis kann entweder
	 * die SimVar bezeichnen oder eine Datensatzart sein.
	 *
	 * @param path Pfad bis zum Verzeichnis der Simulationsvariante
	 *
	 * @return Feld bestehend aus [ObjID, AtgID, AspID SimVar] wenn der Pfad ein gueltiges Verzeichnis ist
	 *
	 * @throws PersistenceException wenn der Pfad kein gueltiges Verzeichnis ist
	 * @see PersistenceManager#isValidObjIdFolderName(String)
	 * @see PersistenceManager#isValidAtgIdFolderName(String)
	 * @see PersistenceManager#isValidAspIdFolderName(String)
	 * @see PersistenceManager#isValidSvFolderName(String)
	 */
	public static long[] extractDIdFromPath(File path) throws PersistenceException {
		// split mit File.separator geht schief unter Windows (wg. \\), deswegen ueber File/Parent

		long[] erg = new long[]{0, 0, 0, 0};
		File tmpDir = path;
		if(tmpDir != null) {
			try {
				if(isValidDataKindFolderName(tmpDir.getName())) tmpDir = tmpDir.getParentFile();

				if(isValidSvFolderName(tmpDir.getName())) {
					erg[3] = Integer.parseInt(getIDPart(tmpDir.getName()));
				}
				else {
					throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.getAbsolutePath() + "'");
				}

				tmpDir = tmpDir.getParentFile();
				String tmp[] = new String[]{"", "", ""};
				do {
					if(tmpDir == null) throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.getAbsolutePath() + "'");
					if(isValidAspIdFolderName(tmpDir.getName())) {
						tmp[2] = getIDPart(tmpDir.getName()) + tmp[2];
					}
					else {
						break;
					}
				}
				while((tmpDir = tmpDir.getParentFile()) != null);
				do {
					if(tmpDir == null) throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.getAbsolutePath() + "'");
					if(isValidAtgIdFolderName(tmpDir.getName())) {
						tmp[1] = getIDPart(tmpDir.getName()) + tmp[1];
					}
					else {
						break;
					}
				}
				while((tmpDir = tmpDir.getParentFile()) != null);
				do {
					if(tmpDir == null) throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.getAbsolutePath() + "'");
					if(isValidObjIdFolderName(tmpDir.getName())) {
						tmp[0] = getIDPart(tmpDir.getName()) + tmp[0];
					}
					else {
						break;
					}
				}
				while((tmpDir = tmpDir.getParentFile()) != null);

				for(int i = 0; i < tmp.length; i++) {
					erg[i] = Long.parseLong(tmp[i]);
				}
				return erg;
			}
			catch(NumberFormatException nex) {
			}
		}
		throw new PersistenceException("Ungueltiger Pfadname einer Datenidentifikation: '" + path.getAbsolutePath() + "'");
	}

	/**
	 * Prueft, ob der gegebene Verzeichnisname ein gueltiger Name innerhalb eines Pfades ist, der eine Objekt-ID repraesentiert.
	 *
	 * @param dirName Verzeichisname
	 *
	 * @return Wahr, wenn der Verzeichnisname korrekt ist, falsch sonst.
	 *
	 * @see PersistenceManager#OBJ_DIRNAME_PAT
	 */
	public static boolean isValidObjIdFolderName(String dirName) {
		return OBJ_DIRNAME_PAT.matcher(dirName).matches();
	}

	/**
	 * Prueft, ob der gegebene Verzeichnisname ein gueltiger Name innerhalb eines Pfades ist, der eine Attributgruppen-ID repraesentiert.
	 *
	 * @param dirName Verzeichisname
	 *
	 * @return Wahr, wenn der Verzeichnisname korrekt ist, falsch sonst.
	 *
	 * @see PersistenceManager#ATG_DIRNAME_PAT
	 */
	public static boolean isValidAtgIdFolderName(String dirName) {
		return ATG_DIRNAME_PAT.matcher(dirName).matches();
	}

	/**
	 * Prueft, ob der gegebene Verzeichnisname ein gueltiger Name innerhalb eines Pfades ist, der eine Aspekt-ID repraesentiert.
	 *
	 * @param dirName Verzeichisname
	 *
	 * @return Wahr, wenn der Verzeichnisname korrekt ist, falsch sonst.
	 *
	 * @see PersistenceManager#ASP_DIRNAME_PAT
	 */
	public static boolean isValidAspIdFolderName(String dirName) {
		return ASP_DIRNAME_PAT.matcher(dirName).matches();
	}

	/**
	 * Prueft, ob der gegebene Verzeichnisname ein gueltiger Name für ein Verzeichnis ist, das eine Simulationsvariante repraesentiert.
	 *
	 * @param dirName Verzeichisname
	 *
	 * @return Wahr, wenn der Verzeichnisname korrekt ist, falsch sonst.
	 *
	 * @see PersistenceManager#SV_DIRNAME_PAT
	 */
	public static boolean isValidSvFolderName(String dirName) {
		return SV_DIRNAME_PAT.matcher(dirName).matches();
	}

	public static boolean isValidDataKindFolderName(String dirName) {
		return dirName.length() == 2 && VALID_SUFFIXES.contains(dirName);
	}

	public static boolean isValidIDFolderName(String dirName) {
		return isValidObjIdFolderName(dirName) || isValidAtgIdFolderName(dirName) || isValidAspIdFolderName(dirName) || isValidSvFolderName(dirName);
	}

	public static String getIDPart(String dirName) {
		Matcher m = NUM_DIR_PAT.matcher(dirName);
		m.find();
		return m.group(1);
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Private Methoden
	//////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Liefert den Archiv-Pfad für die angegebene Datenidentifikation. Wenn das Verzeichnis nicht existiert, wird es angelegt (lazy-Verhalten). Die Methode ist
	 * synchronized weil sie von mehreren Tasks aufgerufen wird und mkdirs in manchen Konstellationen false liefert.
	 *
	 * @param objId Objekt-ID
	 * @param atgId Attributgruppen-ID
	 * @param aspId Aspekt-ID
	 * @param sv    Simulationsvariante
	 *
	 * @return Archiv-Pfad
	 *
	 * @throws PersistenceException
	 */
	private synchronized File getCreatePersDir(long objId, long atgId, long aspId, int sv, ArchiveDataKind archiveDataKind) throws PersistenceException {
		File pDir = new File(getPathName(objId, atgId, aspId, sv, archiveDataKind));
		if(!pDir.exists()) {
			if(!pDir.mkdirs()) throw new PersistenceException("Persistenzverzeichnis '" + pDir.getAbsolutePath() + "' konnte nicht angelegt werden.");
		}
		return pDir;
	}

	/**
	 * Liefert ein File für das angegebene Verzeichnis. Wenn das Verzeichnis nicht existiert, wird es angelegt (lazy-Verhalten). Die Methode ist
	 * synchronized weil sie von mehreren Tasks aufgerufen wird und mkdirs in manchen Konstellationen false liefert.
	 *
	 * @param pathName
	 *
	 * @return Archiv-Pfad
	 *
	 * @throws PersistenceException
	 */
	private synchronized File getCreatePersDir(String pathName) throws PersistenceException {
		File pDir = new File(pathName);
		if(!pDir.exists()) {
			if(!pDir.mkdirs()) throw new PersistenceException("Persistenzverzeichnis '" + pDir.getAbsolutePath() + "' konnte nicht angelegt werden.");
		}
		return pDir;
	}

	/**
	 * Liefert das ContainerFile-Objekt für die angegebene anfragende Task.
	 *
	 * @param task Task
	 *
	 * @return ContainerFile
	 */
	private ContainerFile getContainerFile(Task task) {
		return getContainerFile(task, 0);
	}

	/**
	 * Liefert das ContainerFile-Objekt mit dem angegebenen Index für die angegebene anfragende Task.
	 *
	 * @param task  Anfragender Task
	 * @param cfIdx ContainerFile-Index
	 *
	 * @return ContainerFile
	 */
	private ContainerFile getContainerFile(Task task, long cfIdx) {
		synchronized(containerFileMap) {
			Map<Long, ContainerFile> cfMap = containerFileMap.get(task);
			if(cfMap == null) {
				cfMap = new HashMap<>();
				containerFileMap.put(task, cfMap);
				ContainerFile cf = new ContainerFile();
				cfMap.put(cfIdx, cf);
				return cf;
			}
			else {
				ContainerFile cf = cfMap.get(cfIdx);
				if(cf == null) {
					cf = new ContainerFile();
					cfMap.put(cfIdx, cf);
				}
				return cf;
			}
		}
	}

	/**
	 * Leert die Hashtabelle für die {@link ContainerFile}-Objekte. Dies ist notwendig, wenn (zu Testzwecken) ein neuer {@link de.bsvrz.ars.ars.mgmt.InQueuesMgr}
	 * angelegt wird.
	 */
	public void resetContainerFileMap() {
		synchronized(containerFileMap) {
			containerFileMap.clear();
		}
	}

	/**
	 * Liefert den Wurzelpfad des Persistenzverzeichisses, der beim Systemstart in der Kommandozeile uebergeben wurde.
	 *
	 * @return Wurzelpfad des Persistenzverzeichisses
	 */
	public String getArchivePath() {
		return archivePath;
	}

	/**
	 * Erhoeht nextContainerID um 1 und gibt den Wert zurueck.
	 *
	 * @return als naechste zu verwendende Container-ID
	 */
	public long nextContainerID() {
		return nextContainerID.incrementAndGet();
	}

	/** @return Letzte vergebe ContainerID */
	public long getLastContainerID() {
		return nextContainerID.longValue();
	}

	/**
	 * Verlaengert den Loschschutz gemaess TAnfArS 5.1.2.4.2
	 *
	 * @param task    anfragender Task
	 * @param t_ls    Zeitdauer, um die der LS verlaengert wird (aus atg.archivEinstellung)
	 * @param t_lsMax maximale Zeitdauer, um die verlaengert werden darf (gemessen ab Vorhaltezeitraum)
	 *
	 * @throws PersistenceException
	 */
	public void setAutoExtendLoeschutz(Task task, int cfIdx, long t_ls, long t_lsMax) throws PersistenceException {
		long t_lsMillis = t_ls * 1000;
		long t_lsMaxMillis = t_lsMax * 1000;
		long now = System.currentTimeMillis();
		ContainerFile cf = getContainerFile(task, cfIdx);

		long vorhalte = cf.getContHdrParamAsLong(ContainerHdr.CHP_LOESCHEN);
		if(vorhalte < now + t_lsMillis) {
			long loeschutz = cf.getContHdrParamAsLong(ContainerHdr.CHP_LOESCHUTZ);
			long loeschutzNeu = Math.min(now + t_lsMillis, vorhalte + t_lsMaxMillis);	// insgesamt um max. t_lsMax ab vorhalte
			if(loeschutzNeu > loeschutz) cf.setContHdrParam(ContainerHdr.CHP_LOESCHUTZ, loeschutzNeu);
		}
	}

	/**
	 * Verlaengert den Loschschutz gemaess TAnfArS 5.1.2.4.2
	 *
	 * @param task    anfragender Task
	 * @param t_ls    Zeitdauer, um die der LS verlaengert wird (aus atg.archivEinstellung)
	 * @param t_lsMax maximale Zeitdauer, um die verlaengert werden darf (gemessen ab Vorhaltezeitraum)
	 *
	 * @throws PersistenceException
	 */
	public void setAutoExtendLoeschutz(Task task, long t_ls, long t_lsMax) throws PersistenceException {
		setAutoExtendLoeschutz(task, 0, t_ls, t_lsMax);
	}

	private static final ThreadLocal<byte[]> _tmpMem = ThreadLocal.withInitial(() -> new byte[1024 * 16]);

	private class IndexCloserWorker implements Runnable {
		private final AtomicInteger _anzOA;
		private final AtomicInteger _anzON;
		private final AtomicInteger _anzNA;
		private final AtomicInteger _anzNN;

		private final DataIdentNode _din;
		private final IdDataIdentification _dataIdentification;

		public IndexCloserWorker(final AtomicInteger anzOA, final AtomicInteger anzON, final AtomicInteger anzNA, final AtomicInteger anzNN, final DataIdentNode din, final IdDataIdentification dataIdentification) {
			_anzOA = anzOA;
			_anzON = anzON;
			_anzNA = anzNA;
			_anzNN = anzNN;
			_din = din;
			_dataIdentification = dataIdentification;
		}

		@Override
		public void run() {
			byte[] tmpMem = _tmpMem.get();
			try {
				_anzOA.addAndGet(closeIndex(_din, _dataIdentification, ArchiveDataKind.ONLINE, tmpMem) ? 1 : 0);
				_anzON.addAndGet(closeIndex(_din, _dataIdentification, ArchiveDataKind.ONLINE_DELAYED, tmpMem) ? 1 : 0);
				_anzNA.addAndGet(closeIndex(_din, _dataIdentification, ArchiveDataKind.REQUESTED, tmpMem) ? 1 : 0);
				_anzNN.addAndGet(closeIndex(_din, _dataIdentification, ArchiveDataKind.REQUESTED_DELAYED, tmpMem) ? 1 : 0);
			}
			catch(IndexException e){
				logger.error("Herunterfahren: Fehler beim Schliessen der Indexe: ", e);
			}
		}
	}

	private class ConsistencyCheckWorker implements Runnable {
		private final File _curDir;
		private final DataIdentTree _dit;
		private final String _mode;
		private final AtomicBoolean _allValid;
		private final AtomicInteger _numVisited;

		public ConsistencyCheckWorker(final File curDir, final DataIdentTree dit, final String mode, final AtomicBoolean allValid, final AtomicInteger numVisited) {
			_curDir = curDir;
			_dit = dit;
			_mode = mode;
			_allValid = allValid;
			_numVisited = numVisited;
		}

		public void run() {
			try {
				long[] did = extractDIdFromPath(_curDir);
				DataIdentNode din = new DataIdentNode();
				_dit.putDataIdentNode(did[0], did[1], did[2], (int)did[3], din);
				File[] content = _curDir.listFiles();
				if(content != null) {
					for(File dir : content)         // Ordner der Datensatzarten (oa, on, na, nn) prüfen:
					{
						if(!isValidDataKindFolderName(dir.getName())) {
							logger.fine("Fremde Datei/Verzeichnis im Persistenzverzeichnis: '" + dir.getAbsolutePath() + "'");
							_allValid.set(false);
						}
						else {
							if(_mode.equalsIgnoreCase(ArchiveManager.P_CHECK_PERS_DIR_HDR)) {
								int retCode = ConsistencyCheck.simpleCheck(dir, logger);
								if((retCode & ConsistencyCheck.WARN) == ConsistencyCheck.WARN) {
									logger.fine("Es sind Warnungen im Verzeichnis: '" + dir.getAbsolutePath() + "' aufgetreten.");
								}
								if((retCode & ConsistencyCheck.ERROR) == ConsistencyCheck.ERROR) {
									logger.fine("Es sind Fehler im Verzeichnis: '" + dir.getAbsolutePath() + "' aufgetreten.");
									_allValid.set(false);
								}
							}
							else if(_mode.equalsIgnoreCase(ArchiveManager.P_CHECK_PERS_DIR_ALL)) {
								int retCode = ConsistencyCheck.extendedCheck(dir, logger);
								if((retCode & ConsistencyCheck.WARN) == ConsistencyCheck.WARN) {
									logger.fine("Es sind Warnungen im Verzeichnis: '" + dir.getAbsolutePath() + "' aufgetreten.");
								}
								if((retCode & ConsistencyCheck.ERROR) == ConsistencyCheck.ERROR) {
									logger.fine("Es sind Fehler im Verzeichnis: '" + dir.getAbsolutePath() + "' aufgetreten.");
									_allValid.set(false);
								}
							}
						}
					}
				}
			}
			catch(PersistenceException e) {
				logger.fine("Pfad der Datenidentifikation ungueltig: '" + _curDir.getAbsolutePath() + "'");
				_allValid.set(false);
			}
			finally {
				_numVisited.incrementAndGet();
			}
		}
	}
}
