/*
 *
 * Copyright 2017 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.iter;

import de.bsvrz.ars.ars.persistence.ContainerDataResult;
import de.bsvrz.ars.ars.persistence.PersistenceException;

/**
 * Ein Iterator über die Datensätze von mehreren Containern. Der Iterator verhält sich anders als der Standard-Java-iterator und mehr wie
 * eine Queue. {@link #peek(ContainerDataResult) peek} rufe den aktuellen Datensatz am Kopf der Queue ab, {@link #remove() remove} entfernt diesen,
 * sodass der nächste Datensatz abgerufen werden kann (falls vorhanden). {@link #poll(ContainerDataResult)} gibt den aktuellen Datensatz zurück und entfernt ihn in einem Schritt.
 *
 * Die Iteration mit dieser Klasse erfolg wie beispielsweise folgt:
 * <tt><pre>
 *     ContainerDataResult result = new ContainerDataResult()
 *     while(!iter.isEmpty()) {
 *         iter.poll(result);
 *         // tue was mit result
 *     }
 *     iter.close()
 * </pre></tt>
 *
 * @author Kappich Systemberatung
 */
public interface DataIterator extends AutoCloseable {

	/**
	 * Kopiert die Daten des aktuellen Elements in das angegebene Result-Objekt. Dies ist eine Optimierung, damit nicht ständig neue Result-Objekt angelegt werden müssen.
	 * @param result Ziel-Element
	 * @throws java.util.NoSuchElementException falls kein Element mehr Verfügbar (isEmpty == true)
	 */
	void peek(ContainerDataResult result) throws PersistenceException;

	/**
	 * Ruft den aktuellen Datenindex ab. Entspricht
	 * <tt><pre>
	 *     ContainerDataResult result = new ContainerDataResult()
	 *     this.peek(result)
	 *     return result.getDataIndex()
	 * </pre></tt>
	 * ist aber wahrscheinlich performanter.
	 * @return Datenindex des aktuellen Elements
	 * @throws java.util.NoSuchElementException falls kein Element mehr Verfügbar (isEmpty == true)
	 */
	long peekDataIndex();

	/**
	 * Ruft die aktuelle Datenzeit ab. Entspricht
	 * <tt><pre>
	 *     ContainerDataResult result = new ContainerDataResult()
	 *     this.peek(result)
	 *     return result.getDataTime()
	 * </pre></tt>
	 * ist aber wahrscheinlich performanter.
	 * @return Datenzeit des aktuellen Elements
	 * @throws java.util.NoSuchElementException falls kein Element mehr Verfügbar (isEmpty == true)
	 */
	long peekDataTime();

	/**
	 * Ruft die aktuelle Archivzeit ab. Entspricht
	 * <tt><pre>
	 *     ContainerDataResult result = new ContainerDataResult()
	 *     this.peek(result)
	 *     return result.getArchiveTime()
	 * </pre></tt>
	 * ist aber wahrscheinlich performanter.
	 * @return Archivzeit des aktuellen Elements
	 * @throws java.util.NoSuchElementException falls kein Element mehr Verfügbar (isEmpty == true)
	 */
	long peekArchiveTime();

	/**
	 * Entspricht dem Aufruf von `peek(result)´ gefolgt von `remove()`.
	 * @param result
	 */
	default void poll(ContainerDataResult result) throws PersistenceException {
		peek(result);
		remove();
	}

	/**
	 * Entfernt das aktuelle Element.
	 */
	void remove() throws PersistenceException;

	/**
	 * Gibt <tt>true</tt> zurück, wenn der Iterator erschöpft ist. Ist das der Fall liefern alle anderen Funktionen außer close() {@link java.util.NoSuchElementException}s.
	 * @return <tt>true</tt>, wenn der Iterator erschöpft ist, sonst <tt>false</tt>
	 */
	boolean isEmpty();

	/**
	 * Terminiert das iterieren. Hierdurch werden alle ggf. noch offenen Containerdateien geschlossen.
	 * @throws PersistenceException
	 */
	void close() throws PersistenceException;

	long getCurContFileIdx();

	static DataIterator empty() {
		return EmptyDataIterator.getInstance();
	}
}
